/*******************************************************************************
* fontsel.c -- Nirvana Font Selector main program			       *
* Copyright (c) 1993 Universities Research Association, Inc.		       *
* All rights reserved.							       *
* This material resulted from work developed under a Government Contract and   *
* is subject to the following license:  The Government retains a paid-up,      *
* nonexclusive, irrevocable worldwide license to reproduce, prepare derivative *
* works, perform publicly and display publicly by or for the Government,       *
* including the right to distribute to other Government contractors.  Neither  *
* the United States nor the United States Department of Energy, nor any of     *
* their employees, makes any warrenty, express or implied, or assumes any      *
* legal liability or responsibility for the accuracy, completeness, or         *
* usefulness of any information, apparatus, product, or process disclosed, or  *
* represents that its use would not infringe privately owned rights.           *
* Fermilab Nirvana GUI Library						       *
* June 2, 1993								       *
* Written by Suresh Ravoor (assisted by Mark Edel)			       *
*******************************************************************************/

#define FONTSEL

#include <stdlib.h>
#include <X11/Intrinsic.h>
#include <Xm/Xm.h>
#include <Xm/Protocols.h>
#include <Xm/Form.h>
#include <Xm/PushB.h>
#include <Xm/List.h>
#include <Xm/Label.h>
#include <Xm/Text.h>
#include <Xm/ToggleB.h>
#include <Xm/MessageB.h>
#include <Xm/DialogS.h>
#include <Xm/AtomMgr.h>
#include <math.h>
#include "paint.h"

#define	MAX_ARGS			20
#define MAX_NUM_FONTS   		32767
#define MAX_FONT_NAME_LEN		256
#define MAX_ENTRIES_IN_LIST		512
#define MAX_DISPLAY_SIZE		150
#define DELIM				'-'
#define NUM_COMPONENTS_FONT_NAME	14
#define TEMP_BUF_SIZE			256
#define DISPLAY_HEIGHT			90

enum listSpecifier { NONE, FONT, STYLE, SIZE };

typedef struct
{
  Widget form;
  Widget okButton;
  Widget cancelButton;
  Widget fontList;
  Widget styleList;
  Widget sizeList;
  Widget fontNameField;
  Widget sizeToggle;
  Widget propFontToggle;
  Widget dispLabel;
  Widget sizeFld;
  Widget dimensionLbl;
  Widget aspectRatioTB;
  char **fontData;	/* font name info  */
  int numFonts;		/* number of fonts */
  char *sel1;		/* selection from list 1 */
  char *sel2;		/* selection from list 2 */
  char *sel3;		/* selection from list 3 */
  int showPropFonts;	/* toggle state - show prop fonts */
  int showSizeInPixels;	/* toggle state - size in pixels  */
  char *fontName;	/* current font name */
  XFontStruct *oldFont;	/* font data structure for dispSample */
  XmFontList oldFontList;/* font data structure for dispSample */
  int exitFlag;		/* used for program exit control */
  XFontStruct *retFont;	/* font data to be returned */
  int scalableFont;	/* current font is scalable */
  inpNumStruct *sizeFldInd;/* index of sizeFld in inputArray */
  int textHeight, textWidth;/* width & height of text */
  int keepAspectRatio;	/* keep aspect ratio */
  float maxAspRatioDiff;/* maximal allowed difference in aspect ratio */
} xfselControlBlkType;

/* statically allocated structure which is passed between functions */
static xfselControlBlkType CtrlB;

static struct
{
  Widget drawArea;
  Widget scrollWin;
  Widget clipWinW;
  Widget scrollBX, scrollBY;
  Widget dialogW;
  Rectangle clipWin;
  int boxWidth, boxHeight;
  XFontStruct *font;
  unsigned int DAwidth, DAheight;
  unsigned int charNo;
  unsigned short existing_chars[256];
} charTable;

/* local function prototypes */
static void getStringComponent(char *inStr, int pos, char *outStr);
static void setupScrollLists(int dontChange, xfselControlBlkType CtrlBlk);
static int notPropFont(char *font);
static int styleMatch(xfselControlBlkType CtrlBlk, char *font);
static int sizeMatch(xfselControlBlkType CtrlBlk, char *font);
static int fontMatch(xfselControlBlkType CtrlBlk, char *font);
static int aspectRatioMatch(xfselControlBlkType CtrlBlk, char *font);
static void addItemToList(char **buf, char *item, int *count, int lexOrder);
static void getFontPart(char *font, char *buff1);
static void getStylePart(char *font, char *buff1);
static void getSizePart(char *font, char *buff1, int inPixels);
static void propFontToggleAction(Widget widget, xfselControlBlkType *CtrlBlk,
 XmToggleButtonCallbackStruct *call_data);
static void fontAction(Widget widget, xfselControlBlkType *CtrlBlk,
 XmListCallbackStruct *call_data);
static void styleAction(Widget widget, xfselControlBlkType *CtrlBlk,
 XmListCallbackStruct *call_data);
static void sizeAction(Widget widget, xfselControlBlkType *CtrlBlk,
 XmListCallbackStruct *call_data);
static void choiceMade(xfselControlBlkType *CtrlBlk);
static void dispSample(xfselControlBlkType *CtrlBlk, Boolean initTable);
static void cancelAction(Widget widget, xfselControlBlkType *CtrlBlk,
 XmListCallbackStruct *call_data);
static void okAction(Widget widget, xfselControlBlkType *CtrlBlk,
 XmPushButtonCallbackStruct *call_data);
static void startupFont(xfselControlBlkType *CtrlBlk, char *font);
static void FindBigFont(xfselControlBlkType *CtrlBlk, char *bigFont);
static void aspectRatioToggleAction(Widget widget, xfselControlBlkType *CtrlBlk,
 XmToggleButtonCallbackStruct *call_data);
static void displayText(xfselControlBlkType *CtrlBlk);

static void freeXmString(XmString *string)
{
 if (*string)
  { XmStringFree(*string); *string = NULL; }
}

/* Set up closeCB to be called when the user selects close from the
   window menu.  The close menu item usually activates f.kill which
   sends a WM_DELETE_WINDOW protocol request for the window. */
static void AddMotifCloseCallback (Widget shell, XtCallbackProc closeCB, void *arg)
{
 static Atom wmpAtom, dwAtom = 0;
 Display *display = XtDisplay (shell);

 /* deactivate the built in delete response of killing the application */
 XtVaSetValues (shell, XmNdeleteResponse, XmDO_NOTHING, 0);
 /* add a delete window protocol callback instead */
 if (dwAtom == 0)
  {
   wmpAtom = XInternAtom (display, "WM_PROTOCOLS", TRUE);
   dwAtom = XInternAtom (display, "WM_DELETE_WINDOW", TRUE);
  }
 XmAddProtocolCallback (shell, wmpAtom, dwAtom, closeCB, arg);
}

/* from the txt string eliminate all non displayable chars */
static void checkText(XFontStruct *font, unsigned char *txt)
{
 int l, i;
 unsigned char *cp;

 if ((txt == NULL) || (font == NULL))
   return;
 l = strlen (txt);
 if (l == 0)
   return;
 cp = txt;
 for (i = 0; i < l; i++)
  if ((txt[i] < font->min_char_or_byte2) || (txt[i] > font->max_char_or_byte2)
   || (charTable.existing_chars[txt[i]] == 0))
   continue;
  else
   *cp++ = txt[i];
 *cp = 0;
}

static void exposeCharTbl(Rectangle rect)
{
 int rowSt, colSt, rowNo, colNo, i, j, x, y, lmargin, oldWidth;
 unsigned int currCh, charWidth;
 unsigned long foreground;
 char oneChar[2];
 XGCValues values;
 Font oldFont;
 Boolean whiteRefresh;

 if ((CtrlB.sel1 == NULL) || (CtrlB.sel2 == NULL) ||
     (CtrlB.sel3 == NULL) || (charTable.font == NULL))
   whiteRefresh = True;
 else
  whiteRefresh = False;
 getClipWinDims(charTable.clipWinW, charTable.scrollBX, charTable.scrollBY,
  charTable.DAwidth, charTable.DAheight, &charTable.clipWin);
 XGetGCValues (xinfo.dpy, xinfo.gc, GCLineWidth | GCFont | GCForeground, &values);
 oldFont = values.font;
 oldWidth = values.line_width;
 foreground = values.foreground;
 XSetForeground(xinfo.dpy, xinfo.gc, xinfo.white);
 XFillRectangle(xinfo.dpy, XtWindow (charTable.drawArea), xinfo.gc,
  rect.x, rect.y, rect.width, rect.height);
 if (!whiteRefresh)
  {
   values.font = charTable.font->fid;
   values.line_width = 0;
   values.foreground = xinfo.black;
   XChangeGC(xinfo.dpy, xinfo.gc, GCLineWidth | GCFont | GCForeground, &values);

   rowSt = rect.y / charTable.boxHeight;
   colSt = rect.x / charTable.boxWidth;
   rowNo = (rect.height + charTable.boxHeight - 1) / charTable.boxHeight;
   colNo = (rect.width + charTable.boxWidth - 1) / charTable.boxWidth;

   oneChar[1] = 0;
   y = (rowSt + 1) * charTable.boxHeight;
   for (j = 0; j < rowNo; j++)
    {
     currCh = (rowSt + j) * 16 + colSt;
     x = colSt * charTable.boxWidth;
     for (i = 0; i < colNo; i++)
      {
       if (currCh >= charTable.charNo)
	 break;
       if (charTable.font->per_char == NULL)
	 charWidth = charTable.font->max_bounds.width;
       else
	 charWidth = charTable.font->per_char[currCh].width;
       lmargin = (charTable.boxWidth - charWidth) / 2;
       oneChar[0] = currCh + charTable.font->min_char_or_byte2;
       XDrawString(xinfo.dpy, XtWindow(charTable.drawArea), xinfo.gc,
	x + lmargin, y - charTable.font->descent, oneChar, 1);
       x += charTable.boxWidth;
       currCh++;
      }
     y += charTable.boxHeight;
    }

   /* draw mesh */
   XSetForeground (xinfo.dpy, xinfo.gc, GrayBckGround);
   y = rowSt * charTable.boxHeight;
   x = rect.x;
   while (y < (rect.y + rect.height))
    {
     XDrawLine(xinfo.dpy, XtWindow(charTable.drawArea), xinfo.gc, x, y, x + rect.width, y);
     y += charTable.boxHeight;
    }
   y = rect.y;
   x = colSt * charTable.boxWidth;
   while (x < (rect.x + rect.width))
    {
     XDrawLine(xinfo.dpy, XtWindow(charTable.drawArea), xinfo.gc, x, y, x, y + rect.height);
     x += charTable.boxWidth;
    }
  }
 values.font = oldFont;
 values.line_width = oldWidth;
 values.foreground = foreground;
 XChangeGC(xinfo.dpy, xinfo.gc, GCLineWidth | GCForeground, &values);
}

/* refresh the whole char table */
static void refreshAllCharTable(Boolean checkVisibility)
{
 getClipWinDims(charTable.clipWinW, charTable.scrollBX, charTable.scrollBY,
  charTable.DAwidth, charTable.DAheight, &charTable.clipWin);
 if (checkVisibility)
  {
   if (widgetIsViewable(charTable.drawArea))
     exposeCharTbl(charTable.clipWin);
  }
 else
  exposeCharTbl(charTable.clipWin);
}

/* init char table dialog - set size etc. */
static void initCharTable(XFontStruct *font)
{
 unsigned int minCh, maxCh;
 unsigned short i;

 charTable.font = font;
 if (font == NULL)
  {
   charTable.boxWidth = charTable.boxHeight = charTable.charNo = 0;
   charTable.font = NULL;
   return;
  }
 minCh = font->min_char_or_byte2;
 maxCh = font->max_char_or_byte2;
 charTable.charNo = maxCh - minCh + 1;
 charTable.boxWidth = font->max_bounds.width + 7;
 charTable.boxHeight = font->ascent + font->descent + 7;
 charTable.DAwidth = 16 * charTable.boxWidth;
 charTable.DAheight = ((charTable.charNo + 15) / 16) * charTable.boxHeight;
 for (i = 0; i < 256; i++)
  charTable.existing_chars[i] = 0;

 if ((font->all_chars_exist) || (font->per_char == NULL))
  for (i = 0; i < charTable.charNo; i++)
   charTable.existing_chars[i + minCh] = 1;
 else
  for (i = 0; i < charTable.charNo; i++)
   if (font->per_char[i].width)
    charTable.existing_chars[i + minCh] = 1;
 XtVaSetValues(charTable.drawArea, XmNwidth, charTable.DAwidth,
  XmNheight, charTable.DAheight, NULL);
 refreshAllCharTable (True);
}

/* get mouse coordinate and insert char in text field */
static void InsertChar(Widget w, XtPointer clientData, XEvent *event)
{
 unsigned char insCh[2];
 static int x, y;

 if ((CtrlB.sel1 == NULL) || (CtrlB.sel2 == NULL) ||
     (CtrlB.sel3 == NULL) || (charTable.font == NULL))
   return;
 if (event->type == ButtonPress)
  {
   x = event->xbutton.x / charTable.boxWidth;
   x *= charTable.boxWidth;
   y = event->xbutton.y / charTable.boxHeight;
   y *= charTable.boxHeight;
   XSetFunction(xinfo.dpy, xinfo.gc, GXinvert);
   XFillRectangle(xinfo.dpy, XtWindow (charTable.drawArea), xinfo.gc,
    x, y, charTable.boxWidth, charTable.boxHeight);
   XSetFunction(xinfo.dpy, xinfo.gc, GXcopy);
  }
 else
  {
   XmTextPosition currPos;

   insCh[1] = 0;
   insCh[0] = (y / charTable.boxHeight) * 16 + x / charTable.boxWidth;
   insCh[0] += charTable.font->min_char_or_byte2;
   currPos = XmTextGetInsertionPosition (CtrlB.fontNameField);
   if ((insCh[0] == 0) || (charTable.existing_chars[insCh[0]] == 0)) /* some crazy fonts start with 0 - ignore it */
     XBell (xinfo.dpy, 0);
   else
    XmTextInsert(CtrlB.fontNameField, currPos, insCh);
   XSetFunction(xinfo.dpy, xinfo.gc, GXinvert);
   XFillRectangle(xinfo.dpy, XtWindow (charTable.drawArea), xinfo.gc,
    x, y, charTable.boxWidth, charTable.boxHeight);
   XSetFunction(xinfo.dpy, xinfo.gc, GXcopy);
   displayText(&CtrlB);
  }
}

#define FS_INIT_DA 13
#define FS_INIT_SW 14
#define FS_EXPOSE_CHAR_TBL 15
#define FS_INIT_DLG 17

DFONTSEL void inpFontCB(Widget w, XtPointer clientData, XtPointer callData)
{
 int prm = atoi ((char *) clientData);

 switch (prm)
  {
   case FS_INIT_DLG: charTable.dialogW = w; break;
   case FS_INIT_DA:
     charTable.drawArea = w;
     XtAddEventHandler(w, ButtonPressMask | ButtonReleaseMask, FALSE, (XtEventHandler) InsertChar, NULL);
     break;
   case FS_INIT_SW:
     charTable.scrollWin = w;
     XtVaGetValues(w, XmNclipWindow, &charTable.clipWinW,
      XmNhorizontalScrollBar, &charTable.scrollBX,
      XmNverticalScrollBar, &charTable.scrollBY, NULL);
     break;
   case FS_EXPOSE_CHAR_TBL:
     {
      Rectangle rect;
      XmDrawingAreaCallbackStruct *callS;

      callS = (XmDrawingAreaCallbackStruct *) callData;
      if (callS->event)
       {
	rect.x = callS->event->xexpose.x;
	rect.y = callS->event->xexpose.y;
	rect.width = callS->event->xexpose.width;
	rect.height = callS->event->xexpose.height;
	exposeCharTbl(rect);
       }
     }
     break;
   case 1:
     CtrlB.dispLabel = w;
     setLabelString(w, "");
     break;
   case 2: CtrlB.propFontToggle = w; break;
   case 3:
     CtrlB.fontNameField = w;
     XmTextSetString (w, "");
     break;
   case 4:
     CtrlB.sizeFld = w;
     CtrlB.sizeFldInd = initInputText (w, 12, 5, 999);
     break;
   case 5: CtrlB.fontList = w; break;
   case 6: CtrlB.styleList = w; break;
   case 7: CtrlB.sizeList = w; break;
   case 8: CtrlB.okButton = w; break; /* actually w == form */
   case 10: CtrlB.dimensionLbl = w; setLabelString(w, ""); break;
   case 11:
     if ((CtrlB.sel1 != NULL) && (CtrlB.sel2 != NULL) && (CtrlB.sel3 != NULL))
      choiceMade(&CtrlB);
     break;
   case 12:
     CtrlB.aspectRatioTB = w;
     break;
   case 9:
    {
     int i = 0, j = 0, numChildren;
     WidgetList wList;
     Widget OKCancelHlp[3];

     CtrlB.form = w;
     XtVaGetValues(CtrlB.okButton, XmNnumChildren, &numChildren, XmNchildren, &wList, NULL);
     while (i < numChildren)
      {
       i++;
       if (XmIsPushButton (*wList))
	OKCancelHlp[j++] = *wList;
       wList++;
      }
     CtrlB.okButton = OKCancelHlp[0];
     CtrlB.cancelButton = OKCancelHlp[1];

     XtVaSetValues(w, XmNcancelButton, CtrlB.cancelButton, NULL);
     XtAddCallback(CtrlB.propFontToggle, XmNvalueChangedCallback, (XtCallbackProc) propFontToggleAction, (char *) &CtrlB);
     XtAddCallback(CtrlB.fontList, XmNbrowseSelectionCallback, (XtCallbackProc) fontAction, (char *) &CtrlB);
     XtAddCallback(CtrlB.styleList, XmNbrowseSelectionCallback, (XtCallbackProc) styleAction, (char *) &CtrlB);
     XtAddCallback(CtrlB.sizeList, XmNbrowseSelectionCallback, (XtCallbackProc) sizeAction, (char *) &CtrlB);
     XtAddCallback(CtrlB.okButton, XmNactivateCallback, (XtCallbackProc) okAction, (char *) &CtrlB);
     XtAddCallback(CtrlB.cancelButton, XmNactivateCallback, (XtCallbackProc) cancelAction, (char *) &CtrlB);
     XtAddCallback(CtrlB.aspectRatioTB, XmNvalueChangedCallback, (XtCallbackProc) aspectRatioToggleAction, (char *) &CtrlB);
     XmProcessTraversal(CtrlB.fontList, XmTRAVERSE_CURRENT);
     /* Link Motif Close option to cancel action */
     AddMotifCloseCallback(XtParent (CtrlB.form), (XtCallbackProc) cancelAction, &CtrlB);
    }
   break;
  }
}

DFONTSEL XFontStruct *FontSel(Widget parent, int showPropFonts, char *currFont)
{
 char bigFont[MAX_FONT_NAME_LEN];

 CtrlB.fontData = XListFonts(XtDisplay (parent),
  "-*-*-*-*-*-*-*-*-*-*-*-*-*-*", MAX_NUM_FONTS, &CtrlB.numFonts);
 FindBigFont(&CtrlB, bigFont);
 CtrlB.oldFont = XLoadQueryFont(XtDisplay(parent), bigFont);
 CtrlB.oldFontList = XmFontListCreate(CtrlB.oldFont, XmSTRING_DEFAULT_CHARSET);
 CtrlB.exitFlag = False;
 /* update application's control block structure */
 CtrlB.showSizeInPixels = TRUE;
 CtrlB.sel1 = NULL;
 CtrlB.sel2 = NULL;
 CtrlB.sel3 = NULL;
 CtrlB.fontName = NULL;
 CtrlB.retFont = NULL;
 CtrlB.textWidth = CtrlB.textHeight = 0;
 if (XmToggleButtonGetState(CtrlB.propFontToggle))
  CtrlB.showPropFonts = PREF_PROP;
 else
  CtrlB.showPropFonts = PREF_FIXED;
 CtrlB.keepAspectRatio = XmToggleButtonGetState (CtrlB.aspectRatioTB);
 CtrlB.maxAspRatioDiff = 0.1;
 setupScrollLists (NONE, CtrlB); /* update scroll lists */
 if (showPropFonts == PREF_PROP)
   XmToggleButtonSetState (CtrlB.propFontToggle, TRUE, FALSE);
 XtManageChild (CtrlB.form);
 /* set up current font parameters */
 if (currFont[0] != '\0')
   startupFont (&CtrlB, currFont);
 XtManageChild (charTable.dialogW);
 /* enter event loop */
 while (!CtrlB.exitFlag)
   XtAppProcessEvent (XtWidgetToApplicationContext (CtrlB.form), XtIMAll);
 XtUnmanageChild(CtrlB.form);
 if (CtrlB.oldFont != NULL)
  {
   XFreeFont (XtDisplay (CtrlB.form), CtrlB.oldFont);
   XmFontListFree(CtrlB.oldFontList);
  }
 return(CtrlB.retFont);
}

/* gets a specific substring from a string */
static void getStringComponent(char *inStr, int pos, char *outStr)
{
 int i, j;

 *outStr = '\0';
 if (pos > NUM_COMPONENTS_FONT_NAME)
  return;
 for (i = 0; (pos > 0) && (inStr[i] != '\0'); i++)
   if (inStr[i] == DELIM)
     pos--;
 if (inStr[i] == '\0')
   return;
 for (j = 0; (inStr[i] != DELIM) && (inStr[i] != '\0'); i++, j++)
   outStr[j] = inStr[i];
 outStr[j] = '\0';
}

/* parse through the fontlist data and set up the three scroll lists */
static void setupScrollLists(int dontChange, xfselControlBlkType CtrlBlk)
{
 char *itemBuf1[MAX_ENTRIES_IN_LIST];
 char *itemBuf2[MAX_ENTRIES_IN_LIST];
 char *itemBuf3[MAX_ENTRIES_IN_LIST];
 int itemCount1=0, itemCount2=0, itemCount3=0;
 char buff1[TEMP_BUF_SIZE];
 XmString items[MAX_ENTRIES_IN_LIST];
 int i;
 Boolean scalelableSizes = False;

 for (i = 0; i < CtrlBlk.numFonts; i++)
  {
   if ((dontChange != FONT) &&
       (styleMatch (CtrlBlk, CtrlBlk.fontData[i])) &&
       (sizeMatch (CtrlBlk, CtrlBlk.fontData[i])) &&
       (aspectRatioMatch (CtrlBlk, CtrlBlk.fontData[i])) &&
       ((CtrlBlk.showPropFonts == PREF_PROP) ||
	(notPropFont (CtrlBlk.fontData[i]))))
     {
      getFontPart (CtrlBlk.fontData[i], buff1);
      addItemToList (itemBuf1, buff1, &itemCount1, True);
     }

   if ((dontChange != STYLE) &&
       (fontMatch (CtrlBlk, CtrlBlk.fontData[i])) &&
       (sizeMatch (CtrlBlk, CtrlBlk.fontData[i])) &&
       (aspectRatioMatch (CtrlBlk, CtrlBlk.fontData[i])) &&
       ((CtrlBlk.showPropFonts == PREF_PROP) ||
	(notPropFont (CtrlBlk.fontData[i]))))
     {
      getStylePart (CtrlBlk.fontData[i], buff1);
      addItemToList (itemBuf2, buff1, &itemCount2, True);
     }

   if ((dontChange != SIZE) &&
       (fontMatch (CtrlBlk, CtrlBlk.fontData[i])) &&
       (styleMatch (CtrlBlk, CtrlBlk.fontData[i])) &&
       (aspectRatioMatch (CtrlBlk, CtrlBlk.fontData[i])) &&
       ((CtrlBlk.showPropFonts == PREF_PROP) ||
	(notPropFont (CtrlBlk.fontData[i]))))
     {
      getSizePart(CtrlBlk.fontData[i], buff1, CtrlBlk.showSizeInPixels);
      if (atoi (buff1) == 0) /* scalable font */
       {
	int res;

	/* check dpi - if 0 then is really scalable */
	getStringComponent (CtrlBlk.fontData[i], 7, buff1);
	res = atoi (buff1);
	getStringComponent (CtrlBlk.fontData[i], 8, buff1);
	res += atoi (buff1);
	if (res == 0)
	  scalelableSizes = True;
       }
      else
       addItemToList(itemBuf3, buff1, &itemCount3, False);
     }
  }

 /* if scalelable fonts present create a list of size items */
 if (scalelableSizes)
  {
   int j;

   for (j = 5; j < 200; j++)
    {
     sprintf (buff1, "%d", j);
     addItemToList(itemBuf3, buff1, &itemCount3, False);
    }
  }
 /* recreate all three scroll lists where necessary */
 if (dontChange != FONT)
  {
   for (i = 0; i < itemCount1; i++)
    {
     items[i] = XmStringCreate(itemBuf1[i], XmSTRING_DEFAULT_CHARSET);
     free (itemBuf1[i]);
    }
   XmListDeleteAllItems (CtrlBlk.fontList);
   XmListAddItems (CtrlBlk.fontList, items, itemCount1, 1);
   if (CtrlBlk.sel1 != NULL)
    {
     freeXmString (&items[0]);
     items[0] = XmStringCreate(CtrlBlk.sel1, XmSTRING_DEFAULT_CHARSET);
     XmListSelectItem (CtrlBlk.fontList, items[0], FALSE);
     XmListSetBottomItem(CtrlBlk.fontList, items[0]);
    }
   for (i = 0; i < itemCount1; i++)
    freeXmString(&items[i]);
  }

 if (dontChange != STYLE)
  {
   for (i = 0; i < itemCount2; i++)
    {
     items[i] = XmStringCreate(itemBuf2[i], XmSTRING_DEFAULT_CHARSET);
     free(itemBuf2[i]);
    }
   XmListDeleteAllItems (CtrlBlk.styleList);
   XmListAddItems (CtrlBlk.styleList, items, itemCount2, 1);
   if (CtrlBlk.sel2 != NULL)
    {
     freeXmString (&items[0]);
     items[0] = XmStringCreate(CtrlBlk.sel2, XmSTRING_DEFAULT_CHARSET);
     XmListSelectItem (CtrlBlk.styleList, items[0], FALSE);
     XmListSetBottomItem (CtrlBlk.styleList, items[0]);
    }
   for (i = 0; i < itemCount2; i++)
    freeXmString(&items[i]);
  }

 if ((dontChange != SIZE))
  {
   for (i = 0; i < itemCount3; i++)
    {
     items[i] = XmStringCreate(itemBuf3[i], XmSTRING_DEFAULT_CHARSET);
     free(itemBuf3[i]);
    }
   XmListDeleteAllItems (CtrlBlk.sizeList);
   XmListAddItems (CtrlBlk.sizeList, items, itemCount3, 1);
   if (CtrlBlk.sel3 != NULL)
    {
     freeXmString (&items[0]);
     items[0] = XmStringCreate(CtrlBlk.sel3, XmSTRING_DEFAULT_CHARSET);
     XmListSelectItem (CtrlBlk.sizeList, items[0], FALSE);
     XmListSetBottomItem (CtrlBlk.sizeList, items[0]);
    }
   for (i = 0; i < itemCount3; i++)
    freeXmString(&items[i]);
  }
}

/* returns TRUE if argument is not name of a proportional font */
static int notPropFont(char *font)
{
 char buff1[TEMP_BUF_SIZE];

 getStringComponent (font, 11, buff1);
 if ((strcmp (buff1, "p") == 0) || (strcmp (buff1, "P") == 0))
  return(FALSE);
 else
  return(TRUE);
}

/* returns TRUE if the style portion of the font matches the currently selected style */
static int styleMatch(xfselControlBlkType CtrlBlk, char *font)
{
 char buff[TEMP_BUF_SIZE];

 if (CtrlBlk.sel2 == NULL)
  return(TRUE);
 getStylePart (font, buff);
 if (strcmp (buff, CtrlBlk.sel2) == 0)
  return(TRUE);
 else
  return(FALSE);
}

/* returns TRUE if the size portion of the font matches the currently selected size */
static int sizeMatch(xfselControlBlkType CtrlBlk, char *font)
{
 char buff[TEMP_BUF_SIZE];
 int result;

 result = True;
 if (CtrlBlk.sel3 == NULL)
  return(TRUE);
 getSizePart (font, buff, CtrlBlk.showSizeInPixels);
 if ((atoi (buff) == 0))
   return (True);
 /* scalable font */
 if (strcmp (buff, CtrlBlk.sel3) == 0)
   result = True;
 else
   result = False;
 return(result);
}

/* check if xdpy in ydpy match the requested aspect ratio */
/* valid only when keepAspectRatio is True */
static int  aspectRatioMatch(xfselControlBlkType CtrlBlk, char *font)
{
 char buff[TEMP_BUF_SIZE];
 int xdpy, ydpy;
 float aspectRdpy;
 int result;

 if (!CtrlBlk.keepAspectRatio)
   return (True);

 result = True;
 getStringComponent (font, 9, buff);
 aspectRdpy = xdpy = atoi (buff);
 getStringComponent (font, 10, buff);
 ydpy = atoi (buff);
 if (ydpy)
  aspectRdpy /= ydpy;
 else
  aspectRdpy = 1.0;
 if ((xdpy + ydpy) == 0)
   result = True;
 else if (fabs(densityRatio() - aspectRdpy) > CtrlBlk.maxAspRatioDiff)
   result = False;
 return(result);
}

/* returns TRUE if the font portion of the font matches the currently selected font */
static int fontMatch(xfselControlBlkType CtrlBlk, char *font)
{
 char buff[TEMP_BUF_SIZE];

 if (CtrlBlk.sel1 == NULL)
  return (TRUE);
 getFontPart (font, buff);
 if (strcmp (buff, CtrlBlk.sel1) == 0)
  return (TRUE);
 else
  return (FALSE);
}

/* inserts a string into correct sorted position in a list */
static void addItemToList(char **buf, char *item, int *count, int lexOrder)
{
 int i, j;

 if (*count == MAX_ENTRIES_IN_LIST)
  return;

 if (lexOrder)
  for (i = 0; i < *count; i++)
   {
    if (strcmp(buf[i], item) == 0)
     return;
    if (strcmp(buf[i], item) > 0)
     break;
   }
 else
  for (i = 0; i < *count; i++)
   {
    if (atoi(buf[i]) == atoi(item))
     return;
    if (atoi(buf[i]) > atoi(item))
     break;
   }

 for (j = *count; j > i; j--)
   buf[j] = buf[j - 1];
 buf[i] = malloc(strlen(item) + 1);
 strcpy (buf[i], item);
 (*count)++;
}

/* given a font name this function returns the part used in the first scroll list */
static void getFontPart(char *font, char *buff1)
{
 char buff2[TEMP_BUF_SIZE], buff3[TEMP_BUF_SIZE];
 char buff4[TEMP_BUF_SIZE];

 getStringComponent (font, 2, buff1);
 getStringComponent (font, 1, buff2);
 sprintf (buff3, "%s (%s", buff1, buff2);
 getStringComponent (font, 13, buff1);
 getStringComponent (font, 14, buff4);
 if (((strncmp (buff1, "iso8859", 7) == 0) ||
      (strncmp (buff1, "ISO8859", 7) == 0)) && (strcmp (buff4, "1") == 0))
   sprintf (buff1, "%s)", buff3);
 else
  {
   sprintf (buff2, "%s, %s,", buff3, buff1);
   sprintf (buff1, "%s %s)", buff2, buff4);
  }
}

/* given a font name this function returns the part used in the second scroll list */
static void getStylePart(char *font, char *buff1)
{
 char buff2[TEMP_BUF_SIZE], buff3[TEMP_BUF_SIZE];

 getStringComponent (font, 3, buff3);
 getStringComponent (font, 5, buff2);
 if ((strcmp(buff2, "normal") != 0) && (strcmp(buff2, "Normal") != 0) &&
     (strcmp(buff2, "NORMAL") != 0))
  sprintf (buff1, "%s %s", buff3, buff2);
 else
  strcpy (buff1, buff3);
 getStringComponent (font, 6, buff2);
 if (buff2[0] != '\0')
   sprintf (buff3, "%s %s", buff1, buff2);
 else
   strcpy (buff3, buff1);
 getStringComponent (font, 4, buff2);
 if ((strcmp (buff2, "o") == 0) || (strcmp (buff2, "O") == 0))
   sprintf (buff1, "%s oblique", buff3);
 else if ((strcmp (buff2, "i") == 0) || (strcmp (buff2, "I") == 0))
   sprintf (buff1, "%s italic", buff3);
 if (strcmp (buff1, " ") == 0)
   strcpy (buff1, "-");
}

/* given a font name this function returns the part used in the third scroll list */
static void getSizePart(char *font, char *buff1, int inPixels)
{
 int size;
 float temp;

 if (inPixels)
  {
   getStringComponent (font, 7, buff1);
   size = atoi (buff1);
   sprintf (buff1, "%d", size);
  }
 else
  {
   getStringComponent (font, 8, buff1);
   size = atoi (buff1);
   temp = (float) size / 10.0;
   size = temp;
   if (buff1[strlen (buff1) - 1] == '0')
    sprintf(buff1, "%d", size);
   else
    sprintf(buff1, "%4.1f", temp);
  }
}

/* Callback functions start from here - suffix Action in the function name
   is for the callback function for the corresponding widget */
static void propFontToggleAction(Widget widget, xfselControlBlkType *CtrlBlk,
 XmToggleButtonCallbackStruct *call_data)
{
 if (call_data->reason == XmCR_VALUE_CHANGED)
  {
   if (CtrlBlk->showPropFonts == PREF_FIXED)
    CtrlBlk->showPropFonts = PREF_PROP;
   else
    CtrlBlk->showPropFonts = PREF_FIXED;
   if (CtrlBlk->sel1 != NULL)
     free (CtrlBlk->sel1);
   CtrlBlk->sel1 = NULL;
   if (CtrlBlk->sel2 != NULL)
     free (CtrlBlk->sel2);
   CtrlBlk->sel2 = NULL;
   if (CtrlBlk->sel3 != NULL)
     free (CtrlBlk->sel3);
   CtrlBlk->sel3 = NULL;
   setupScrollLists (NONE, *CtrlBlk);
   setLabelString(CtrlBlk->dispLabel, "");
   setLabelString(CtrlBlk->dimensionLbl, "");
   refreshAllCharTable(False);
  }
}

/* basically this functions sets the global variable keepAspectRatio and
   redisplays the scrolled list according to the new value */
static void aspectRatioToggleAction(Widget widget,  xfselControlBlkType *CtrlBlk,
 XmToggleButtonCallbackStruct *call_data)
{
 if (call_data->reason == XmCR_VALUE_CHANGED)
  {
   CtrlBlk->keepAspectRatio = call_data->set;
   if (CtrlBlk->sel1 != NULL)
     free (CtrlBlk->sel1);
   CtrlBlk->sel1 = NULL;
   if (CtrlBlk->sel2 != NULL)
     free (CtrlBlk->sel2);
   CtrlBlk->sel2 = NULL;
   if (CtrlBlk->sel3 != NULL)
     free (CtrlBlk->sel3);
   CtrlBlk->sel3 = NULL;
   setupScrollLists (NONE, *CtrlBlk);
   setLabelString(CtrlBlk->dispLabel, "");
   setLabelString(CtrlBlk->dimensionLbl, "");
   refreshAllCharTable (False);
  }
}

static void fontAction(Widget widget, xfselControlBlkType *CtrlBlk, XmListCallbackStruct *call_data)
{
 char *sel;

 XmStringGetLtoR (call_data->item, XmSTRING_DEFAULT_CHARSET, &sel);

 if (CtrlBlk->sel1 == NULL)
  {
   CtrlBlk->sel1 = malloc (strlen (sel) + 1);
   strcpy(CtrlBlk->sel1, sel);
  }
 else
  {
   if (strcmp (CtrlBlk->sel1, sel) == 0)
    { /* Unselecting current selection */
     free (CtrlBlk->sel1);
     CtrlBlk->sel1 = NULL;
     XmListDeselectItem (widget, call_data->item);
    }
   else
    {
     free (CtrlBlk->sel1);
     CtrlBlk->sel1 = malloc (strlen (sel) + 1);
     strcpy (CtrlBlk->sel1, sel);
    }
  }

 free (sel);
 setupScrollLists (FONT, *CtrlBlk);
 if ((CtrlBlk->sel1 != NULL) && (CtrlBlk->sel2 != NULL) && (CtrlBlk->sel3 != NULL))
  choiceMade(CtrlBlk);
 else
  {
   setLabelString(CtrlBlk->dispLabel, "");
   setLabelString(CtrlBlk->dimensionLbl, "");
   refreshAllCharTable(FALSE);
  }
}

static void styleAction(Widget widget, xfselControlBlkType *CtrlBlk, XmListCallbackStruct *call_data)
{
 char *sel;

 XmStringGetLtoR(call_data->item, XmSTRING_DEFAULT_CHARSET, &sel);

 if (CtrlBlk->sel2 == NULL)
  {
   CtrlBlk->sel2 = malloc(strlen(sel) + 1);
   strcpy(CtrlBlk->sel2, sel);
  }
 else
   {
    if (strcmp (CtrlBlk->sel2, sel) == 0)
     { /* unselecting current selection */
      free (CtrlBlk->sel2);
      CtrlBlk->sel2 = NULL;
      XmListDeselectItem (widget, call_data->item);
     }
    else
     {
      free(CtrlBlk->sel2);
      CtrlBlk->sel2 = malloc (strlen (sel) + 1);
      strcpy(CtrlBlk->sel2, sel);
     }
   }
 free(sel);
 setupScrollLists(STYLE, *CtrlBlk);
 if ((CtrlBlk->sel1 != NULL) && (CtrlBlk->sel2 != NULL) && (CtrlBlk->sel3 != NULL))
  choiceMade(CtrlBlk);
 else
  {
   setLabelString(CtrlBlk->dispLabel, "");
   setLabelString(CtrlBlk->dimensionLbl, "");
   refreshAllCharTable(False);
  }
}

static void sizeAction(Widget widget, xfselControlBlkType *CtrlBlk,
 XmListCallbackStruct *call_data)
{
 char *sel;

 XmStringGetLtoR (call_data->item, XmSTRING_DEFAULT_CHARSET, &sel);

 if (CtrlBlk->sel3 == NULL)
  {
   CtrlBlk->sel3 = malloc (strlen (sel) + 1);
   strcpy (CtrlBlk->sel3, sel);
  }
 else
  {
   if (strcmp (CtrlBlk->sel3, sel) == 0)
    { /* unselecting current selection */
     free (CtrlBlk->sel3);
     CtrlBlk->sel3 = NULL;
     XmListDeselectItem (widget, call_data->item);
    }
   else
    {
     free (CtrlBlk->sel3);
     CtrlBlk->sel3 = malloc (strlen (sel) + 1);
     strcpy (CtrlBlk->sel3, sel);
    }
  }
 free (sel);
 setupScrollLists (SIZE, *CtrlBlk);
 if ((CtrlBlk->sel1 != NULL) && (CtrlBlk->sel2 != NULL) && (CtrlBlk->sel3 != NULL))
   choiceMade(CtrlBlk);
 else
  {
   setLabelString(CtrlBlk->dispLabel, "");
   setLabelString(CtrlBlk->dimensionLbl, "");
   refreshAllCharTable (False);
  }
}

/* function called when all three choices have been made; sets up font
   name and displays sample font */
static void choiceMade(xfselControlBlkType *CtrlBlk)
{
 int i;

 if (CtrlBlk->fontName != NULL)
   free(CtrlBlk->fontName);
 CtrlBlk->fontName = NULL;

 for (i = 0; i < CtrlBlk->numFonts; i++)
  {
   if ((fontMatch(*CtrlBlk, CtrlBlk->fontData[i])) &&
    (styleMatch(*CtrlBlk, CtrlBlk->fontData[i])) &&
    (sizeMatch(*CtrlBlk, CtrlBlk->fontData[i])) &&
    (aspectRatioMatch(*CtrlBlk, CtrlBlk->fontData[i])))
    {
     CtrlBlk->fontName = malloc(strlen(CtrlBlk->fontData[i]) + 1);
     strcpy(CtrlBlk->fontName, CtrlBlk->fontData[i]);
     break;
    }
  }
 if (CtrlBlk->fontName)
  dispSample(CtrlBlk, True);
 else
  manageErrorDialog(getTxtI8N(INVALID_FONT_MSG), NULL, NULL, getTxtI8N(OK_MSG));
}

static void measureStringExtensions(XFontStruct *font, unsigned char *txt,
 int *width, int *ascent, int *descent, int *l_offset)
{
 int l;

 *width = *ascent = *descent = 0;
 if ((!txt) || (!font))
   return;
 l = strlen (txt);

 if (font->per_char == NULL)
  {
   *width = XTextWidth (font, txt, l);
   *ascent = font->ascent;
   *descent = font->descent;
  }
 else
  {
   int i, totalW, maxDesc, maxAsc, minByte, maxByte, ind, tmp;

   minByte = font->min_char_or_byte2;
   maxByte = font->max_char_or_byte2;
   totalW = 0;
   maxDesc = maxAsc = 0;
   for (i = 0; i < l; i++)
    {
     ind = txt[i] - minByte;
     if ((ind < 0) || (txt[i] > maxByte))
      {
       totalW += font->max_bounds.width;
       tmp = font->max_bounds.ascent;
       if (tmp > maxAsc)
	 maxAsc = tmp;
       tmp = font->max_bounds.descent;
       if (tmp > maxDesc)
	 maxDesc = tmp;
       continue;
      }
     totalW += font->per_char[ind].width;
     if (i == 0)
      {
       *l_offset = font->per_char[ind].lbearing;
       totalW -= *l_offset;
      }
     if (i == (l - 1))
      totalW -= font->per_char[ind].width - font->per_char[ind].rbearing;

     tmp = font->per_char[ind].ascent;
     if (tmp > maxAsc)
       maxAsc = tmp;
     tmp = font->per_char[ind].descent;
     if (tmp > maxDesc)
      maxDesc = tmp;
    }
   *width = totalW;
   *ascent = maxAsc;
   *descent = maxDesc;
  }
}

/* display text line using current font */
static void displayText(xfselControlBlkType *CtrlBlk)
{
 unsigned char *line, buff[100];
 int f_asc, f_desc, ww, l_offs;

 line = XmTextGetString(CtrlBlk->fontNameField);
 checkText(CtrlBlk->oldFont, line);
 setLabelString(CtrlBlk->dispLabel, line);
 /* get dimensions of the string */
 measureStringExtensions(CtrlBlk->oldFont, line, &ww, &f_asc, &f_desc, &l_offs);
 CtrlBlk->textWidth = ww;
 CtrlBlk->textHeight = f_asc + f_desc;
 sprintf(buff, "%s %dx%d (%.2fx%.2f %s)", getTxtI8N (SIZE_MSG),
  CtrlBlk->textWidth, CtrlBlk->textHeight,
  CtrlBlk->textWidth/(1.0*ApData.width/getImageXmm()),  /* widthMM */
  CtrlBlk->textHeight/(1.0*ApData.height/getImageYmm()),/* heightMM */
  metricSystem ? "mm" : "in");
 setLabelString(CtrlBlk->dimensionLbl, buff);
 free(line);
}

/* loads selected font and displays sample text in that font */
static void dispSample(xfselControlBlkType * CtrlBlk, Boolean initTable)
{
 XFontStruct *font;
 XmFontList fontList;
 Display *display;
 char fontName[MAX_FONT_NAME_LEN];

 display = XtDisplay (CtrlBlk->form);
 /* check if scalable font - then modify the fontName accordingly */
 getSizePart (CtrlBlk->fontName, &fontName[0], CtrlBlk->showSizeInPixels);
 if (atoi (fontName) == 0)
  {
   int i, pos = 7;

   for (i = 0; (pos > 0) && (CtrlBlk->fontName[i] != '\0'); i++)
     if (CtrlBlk->fontName[i] == DELIM)
       pos--;
   if (CtrlBlk->fontName[i] == '\0')
     return;
   strncpy (fontName, CtrlBlk->fontName, i);
   fontName[i] = '\0';
   strcat (fontName, CtrlBlk->sel3);
   strcat (fontName, "-0-0-0-");
   pos = 11;
   for (i = 0; (pos > 0) && (CtrlBlk->fontName[i] != '\0'); i++)
     if (CtrlBlk->fontName[i] == DELIM)
       pos--;
   if (CtrlBlk->fontName[i] == '\0')
     return;
   strcat (fontName, &(CtrlBlk->fontName[i]));
   free (CtrlBlk->fontName);
   CtrlBlk->fontName = malloc (strlen (fontName) + 1);
   strcpy (CtrlBlk->fontName, &fontName[0]);
  }
 if (CtrlBlk->keepAspectRatio)
  {
   int i, pos = 9, dpyX, dpyY;
   float ratio, diffRatio;

   getStringComponent (CtrlBlk->fontName, 9, &fontName[0]);
   dpyX = atoi (&fontName[0]);
   getStringComponent (CtrlBlk->fontName, 10, &fontName[0]);
   dpyY = atoi (&fontName[0]);
   if (dpyY == 0)
     ratio = 1.0;
   else
     ratio = 1.0 * dpyX / dpyY;

   diffRatio = fabs(densityRatio() - ratio);
   if (diffRatio > CtrlBlk->maxAspRatioDiff)
    {
     unsigned char buff[100];

     for (i = 0; (pos > 0) && (CtrlBlk->fontName[i] != '\0'); i++)
       if (CtrlBlk->fontName[i] == DELIM)
	 pos--;

     if (CtrlBlk->fontName[i] == '\0')
       return;
     strncpy (fontName, CtrlBlk->fontName, i);
     fontName[i] = '\0';
     sprintf (buff, "%d-%d-", (int)ApData.imageXdpi, (int)ApData.imageYdpi);
     strcat (fontName, buff);
     pos = 11;
     for (i = 0; (pos > 0) && (CtrlBlk->fontName[i] != '\0'); i++)
       if (CtrlBlk->fontName[i] == DELIM)
	 pos--;
     if (CtrlBlk->fontName[i] == '\0')
       return;
     strcat (fontName, &(CtrlBlk->fontName[i]));
     free (CtrlBlk->fontName);
     CtrlBlk->fontName = malloc (strlen (fontName) + 1);
     strcpy (CtrlBlk->fontName, &fontName[0]);
    }
  }
 font = XLoadQueryFont (display, CtrlBlk->fontName);
 fontList = XmFontListCreate (font, XmSTRING_DEFAULT_CHARSET);
 XtVaSetValues(CtrlBlk->dispLabel, XmNfontList, fontList, NULL);
 initCharTable(font);
 displayText(CtrlBlk);
 if (CtrlBlk->oldFont != NULL)
  {
   XFreeFont(display, CtrlBlk->oldFont);
   XmFontListFree(CtrlBlk->oldFontList);
  }
 CtrlBlk->oldFont = font;
 CtrlBlk->oldFontList = fontList;
 displayText(&CtrlB);
}

static void cancelAction(Widget widget, xfselControlBlkType * CtrlBlk,
 XmListCallbackStruct * call_data)
{
 if (CtrlBlk->sel1 != NULL)
   free (CtrlBlk->sel1);
 if (CtrlBlk->sel2 != NULL)
   free (CtrlBlk->sel2);
 if (CtrlBlk->sel3 != NULL)
   free (CtrlBlk->sel3);
 if (CtrlBlk->fontName != NULL)
   free (CtrlBlk->fontName);
 CtrlBlk->fontName = NULL;
 XFreeFontNames(CtrlBlk->fontData);
 CtrlBlk->exitFlag = TRUE;
}

static void okAction(Widget widget, xfselControlBlkType *CtrlBlk,
 XmPushButtonCallbackStruct *call_data)
{
 char **fontName;
 int i;

 fontName = XListFonts (XtDisplay (CtrlBlk->form), CtrlBlk->fontName, 1, &i);
 if (i != 1)
  {
   manageErrorDialog(getTxtI8N(INVALID_FONT_MSG), NULL, NULL, getTxtI8N(OK_MSG));
   XFreeFontNames(fontName);
  }
 else
  {
   char *line;
   XPoint *np = &Undo.points[0];
   int ww, f_asc, f_desc, l_offs;

   if (CtrlBlk->fontName != NULL)
     free (CtrlBlk->fontName);
   CtrlBlk->fontName = malloc (strlen (fontName[0]) + 1);
   strcpy (CtrlBlk->fontName, fontName[0]);

   if (CtrlBlk->sel1 != NULL)
     free (CtrlBlk->sel1);
   if (CtrlBlk->sel2 != NULL)
     free (CtrlBlk->sel2);
   if (CtrlBlk->sel3 != NULL)
     free (CtrlBlk->sel3);
   XFreeFontNames (fontName);
   XFreeFontNames (CtrlBlk->fontData);
   line = XmTextGetString (CtrlBlk->fontNameField);
   checkText (CtrlBlk->oldFont, line);
   measureStringExtensions (CtrlBlk->oldFont, line, &ww, &f_asc, &f_desc, &l_offs);
   np->x = -l_offs; /* coordinates */
   np->y = 0;
   np++;
   np->x = (short) ww;
   np->y = (short) (f_asc + f_desc);
   np++;
   np->x = (short) strlen (line);
   np->y = (short) f_asc;
   np++;
   for (i = 0; i < (int) strlen (line); i++)
    { np->x = (short) line[i]; np++; }
   Undo.npoints = 3 + strlen (line);
   free (line);
   CtrlBlk->retFont = CtrlBlk->oldFont;
   CtrlBlk->oldFont = NULL;
   CtrlBlk->exitFlag = TRUE;
  }
}

/* if current font is passed as an argument then this function is
   invoked and sets up initial entries */
static void startupFont(xfselControlBlkType * CtrlBlk, char *font)
{
 int i;
 char **fontName;
 char part[TEMP_BUF_SIZE];
 XmString str;

 fontName = XListFonts(XtDisplay(CtrlBlk->form), font, 1, &i);

 if (i == 0)
  { /* invalid font passed in at startup */
   XFreeFontNames(fontName);
   return;
  }

 CtrlBlk->fontName = malloc (strlen (fontName[0]) + 1);
 strcpy (CtrlBlk->fontName, fontName[0]);

 getFontPart (fontName[0], part);
 XFreeFontNames (fontName);
 str = XmStringCreate (part, XmSTRING_DEFAULT_CHARSET);
 XmListSetBottomItem (CtrlBlk->fontList, str);
 XmListSelectItem (CtrlBlk->fontList, str, TRUE);
 freeXmString (&str);

 dispSample (CtrlBlk, True);
}

/* finds the name of the biggest font less than the given limit
   MAX_DISPLAY_SIZE used to set up the initial height of the display widget */
static void FindBigFont(xfselControlBlkType * CtrlBlk, char *bigFont)
{
 int i, maxSize, size, index = 0;
 char sizeStr[10];

 for (i = 0, maxSize = 0; i < CtrlBlk->numFonts; i++)
  {
   getStringComponent (CtrlBlk->fontData[i], 7, sizeStr);
   size = atoi (sizeStr);
   if ((size > maxSize) && (size < MAX_DISPLAY_SIZE))
    { index = i; maxSize = size; }
  }
 strcpy(bigFont, CtrlBlk->fontData[index]);
}

DFONTSEL void moveString(Widget w, XtPointer prm, XEvent * event)
{
 int zoomXF, zoomYF;
 static int moveIt = 0;

 if (prm)
  { zoomXF = zoomDraw.factorX; zoomYF = zoomDraw.factorY; }
 else
  zoomXF = zoomYF = 1;

 if (event->type == ButtonRelease)
  moveIt = 0;
 else if ((event->type == ButtonPress) && (Undo.points[2].x))
  {
   if (((event->xbutton.x / zoomXF) >= Undo.points[0].x) &&
       ((event->xbutton.x / zoomXF) <= (Undo.points[0].x + Undo.points[1].x)) &&
       ((event->xbutton.y / zoomYF) >= Undo.points[0].y) &&
   ((event->xbutton.y / zoomYF) <= (Undo.points[0].y + Undo.points[1].y)))
    {
     ApData.orig_xy.x = event->xbutton.x / zoomXF;
     ApData.orig_xy.y = event->xbutton.y / zoomYF;
     Select.rect_on = False;
     moveIt = 1;
     opBounds.width = Undo.points[1].x;
     opBounds.height = Undo.points[1].y;
    }
  }
 else if (moveIt)
  {
   MyPutImage(drawAreaWin, xinfo.gc, ApData.img,
    Undo.points[0].x, Undo.points[0].y, Undo.points[0].x, Undo.points[0].y,
    Undo.points[1].x, Undo.points[1].y, NULL);
   refreshZoomWin(opBounds.x, opBounds.y, opBounds.width, opBounds.height, True);
   Undo.points[0].x += event->xbutton.x / zoomXF - ApData.orig_xy.x;
   Undo.points[0].y += event->xbutton.y / zoomYF - ApData.orig_xy.y;
   opBounds.x = Undo.points[0].x;
   opBounds.y = Undo.points[0].y;
   LastDrawOnAllWin (False);
   ApData.orig_xy.x = event->xbutton.x / zoomXF;
   ApData.orig_xy.y = event->xbutton.y / zoomYF;
  }
}
