#define HELP
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <Xm/Xm.h>
#include <Xm/RowColumn.h>
#include <Xm/MessageB.h>
#include <Xm/Label.h>
#include <Xm/LabelG.h>
#include <Xm/BulletinB.h>
#include <Xm/Scale.h>
#include <Xm/Text.h>
#include "paint.h"

#define TxtSizeLimit 5000
#define BufLimit     200

static char *keyWords[] = {"topic", "subtopic", "reference"};
static char delimSt[] = "<<";
static char delimEnd[] = ">>";
static char delimKeyWst[] = "<#";
static char delimKeyWend[] = "#>";

typedef enum { nothing, topic, subtopic, reference } keyWordType;

static struct
{
 FILE *f;		/* help file pointer */
 char fName[BufLimit];	/* help file name */
 long pos;		/* help file position */
 char buf[BufLimit];
 Boolean enabled;
} Help;

static struct
{
 keyWordType type;
 char *msg;
 int count;
} HelpItem;

#define MSG_NUM 1000
static char *msgArr[MSG_NUM];

DHELP Boolean initHelp(char *helpFName)
{
 Help.f = fopen (helpFName, "r");
 Help.enabled = (Help.f != NULL);
 if (!Help.f) /* error opening file */
  return (False);
 else
  return (True);
}

DHELP void endHelp()
{
 fclose(Help.f);
}

static Boolean getKeyValue(char **line, keyWordType * key, char *value)
{
 Boolean Found = False;
 int i = 0, j = 0, end;
 char *p = *line;
 char *q = NULL;

 *key = nothing;
 end = strlen (*line) - 1; /* length of the line */

 q = strstr (*line, delimKeyWend);
 if (q)
  while ((i < XtNumber (keyWords)) && !Found)
   {
    p = strstr (*line, keyWords[i]);	/*check for keyword */
    if (p != NULL)
     {
      p += strlen (keyWords[i]) + 1; /* + semicolon */

      j = 0; /* find value */
      while ((strncmp (p + j, " ", 1) == 0) && (j < end))
	j++;
      p += j;
      /* get value */
      j = strlen (p) - strlen (q);
      strncpy (value, p, j);
      /* set pointer to current position + 1 for CR or space */
      *line = q + strlen (delimKeyWend) + 1;
      Found = True;
      *key = i + 1;
     }
    else
     i++;
   }
 value[j] = 0; /* add zero to the end */
 return (Found);
}

static Boolean findTopic(char *hlpTopic, char **restLn)
{
 Boolean Found = False;
 char *p = NULL, *p1 = NULL;
 char param[100];
 keyWordType key;

 Help.pos = 0;
 fsetpos (Help.f, &Help.pos);

 *restLn = NULL;

 while ((!feof (Help.f)) && !Found)
  if (fgets (Help.buf, BufLimit, Help.f) != NULL)
   {
    p = strstr (Help.buf, delimSt); /* begin of help item block */
    if (p)
      {
	p += strlen (delimSt);
       p1 = strstr (p, delimKeyWst);
       if (p1)
	 {
	   p1 += strlen (delimKeyWst);
	   Found = getKeyValue (&p1, &key, param);
	   if ((Found) && (key == topic))
	     if (strcmp (param, hlpTopic) == 0)
	      {
	       Found = True;
	       if (strlen (p1) > 0)
		 *restLn = p1;
	      }
	     else
	      Found = False;
	   else
	    Found = False;
	 }
      }
   }
 return(Found);
}

/* deletes all \n chars if the next char is not \t or \n */
static void parseText(char *text)
{
 int l1, l, copylen;
 char *new, *p, *q;

 l = strlen (text);
 new = calloc (l + 2, sizeof (char));
 new = strncpy (new, text, l);
 text[0] = 0;

 p = new;
 while (l > 0)
  {
   q = strchr (p, '\n');
   if (q)
     l1 = strlen (q);
   else
     l1 = 0;
   copylen = l - l1;
   if (l1 > 0)
     if (q)
      {
       if (strcspn (q + 1, "\n\t") == 0)
	copylen += 2;
       else
	 {
	  *q = ' ';		/*overwrite \n */
	  copylen += 1;
	 }
      }
   text = strncat (text, p, copylen);
   p += copylen;
   l = strlen (p);
  }
 free(new);
}

static Boolean getTopicText(char *restOLn)
{
 Boolean EndMsg = False;
 char *searchS, *p;
 int q;

 searchS = delimEnd;
 HelpItem.msg = calloc (TxtSizeLimit, sizeof (char));
 HelpItem.count = 0;
 if (restOLn != NULL)
  {
   p = strstr (restOLn, searchS);
   if (p)
    {
     HelpItem.count = strlen (restOLn) - strlen (p);
     EndMsg = True;
    }
   else
    HelpItem.count = strlen (restOLn);
   strncpy (HelpItem.msg, restOLn, HelpItem.count);
  }

 while ((!feof (Help.f)) && !EndMsg)
  if (fgets (Help.buf, BufLimit, Help.f) != NULL)
   {
     q = strlen (Help.buf);
     p = strstr (Help.buf, searchS);
     if (p) /*end text block delimiter found */
      {
       q = strlen (Help.buf) - strlen (p);
       EndMsg = True;
      }
     p = Help.buf;
     if (q > 0)
      {
       if ((HelpItem.count + q) < TxtSizeLimit)
	{
	 strncat (HelpItem.msg, p, q);
	 HelpItem.count += q;
	}
       else
	EndMsg = True;
      }
   }
 return(HelpItem.count > 0);
}

DHELP void helpCB(Widget w, XtPointer clientData, XtPointer callData)
{
 char *restLn;

 if (!initHelp(ApData.helpFile))
  manageErrorDialog(getTxtI8N(NO_HELP_FILE_MSG), NULL, NULL, getTxtI8N(OK_MSG));
 else
  {
   char buf[100];
   strcpy (buf, (char *) clientData);
   if (!findTopic (buf, &restLn))
     manageErrorDialog (getTxtI8N (NO_RELATED_HELP_TEXT_MSG), NULL, NULL, getTxtI8N (OK_MSG));
   else
    {
     if (!getTopicText (restLn))
      manageErrorDialog (getTxtI8N (NO_APPROPRIATE_HELP_MSG), NULL, NULL, getTxtI8N (OK_MSG));
     else
       {
	parseText(HelpItem.msg);

	if (strcmp (buf, "Program info") == 0)
	  {
           char msg[500];

	   sprintf(msg,"%s\n%s", HelpItem.msg, ApData.version);
	   XmTextSetString(helpTextW, msg);
	  }
	 else
	  XmTextSetString(helpTextW, HelpItem.msg);
	XtManageChild(helpDialogW);
       }
      free(HelpItem.msg);
     }
    endHelp();
  }
}

/* help inside drawing area as a result of pressing middle mouse button */
DHELP void helpTools(Sel_tool tool)
{
 char param[100], *p = &param[0];

 switch (tool)
  {
   case selectA: p = strcpy (param, "Select"); break;
   case selectPolygonA: p = strcpy (param, "Select polygon"); break;
   case copyA:
   case pasteA:
   case moveA: p = strcpy (param, "Move & resize selection"); break;
   case curveDraw: p = strcpy (param, "Curve draw"); break;
   case lineDraw: p = strcpy (param, "Line draw"); break;
   case polyDraw: p = strcpy (param, "Polygon draw"); break;
   case rectDraw:
   case squareDraw:
   case circDraw:
   case ellipseDraw: p = strcpy (param, "Rect & square & ellipse & circle draw"); break;
   case arcDraw: p = strcpy (param, "Arc draw"); break;
   case shiftHor:
   case shiftVer: p = strcpy (param, "Shift"); break;
   case fill: p = strcpy (param, "Fill area"); break;
   case contourT: p = strcpy (param, "Contour main"); break;
   case spray: p = strcpy (param, "Spray");  break;
   default: p = strcpy (param, "Visible tools"); break;
  }
 helpCB (NULL, param, NULL);
}

/* get message/string index */
static int getIndex(char *ln, int lnSize)
{
 int v = -1;
 char buf[100], *p = &buf[0];

 if (lnSize > 100)
   return (v);
 p = strncpy (buf, ln, lnSize);
 buf[lnSize] = 0;
 v = atoi (p);
 if (v > MSG_NUM)
   v = -1;
 return (v);
}

DHELP int ReadMsgFile(char *msgFile)
{
 FILE *f;
 char msgBuf[BufLimit], msgTxt[TxtSizeLimit];
 Boolean OK = True, ret_v = True;
 int count = 0, i;

 f = fopen(msgFile, "r");
 if (f == NULL) /* error opening file */
  {
   printf("Can not open/find %s\n", msgFile);
   return(False);
  }

 memset (msgTxt, 0, TxtSizeLimit);

 for (i = 0; i < MSG_NUM; i++)
  msgArr[i] = NULL;
 while ((OK) && (!feof (f)))
  {
   char *p, *p1;

   memset (msgBuf, 0, BufLimit); /* clear buffer */
   p = fgets (msgBuf, BufLimit - 1, f); /* read line from file */
   OK = ((!ferror (f)) && (p != NULL));
   p = strchr (msgBuf, '#'); /* find start delimiter in line */
   if ((p) && (*p == msgBuf[0]))
    {
     if (count) /* store previous msg into array */
      {
       if (strlen (msgTxt))
	{
	 msgTxt[strlen (msgTxt) - 1] = '\0';
	 if (msgArr[count])
	  {
	   fprintf(stderr, "Error in msg file #%d, overriding %s with %s\n", count, msgArr[count], msgTxt);
	   free (msgArr[count]);
	   ret_v = False;
	  }
	 msgArr[count] = calloc (strlen (msgTxt) + 1, sizeof (char));
	 strcpy (msgArr[count], msgTxt);
	 memset (msgTxt, 0, TxtSizeLimit);
	}
       else /* no text for message number count */
	{
	 msgArr[count] = calloc (1, sizeof (char));
	 ret_v = False;
	 fprintf(stderr, "Error in msg file: no text for message %d\n", count);
	}
      }
     p++;
     p1 = strchr (p, ' ');
     if (p1 == NULL)
      {
       OK = False;
       if (strstr (p, "END") == NULL) /* file corrupted or not properly terminated */
	 ret_v = False;
      }
     else
      count = getIndex(p, strlen (p) - strlen (p1));
    }
   else
    strcat (msgTxt, msgBuf);
  }
 fclose(f);
 return(ret_v);
}

static void setWidgetStr(Widget w, int ind, int dots)
{
 char *sett, setting[1000];

 if ((ind == 0) || (ind > MSG_NUM))
  return;

 if (msgArr[ind] == NULL)
  {
   fprintf (stderr, "Message %d missing!\n", ind);
   return;
  }
 if (dots)
  {
   strcpy(setting, msgArr[ind]);
   strcat(setting, "...");
   sett=setting;
  }
 else
  sett=msgArr[ind];

// printf("setting %s\n", sett); /* enable this to debug widget creation */
 if ((XtIsSubclass (w, xmLabelWidgetClass)) ||
     (XtIsSubclass (w, xmLabelGadgetClass)) ||
     (XtIsSubclass (w, xmRowColumnWidgetClass)))
   setLabelString(w, sett);
 else if (XtIsSubclass (w, xmBulletinBoardWidgetClass))
  setDialogTitle(w, sett);
 else if (XtIsSubclass (w, xmScaleWidgetClass))
  setTitleString(w, sett);
}

DHELP void SetWidgetStrDotsCB(Widget w, XtPointer clientData, XtPointer callData)
{
 setWidgetStr(w, atoi ((char *) clientData), TRUE);
}

DHELP void SetWidgetStrCB(Widget w, XtPointer clientData, XtPointer callData)
{
 setWidgetStr(w, atoi ((char *) clientData), FALSE);
}

DHELP char *getTxtI8N(int ind)
{
 if ((ind > 0) && (ind <= MSG_NUM))
  return (msgArr[ind]);
 else
  return (NULL);
}
