#define MAIN
#include <Xm/Xm.h>
#include <X11/Wc/WcCreate.h>
#include <X11/Xmp/Xmp.h>
#include <Xm/MessageB.h>
#include <Xm/Protocols.h>
#include <Xm/AtomMgr.h>
#include <Xm/ToggleB.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include "paint.h"

#define NO_ERR 0
#define CREATE_PIXMAP_ERR 1
#define COLOR_ALLOC_ERR 2
#define PROGRAM_PATH_ERR 4
#define CFG_FILE_ERR 8
#define MSG_FILE_ERR 16
#define NOT_ENOUGH_MEMORY_ERR 32
#define WRONG_COLOR_DEPTH_ERR 64
static Widget editMW, toolsMW, zoomAspectW;

static struct
{
 EventMask eventMask;
 XtEventHandler currEH;
 Boolean repeatsUnset;
} viewData;

/* All Wcl applications should provide at least these Wcl options: */
static XrmOptionDescRec options[] = { WCL_XRM_OPTIONS };


/* This callback is invoked when the user selects `Close' from
   the mwm frame menu on the upper left of the window border. */
static void DeleteWindowCB(Widget w, XtPointer clientData, XtPointer callData)
{
 XtAddTimeOut(3000, (XtTimerCallbackProc) * reallyExitCB, "0");
}

static void stateWatcher(Widget w, XtPointer clientData, XEvent *event)
{
 if (event->type == UnmapNotify)
  setIcon(w, ApData.img);
}

/* init default values in case configuration file is missing or is corrupted */
static void firstInit()
{
 initWidth = initHeight = 400;
 maxWidth = maxHeight = MAX_IMAGE;
 screenHeightMM = 234;
 screenWidthMM = 314;
 maxColors = 32;
 zoomDraw.thickX = zoomDraw.thickY = 10;
 foregroundDraw = TRUE;
 metricSystem = TRUE;

 ApData.imageXdpi = ApData.imageYdpi = 100.0;
 ApData.imgPath = strdup("../data/img");
 ApData.msgFile = strdup("msg/apaint.eng.msg");
 ApData.helpFile = strdup("msg/apaint.eng.hlp");
 enableSHM = True;
 pathCommon = strdup("../common");
 printCommand = strdup("/usr/bin/lpr -Praw -#%d ");
 defaultPrinter = strdup("Stylus800");
 copy2dosCommand = strdup("/usr/bin/mcopy -o");
 ApData.version= strdup("ArahPaint 2.8e");
}

static void initZoomWin(int clientData, XtIntervalId * id)
{
 if(atoi((char *) clientData)==1)
  XtAddEventHandler(zoomDrawAreaW, PointerMotionMask | LeaveWindowMask | KeyReleaseMask,
   FALSE, (XtEventHandler) mousePosZoom, NULL);
 else
  {
   setExclEventHandler (ApData.CurrEventHandler, ApData.eventMask, True);
   ChangeCursor (XtWindow (zoomDrawAreaW), ApData.cursorShape, False);
  }
}

static int initMyXstructures(Widget w)
{
 XGCValues values;
 int i = 0, colorNo = 0, resultMS, res;
 XVisualInfo viTmpl;
 char dash_list[2] =
 {4, 4};

 resultMS = NO_ERR;
 xinfo.dpy = XtDisplay (w);
 xinfo.screenNo = XDefaultScreen (xinfo.dpy);
 xinfo.visual = XDefaultVisual (xinfo.dpy, xinfo.screenNo);
 xinfo.depth = XDefaultDepth (xinfo.dpy, xinfo.screenNo);
 xinfo.trueColor = xinfo.depth > 8;
/*  xinfo.gc = XDefaultGC (xinfo.dpy, xinfo.screenNo); */
 xinfo.cmap = XDefaultColormap (xinfo.dpy, xinfo.screenNo);
 xinfo.dpyw = XDisplayWidth (xinfo.dpy, xinfo.screenNo);
 xinfo.dpyh = XDisplayHeight (xinfo.dpy, xinfo.screenNo);

 viTmpl.screen = xinfo.screenNo;
 viTmpl.depth = xinfo.depth;
 xinfo.visualInfo = XGetVisualInfo(xinfo.dpy, VisualScreenMask |
  VisualDepthMask, &viTmpl, &res);
 maxColors = colorNo = initAllAvailableColors();
 if (colorNo < 2)
   resultMS += COLOR_ALLOC_ERR;

/* shared memory initialization */
 Shm.present = False;
 Shm.alloc = False;
 Shm.img = NULL;
#ifdef MITSHM
 Shm.present = (Boolean) checkLocalAndMITSHM();
#endif

 if (enableSHM == False)
  Shm.present = False;
 values.line_width = 2;
 values.line_style = LineOnOffDash;
 values.join_style = JoinMiter;
 values.cap_style = CapButt;
 values.foreground = xinfo.white = WhitePixel (xinfo.dpy, xinfo.screenNo);
 values.background = xinfo.black = BlackPixel (xinfo.dpy, xinfo.screenNo);

 Select.gc = XCreateGC (xinfo.dpy, RootWindow (xinfo.dpy, xinfo.screenNo),
  GCLineStyle | GCLineWidth | GCForeground | GCBackground, &values);
 Select.moveGC = XCreateGC (xinfo.dpy, RootWindow (xinfo.dpy, xinfo.screenNo),
  GCForeground | GCBackground, &values);
 XSetDashes (xinfo.dpy, Select.gc, 0, dash_list, 2);
 Select.state = undef;
 Select.resLeft = Select.resRight = Select.resUp = Select.resDown = False;
 Select.allocPoints = 5000;
 Select.points = (XPoint *) (calloc (Select.allocPoints, sizeof (XPoint)));
 Select.npoints = 0;
 Select.rectangular = True;
 Select.reg = NULL;
 Select.fname[0] = 0;
 Select.imgAllocated = Select.colorCons = Select.extraColors = False;
 Select.img = NULL;

 if (xinfo.trueColor)
  {
   ApData.drawinfo.backgrndC = xinfo.pixelTC[xinfo.allocCells[1].pixel];
   ApData.drawinfo.foregrndC = xinfo.pixelTC[xinfo.allocCells[0].pixel];
  }
 else
  {
   ApData.drawinfo.backgrndC = xinfo.allocCells[1].pixel;
   ApData.drawinfo.foregrndC = xinfo.allocCells[0].pixel;
  }

 values.foreground = ApData.drawinfo.foregrndC;
 values.background = ApData.drawinfo.backgrndC;
 ApData.drawinfo.lineWidth = values.line_width = 1;
 ApData.drawinfo.lineStyle = values.line_style = LineSolid;
 values.function = GXcopy;
 values.join_style = JoinRound;
 values.cap_style = CapRound;
 if (values.line_width == 1)
   values.line_width--;
 xinfo.gc = XCreateGC(xinfo.dpy, RootWindow (xinfo.dpy, xinfo.screenNo),
  GCLineStyle | GCLineWidth | GCForeground |
  GCBackground | GCFunction | GCCapStyle | GCJoinStyle, &values);

 values.foreground = 1;
 values.background = 0;

 /* dummy pixmap allocation - only for creation of GC */
 ApData.pixy = safeCreatePixmap(xinfo.dpy, RootWindow(xinfo.dpy, xinfo.screenNo), 16, 16, 1);
 xinfo.gc1bit = XCreateGC (xinfo.dpy, ApData.pixy,
  GCLineStyle | GCLineWidth | GCForeground | GCBackground |
  GCFunction | GCCapStyle | GCJoinStyle, &values);
 XFreePixmap(xinfo.dpy, ApData.pixy);

 values.foreground = xinfo.white;
 values.background = xinfo.black;

 ApData.img = createMainImage(initWidth, initHeight);
 if (ApData.img == NULL)
  resultMS += CREATE_PIXMAP_ERR;

 ApData.width = initWidth;
 ApData.height = initHeight;

 ApData.fname[0] = 0;
 ApData.clipWin.x = ApData.clipWin.y = 0;

 ApData.ColorNo = colorNo;
 ApData.ProtColNo = ApData.TranspColNo = 0;
 ApData.viewMode = normalView;
 ApData.drawinfo.fill = False;
 ApData.drawinfo.fillType = solidFill;
 ApData.pic_change = False;
 ApData.oldCursorShape = ApData.cursorShape = XC_left_ptr;
 ApData.aspectRatio = (1.0*ApData.width)/ApData.height;
 ApData.zoomX = ApData.zoomY = 100;
 ApData.imgAllocated = True;
 ApData.CurrEventHandler = NULL;
 ApData.eventMask = 0;
 ApData.currFont = NULL;

 zoomDraw.factorX = zoomDraw.factorY = 2;
 zoomDraw.newImage = True;
 zoomDraw.drawMesh = True;

 XtAddTimeOut (500, (XtTimerCallbackProc) * initZoomWin, "1");

 Undo.gc = XCreateGC (xinfo.dpy, RootWindow (xinfo.dpy, xinfo.screenNo), 0, NULL);
 XCopyGC(xinfo.dpy, xinfo.gc,
  GCLineStyle | GCLineWidth | GCForeground | GCBackground | GCCapStyle | GCJoinStyle, Undo.gc);

 Undo.allocPoints = 20000;
 Undo.points = (XPoint *) (calloc (Undo.allocPoints, sizeof (XPoint)));
 Undo.npoints = 0;
 memcpy (&Undo.drawinfo, &ApData.drawinfo, sizeof (DrawInformation));
 Undo.imgAllocated = False;

 opBounds.x = opBounds.y = opBounds.width = opBounds.height = 0;

 /* init repeat structure */
 repeats.currNum = repeats.num = repeats.rows = repeats.cols = 1;
 repeats.crop.width = repeats.width = ApData.DAwidth = ApData.width;
 repeats.crop.height = repeats.height = ApData.DAheight = ApData.height;
 repeats.draw[0] = True;
 repeats.crop.x = repeats.crop.y = repeats.firstDraw = repeats.r[0].x = repeats.r[0].y = 0;
 repeats.mode = 0;
 repeats.distX = repeats.distY = 0;
 repeats.drawGrid = False;

 for (i = 1; i < MAX_ALL_REPEATS; i++)
   {
     repeats.draw[i] = False;
     repeats.r[i].x = repeats.r[i].y = 0;
     repeats.r[i].rType = normal;
   }

 /*set init values of input field structure array */
 for (i = 0; i < INP_NUM_COUNT; i++)
  {
   inpNumArr[i].inpField = NULL;
   inpNumArr[i].lowerB = inpNumArr[i].upperB = inpNumArr[i].value = 0;
   inpNumArr[i].enableChange = True;
  }

 sprayData.delay = 200;
 sprayData.sprayPts = 80;
 sprayData.radius = 25;
 ColorMatrix.mainWinWidth = ColorMatrix.mainWinHeight = 0;

 return (resultMS);
}

/* Scroll windows after children Callback */
static void previewEH(Widget w, XtPointer unused, XEvent * event)
{
 if (event->xbutton.button == Button1)
  manageErrorDialog(getTxtI8N (VIEW_ONLY_IN_THIS_MODE_MODE_MSG),
   NULL, NULL, getTxtI8N (OK_MSG));
 else if (event->xbutton.button == Button3)
  helpCB(NULL, "Preview in repeat (density) mode", NULL);
}


/* set repeat view mode - init all necesarry variables and structures */
DMAIN void setRepeatViewMode()
{
 int wi, h, tmp;

 repeats.currNum = repeats.num;
 if (repeats.currNum == 1)
  {
   int r, c, ww, hh, i, j;

   wi = (xinfo.dpyw / ApData.width) * ApData.width;
   h = (xinfo.dpyh / ApData.height) * ApData.height;
   r = wi / ApData.width;
   c = h / ApData.height;
   r = MAX (2, r);
   r = MIN (15, r);
   c = MAX (2, c);
   c = MIN (15, c);
   h = r * ApData.height;
   wi = c * ApData.width;
   hh = 0;
   repeats.currNum = 0;
   for (j = 0; j < r; j++)
    {
     ww = 0;
     for (i = 0; i < c; i++)
      {
       repeats.r[repeats.currNum/*j*MAX_LINE_REPEAT+i*/].x = ww;
       repeats.r[repeats.currNum/*j*MAX_LINE_REPEAT+i*/].y = hh;
       ww += ApData.width;
       repeats.currNum++;
      }
     hh += ApData.height;
    }
   viewData.repeatsUnset = True;
   repeats.width = wi;
   repeats.height = h;
   repeats.cols = c;
   repeats.rows = r;
  }
 else
  {
   wi = repeats.width = repeats.cols * repeats.distX;
   h = repeats.height = repeats.rows * repeats.distY;
  }
 ApData.DAwidth = wi;
 ApData.DAheight = h;
 switch (ApData.viewMode)
  {
   case normalView:
     viewData.currEH = ApData.CurrEventHandler;
     viewData.eventMask = ApData.eventMask;
     break;
   case densityView:
     XtVaSetValues(editMW, XmNsensitive, True, NULL);
     XtVaSetValues(toolsMW, XmNsensitive, True, NULL);
     break;
   default:	/* to make lint or -Wall switch happy */
     break;
  }
 XtVaSetValues(drawAreaWID, XmNwidth, wi, XmNheight, h, NULL);
 XtVaGetValues(clipWin, XmNheight, &ApData.clipWin.height,
  XmNwidth, &ApData.clipWin.width, NULL);
 XtVaGetValues (scrollBX, XmNvalue, &tmp, NULL);
 ApData.clipWin.x = tmp;
 XtVaGetValues (scrollBY, XmNvalue, &tmp, NULL);
 ApData.clipWin.y = tmp;
 ApData.viewMode = repeatView;
 getClipWinDims (clipWin, scrollBX, scrollBY, ApData.DAwidth,
  ApData.DAheight, &ApData.clipWin);
 setExclEventHandler (viewData.currEH, viewData.eventMask, True);
 exposeRepeat(XtWindow(drawAreaWID), ApData.clipWin.x, ApData.clipWin.y,ApData.clipWin.width,
  ApData.clipWin.height, ApData.clipWin.x, ApData.clipWin.y, repeats.drawGrid, True);
 zoomDraw.newImage = True;
 zoomMapWindow();
 XmToggleButtonSetState(repeatViewTBW, True, True);
}

#define CLIPWIN_INIT 1
#define SCROLLWIN_HLP 3
#define SCROLLBAR_INIT 4
#define NORMAL_VIEW 5
#define REPEAT_VIEW 6
#define DENSITY_VIEW 7
#define RESIZE_INIT 8
#define ZOOM_DRAW_UP 9
#define NORMAL_VIEW_INIT 11
#define REPEAT_VIEW_INIT 12

/* init DrawingArea widget */
DMAIN void initDrawAreaCB(Widget w, XtPointer clientData, XtPointer callData)
{
 int prm = atoi ((char *) clientData);

 switch (prm)
  {
   case NORMAL_VIEW_INIT:
     normalViewTBW = w;
     break;
   case REPEAT_VIEW_INIT:
     repeatViewTBW = w;
     break;
   case CLIPWIN_INIT:
     XtVaGetValues(scrollWinW, XmNclipWindow, &clipWin, NULL);
     XtVaSetValues(drawAreaWID, XmNwidth, initWidth, XmNheight, initHeight, NULL);
     XtAddEventHandler(w, PointerMotionMask | LeaveWindowMask | KeyReleaseMask, FALSE,
      (XtEventHandler) mousePos, NULL);
     viewData.repeatsUnset = False;
     break;
   case REPEAT_VIEW:
     {
      XmToggleButtonCallbackStruct *cbs = (XmToggleButtonCallbackStruct *) callData;

      if ((ApData.viewMode == repeatView) || (cbs->set == False))
       return;
      setRepeatViewMode();
     }
     break;
   case DENSITY_VIEW:
     {
      int wi, h, tmp;
      XmToggleButtonCallbackStruct *cbs = (XmToggleButtonCallbackStruct *) callData;

      if ((ApData.viewMode == densityView) || (cbs->set == False))
	return;
      wi = h = REAL_VIEW_SIZE;

      finalDraw(ApData.pixy);
      disableUndo();
      ApData.tool = none;
      viewData.currEH = ApData.CurrEventHandler;
      viewData.eventMask = ApData.eventMask;
      XtVaSetValues (editMW, XmNsensitive, False, NULL);
      XtVaSetValues (toolsMW, XmNsensitive, False, NULL);
      setExclEventHandler ((XtEventHandler) previewEH, ButtonPressMask, True);
      XtVaSetValues (drawAreaWID, XmNwidth, wi, XmNheight, h, NULL);

      setLabelString(posXYW, " ");
      XtVaGetValues (clipWin, XmNheight, &ApData.clipWin.height,
       XmNwidth, &ApData.clipWin.width, NULL);
      XtVaGetValues (scrollBX, XmNvalue, &tmp, NULL);
      ApData.clipWin.x = tmp;
      XtVaGetValues (scrollBY, XmNvalue, &tmp, NULL);
      ApData.clipWin.y = tmp;
      XtVaSetValues(scrollBX, XmNvalue, 0, NULL);
      XtVaSetValues(scrollBY, XmNvalue, REAL_VIEW_SIZE-ApData.clipWin.height-1, NULL);
      patchScrollBarPos(scrollBX, scrollBY, 0, 0);

      exposeTkanina(ApData.clipWin.x, ApData.clipWin.y,
       ApData.clipWin.width, ApData.clipWin.height, drawAreaWID);
      ApData.viewMode = densityView;
      viewData.repeatsUnset = False;
      XtManageChild(zoomAspectW);
      highlightButton(NULL);
     }
     break;
   case NORMAL_VIEW:
     {
      XmToggleButtonCallbackStruct *cbs = (XmToggleButtonCallbackStruct *) callData;

      if ((ApData.viewMode == normalView) || (cbs->set == False))
       return;
      if (ApData.viewMode == densityView)
       setExclEventHandler(viewData.currEH, viewData.eventMask, True);
      else
       {
	ApData.DAwidth = ApData.width;
	ApData.DAheight = ApData.height;
	zoomDraw.newImage = True;
	zoomMapWindow();
       }
      XtVaSetValues(drawAreaWID, XmNwidth, ApData.width, XmNheight, ApData.height, NULL);
      ApData.viewMode = normalView;
      XtVaSetValues (editMW, XmNsensitive, True, NULL);
      XtVaSetValues (toolsMW, XmNsensitive, True, NULL);
      repeats.currNum = 1;
      repeats.crop.x = repeats.crop.y = repeats.firstDraw = repeats.r[0].x = repeats.r[0].y = 0;
      repeats.width = repeats.crop.width = ApData.width;
      repeats.height = repeats.crop.height = ApData.height;
      repeats.draw[0] = True;
      if (viewData.repeatsUnset)
       {
	repeats.cols = repeats.rows = 1;
	viewData.repeatsUnset = False;
       }
      refreshDrawingArea(True, True);
     }
     break;
   case ZOOM_DRAW_UP:
     if (ApData.viewMode == densityView)
      {
       XtPopdown (GetShellParent (zoomDrawAreaW));
       return;
      }
     XtAddTimeOut (500, (XtTimerCallbackProc) * initZoomWin, "2");
     XtManageChild ((XtParent (XtParent (zoomDrawAreaW))));
     break;
   case SCROLLWIN_HLP:
     helpTools (ApData.tool);
     break;
   case RESIZE_INIT:
     getClipWinDims(clipWin, scrollBX, scrollBY, ApData.DAwidth,
      ApData.DAheight, &ApData.clipWin);
     break;
   case SCROLLBAR_INIT:
     XtVaGetValues(scrollWinW, XmNclipWindow, &clipWin,
      XmNhorizontalScrollBar, &scrollBX, XmNverticalScrollBar, &scrollBY, NULL);
     break;
  }
}

/* expose DrawingArea Callback */
static void redrawDAreaCB(Widget w, XtPointer clientData, XtPointer callData)
{
 Rectangle rect;
 XmDrawingAreaCallbackStruct *callS;

 callS = (XmDrawingAreaCallbackStruct *) callData;

 rect.x = callS->event->xexpose.x;
 rect.y = callS->event->xexpose.y;
 rect.width = callS->event->xexpose.width;
 rect.height = callS->event->xexpose.height;

 switch (ApData.viewMode)
  {
   case normalView:
    if (Select.selected && Select.rect_on) /* undraw dashed selection rectangle */
     drawSelectionBoundariesWrapped (drawAreaWin, 1, 1, 0, 0);
    if (ApData.img)
     refreshWindow(drawAreaWin, ApData.img, callS->event->xexpose.x, callS->event->xexpose.y,
      rect.width, rect.height, callS->event->xexpose.x, callS->event->xexpose.y, True);
    break;
   case repeatView:
    if (Select.selected && Select.rect_on) /* undraw dashed selection rectangle */
     drawSelectionBoundariesWrapped (drawAreaWin, 1, 1, 0, 0);
    exposeRepeat(XtWindow (drawAreaWID), callS->event->xexpose.x, callS->event->xexpose.y,
     callS->event->xexpose.width, callS->event->xexpose.height,
     callS->event->xexpose.x, callS->event->xexpose.y, repeats.drawGrid, True);
    break;
   case densityView:
    exposeTkanina(callS->event->xexpose.x, callS->event->xexpose.y,
     callS->event->xexpose.width, callS->event->xexpose.height, drawAreaWID);
    break;
  }
}

/* draw grid in repeat view */
static void drawRepeatGrid(Drawable drawable, int x, int y, int widthE, int heightE, int destX, int destY)
{
 int width, height, x1, x2, y1, y2, w1, h1, ind, offsX, offsY;
 GC gc;

 gc = Select.gc;
 offsX = destX - x;
 offsY = destY - y;
 XSetFunction (xinfo.dpy, gc, GXcopy);
 XSetForeground (xinfo.dpy, gc, xinfo.black);

 switch (repeats.mode)
  {
   case 0:
     width = widthE;
     w1 = repeats.crop.width - (x % repeats.crop.width);
     if (w1 == repeats.crop.width)
       XDrawLine(xinfo.dpy, drawable, gc, x + offsX, y + offsY, x + offsX, y + heightE + offsY);
     x2 = x + w1;
     width -= w1;
     while (width >= 0)
       {
	 XDrawLine(xinfo.dpy, drawable, gc, x2 + offsX, y + offsY, x2 + offsX, y + heightE + offsY);
	 width -= repeats.crop.width;
	 x2 += repeats.crop.width;
       }
     height = heightE;
     h1 = repeats.crop.height - (y % repeats.crop.height);
     if (h1 == repeats.crop.height)
       XDrawLine(xinfo.dpy, drawable, gc, x + offsX, y + offsY, x + widthE + offsX, y + offsY);
     y2 = y + h1;
     height -= h1;
     while (height >= 0)
       {
	 XDrawLine(xinfo.dpy, drawable, gc, x + offsX, y2 + offsY, x + widthE + offsX, y2 + offsY);
	 height -= repeats.crop.height;
	 y2 += repeats.crop.height;
       }
     break;
   case 1:
     height = heightE;
     h1 = repeats.crop.height - (y % repeats.crop.height);
     if (h1 == repeats.crop.height)
       XDrawLine(xinfo.dpy, drawable, gc, x + offsX, y + offsY, x + widthE + offsX, y + offsY);
     y2 = y + h1;
     height -= h1;
     do
      {
       x1 = x;
       width = widthE;
       ind = pointInRepeat (x, y2 - h1, repeats.currNum);
       if (ind == -1)
	 break;
       w1 = repeats.r[ind].x + repeats.crop.width - x;
       if (w1 == repeats.crop.width)
	 XDrawLine(xinfo.dpy, drawable, gc, x + offsX, y2 - h1 + offsY, x + offsX, y2 + offsY);
       x2 = x + w1;
       width -= w1;
       while (width >= 0)
	 {
	   XDrawLine(xinfo.dpy, drawable, gc, x2 + offsX, y2 - h1 + offsY, x2 + offsX, y2 + offsY);
	   width -= repeats.crop.width;
	   x2 += repeats.crop.width;
	 }
       if (height >= 0)
	 XDrawLine(xinfo.dpy, drawable, gc, x + offsX, y2 + offsY, x + widthE + offsX, y2 + offsY);
       height -= repeats.crop.height;
       y2 += repeats.crop.height;
       h1 = repeats.crop.height;
      }
     while (height > -(int) repeats.crop.height);
     break;
   case 2:
     width = widthE;
     w1 = repeats.crop.width - (x % repeats.crop.width);
     if (w1 == repeats.crop.width)
       XDrawLine (xinfo.dpy, drawable, gc, x + offsX, y + offsY, x + offsX, y + heightE + offsY);
     x2 = x + w1;
     width -= w1;
     do
      {
       y1 = y;
       height = heightE;
       ind = pointInRepeat (x2 - w1, y, repeats.currNum);
       if (ind == -1)
	 return;
       h1 = repeats.r[ind].y + repeats.crop.height - y;
       if (h1 == repeats.crop.height)
	 XDrawLine(xinfo.dpy, drawable, gc, x2 - w1 + offsX, y + offsY, x2 + offsX, y + offsY);
       y2 = y + h1;
       height -= h1;
       while (height >= 0)
	 {
	   XDrawLine(xinfo.dpy, drawable, gc, x2 - w1 + offsX, y2 + offsY, x2 + offsX, y2 + offsY);
	   height -= repeats.crop.height;
	   y2 += repeats.crop.height;
	 }
       if (width >= 0)
	 XDrawLine(xinfo.dpy, drawable, gc, x2 + offsX, y + offsY, x2 + offsX, y + heightE + offsY);
       width -= repeats.crop.width;
       x2 += repeats.crop.width;
       w1 = repeats.crop.width;
      }
     while (width > -(int) repeats.crop.width);
     break;
  }
 XSetForeground(xinfo.dpy, gc, xinfo.white);
 XSetFunction(xinfo.dpy, gc, GXinvert);
}
/* expose function for repeatView expose callback */
DMAIN void exposeRepeat(Drawable drawable, int x, int y, int widthE, int heightE,
 int destX, int destY, Boolean drawGrid, Boolean final_draw)
{
  int width, height, x1, x2, y1, y2, w1=0, h1=0, ind, offsX, offsY;
  Rectangle rp, exp, res, expTrim;

  offsX = destX - x;
  offsY = destY - y;
  rp.x = rp.y = 0;
  rp.width = repeats.width;
  rp.height = repeats.height;
  exp.x = x;
  exp.y = y;
  exp.width = widthE;
  exp.height = heightE;
  intersectRect (&rp, &exp, &expTrim);
  y2 = expTrim.y;
  height = expTrim.height;
  switch (repeats.mode)
   {
    case 0:
     while (height > 0)
      {
       x1 = x2 = expTrim.x;
       width = expTrim.width;
       y1 = y2 % repeats.crop.height;
       h1 = repeats.crop.height - y1;
       y1 += repeats.crop.y;
       if (h1 > height)
	h1 = height;
       while (width > 0)
	{
	 ind = pointInRepeat (x2, y2, repeats.currNum);
	 if (ind == -1)
	   break;
	 x1 = x2 % repeats.crop.width;
	 w1 = repeats.crop.width - x1;
	 x1 += repeats.crop.x;
	 if (w1 > width)
	  w1 = width;
	 repeatCopyArea(drawable, ApData.img, xinfo.gc, x1, y1, w1, h1, x2 + offsX, y2 + offsY,
	  False, repeats.r[ind].rType, repeats.crop.width, repeats.crop.height, 0, 0, NULL);
	 x2 += w1;
	 width -= w1;
	}
       height -= h1;
       y2 += h1;
      }
      break;
    case 1:
     while (height > 0)
      {
       x1 = x2 = expTrim.x;
       width = expTrim.width;
       y1 = y2 % repeats.crop.height;
       y1 += repeats.crop.y;
       while (width > 0)
	{
	 ind = pointInRepeat (x2, y2, repeats.currNum);
	 if (ind == -1)
	   break;
	 rp.x = repeats.r[ind].x;
	 rp.y = repeats.r[ind].y;
	 rp.width = repeats.crop.width;
	 rp.height = repeats.crop.height;
	 exp.x = x2;
	 exp.y = y2;
	 exp.width = width;
	 exp.height = height;
	 intersectRect (&rp, &exp, &res);
	 x1 = res.x - rp.x;
	 x1 += repeats.crop.x;
	 w1 = res.width;
	 h1 = res.height;
	 if (w1 > width)
	   w1 = width;
	 repeatCopyArea(drawable, ApData.img, xinfo.gc, x1, y1, w1, h1, x2 + offsX, y2 + offsY,
	   False, repeats.r[ind].rType, repeats.crop.width, repeats.crop.height, 0, 0, NULL);
	 x2 += w1;
	 width -= w1;
	}
       height -= h1;
       y2 += h1;
      }
     break;
    case 2:
      x2 = expTrim.x;
      width = expTrim.width;
      while (width > 0)
	{
	  y1 = y2 = expTrim.y;
	  height = expTrim.height;
	  x1 = x2 % repeats.crop.width;
	  x1 += repeats.crop.x;
	  while (height > 0)
	    {
	     ind = pointInRepeat (x2, y2, repeats.currNum);
	     if (ind == -1)
	       break;
	     rp.x = repeats.r[ind].x;
	     rp.y = repeats.r[ind].y;
	     rp.width = ApData.width;
	     rp.height = ApData.height;
	     exp.x = x2;
	     exp.y = y2;
	     exp.width = width;
	     exp.height = height;
	     intersectRect (&rp, &exp, &res);
	     y1 = res.y - rp.y;
	     y1 += repeats.crop.y;
	     w1 = res.width;
	     h1 = res.height;
	     if (h1 > height)
	       h1 = height;
	     repeatCopyArea(drawable, ApData.img, xinfo.gc, x1, y1, w1, h1, x2 + offsX, y2 + offsY,
	      False, repeats.r[ind].rType, repeats.crop.width, repeats.crop.height, 0, 0, NULL);
	     y2 += h1;
	     height -= h1;
	    }
	  width -= w1;
	  x2 += w1;
	}
      break;
    }
  if ((drawGrid) && (repeats.currNum > 1))
    drawRepeatGrid(drawable, expTrim.x, expTrim.y, expTrim.width, expTrim.height, destX, destY);
  if (final_draw)
    finalDraw (drawable);
}

DMAIN XImage *GetRepeatImage(XImage *srcImg, int x, int y, int width, int height)
{
  XImage *res, *tmp;
  Rectangle rp, part, resR;
  int ind, widthR, heightR, x1, x2, y1, y2, w1=0, h1=0;

  res = tmp = NULL;
  y2 = y;
  res = doXImage (8, width, height);
  if (res == NULL)
    return (res);
  rp.width = repeats.crop.width;
  rp.height = repeats.crop.height;
  tmp = SubImage (srcImg, repeats.crop.x, repeats.crop.y, rp.width, rp.height);
  if (tmp == NULL)
    {
     undoXImage(&res);
     return(res);
    }
  switch (repeats.mode)
    {
    case 0:
    case 1:
      heightR = height;
      while (heightR > 0)
	{
	  x1 = x2 = x;
	  widthR = width;
	  y1 = y2 % repeats.crop.height;
	  y1 += repeats.crop.y;
	  while (widthR > 0)
	    {
	     ind = pointInRepeat (x2, y2, repeats.currNum);
	     if (ind == -1)
	       break;
	     rp.x = repeats.r[ind].x;
	     rp.y = repeats.r[ind].y;
	     part.x = x2;
	     part.y = y2;
	     part.width = widthR;
	     part.height = heightR;
	     intersectRect (&rp, &part, &resR);
	     x1 = resR.x - rp.x;
	     x1 += repeats.crop.x;
	     w1 = resR.width;
	     h1 = resR.height;
	     if (w1 > widthR)
	       w1 = widthR;
	     mirrorImg (tmp, repeats.r[ind].rType);
	     CopyAreaImg (tmp, res, NULL, 0, 0, x1, y1, w1, h1, x2 - x, y2 - y);
	     mirrorImg (tmp, repeats.r[ind].rType);
	     x2 += w1;
	     widthR -= w1;
	    }
	  heightR -= h1;
	  y2 += h1;
	}
      break;
    case 2:
      widthR = width;
      x2 = x;
      while (widthR > 0)
	{
	  y1 = y2 = y;
	  heightR = height;
	  x1 = x2 % repeats.crop.width;
	  x1 += repeats.crop.x;
	  while (heightR > 0)
	   {
	    ind = pointInRepeat (x2, y2, repeats.currNum);
	    if (ind == -1)
	      break;
	    rp.x = repeats.r[ind].x;
	    rp.y = repeats.r[ind].y;
	    part.x = x2;
	    part.y = y2;
	    part.width = widthR;
	    part.height = heightR;
	    intersectRect (&rp, &part, &resR);
	    y1 = resR.y - rp.y;
	    y1 += repeats.crop.y;
	    w1 = resR.width;
	    h1 = resR.height;
	    if (h1 > heightR)
	      h1 = heightR;
	    mirrorImg (tmp, repeats.r[ind].rType);
	    CopyAreaImg (tmp, res, NULL, 0, 0, x1, y1, w1, h1, x2 - x, y2 - y);
	    mirrorImg (tmp, repeats.r[ind].rType);
	    y2 += h1;
	    heightR -= h1;
	   }
	  widthR -= w1;
	  x2 += w1;
	}
      break;
    }
  undoXImage (&tmp);
  return (res);
}

static void getWidgetIDCB(Widget w, XtPointer clientData, XtPointer callData)
{
 char *prm = (char *) clientData;

 if (EQUAL(prm, "mainW"))
   { mainW = w; return; }
 else if (EQUAL(prm, "scrollWinW"))
   { scrollWinW = w; return; }
 else if (EQUAL(prm, "errorDialogW"))
   { errorDialogW = w; return; }
 else if (EQUAL(prm, "picW"))
   { drawAreaWID = w; return; }
 else if (EQUAL(prm, "zoomDialogW"))
   { zoomDialogW = w; return; }
 else if (EQUAL(prm, "warnDialogW"))
   { warnDialogW = w; return; }
 else if (EQUAL(prm, "Undo"))
   { UndoBW = w; return; }
 else if (EQUAL(prm, "Cut"))
   { CutBW = w; return; }
 else if (EQUAL(prm, "Paste"))
   { PasteBW = w; return; }
 else if (EQUAL(prm, "Copy"))
   { CopyBW = w; return; }
 else if (EQUAL(prm, "DRectB"))
   { DRectBW = w; return; }
 else if (EQUAL(prm, "DSquareB"))
   { DSquareBW = w; return; }
 else if (EQUAL(prm, "DCircB"))
   { DCircBW = w; return; }
 else if (EQUAL(prm, "DArcB"))
   { DArcBW = w; return; }
 else if (EQUAL(prm, "unselectB"))
   { unselectBW = w; return; }
 else if (EQUAL(prm, "fillB"))
   { fillBW = w; return; }
 else if (EQUAL(prm, "DEllipseB"))
   { DEllipseBW = w; return; }
 else if (EQUAL(prm, "color_m"))
   { colSelW = w; return; }
 else if (EQUAL(prm, "colSelDialog"))
   { colSelDialogW = w; return; }
 else if (EQUAL(prm, "color_comp1"))
   { color_comp1W = w; return; }
 else if (EQUAL(prm, "color_comp2"))
   { color_comp2W = w; return; }
 else if (EQUAL(prm, "color_comp3"))
   { color_comp3W = w; return; }
 else if (EQUAL(prm, "rotateDialogW"))
   { rotateDialogW = w; return; }
 else if (EQUAL(prm, "fileSelection"))
   { fileSelectionW = w; return; }
 else if (EQUAL(prm, "widthPrnL"))
   { widthPrnLW = w; return; }
 else if (EQUAL(prm, "heightPrnL"))
   { heightPrnLW = w; return; }
 else if (EQUAL(prm, "helpDialog"))
   { helpDialogW = w; return; }
 else if (EQUAL(prm, "helpText"))
   { helpTextW = w; return; }
 else if (EQUAL(prm, "printDialog"))
   { printDialogW = w; return; }
 else if (EQUAL(prm, "zoomDrawArea"))
   { zoomDrawAreaW = w; return; }
 else if (EQUAL(prm, "posXY"))
   { posXYW = w; return; }
 else if (EQUAL(prm, "zoom undo"))
   { undoZoomBW = w; return; }
 else if (EQUAL(prm, "workingDialogW"))
   { workingDialogW = w; return; }
 else if (EQUAL(prm, "infoDialogW"))
   { infoDialogW = w; return; }
 else if (EQUAL(prm, "printingTW"))
   { printingTW = w; return; }
 else if (EQUAL(prm, "printingInfoLW"))
   { printingInfoLW = w; return; }
 else if (EQUAL(prm, "printProgressSW"))
   { printProgressSW = w; return; }
 else if (EQUAL(prm, "printCancelW"))
   { printCancelW = w; return; }
 else if (EQUAL(prm, "printerSelBW"))
   { printerSelBW = w; return; }
 else if (EQUAL(prm, "printModeSelBW"))
   { printModeSelBW = w; return; }
 else if (EQUAL(prm, "printXdpiSelBW"))
   { printXdpiSelBW = w; return; }
 else if (EQUAL(prm, "printYdpiSelBW"))
   { printYdpiSelBW = w; return; }
 else if (EQUAL(prm, "printToFileIW"))
   { printToFileIW = w; return; }
 else if (EQUAL(prm, "printToFileTBW"))
   { printToFileTBW = w; return; }
 else if (EQUAL(prm, "printToFileDstW"))
   { printToFileDstW = w; return; }
 else if (EQUAL(prm, "copiesW"))
   { copiesW = w; return; }
 else if (EQUAL(prm, "editM"))
   { editMW = w; return; }
 else if (EQUAL(prm, "toolsM"))
   { toolsMW = w; return; }
 else if (EQUAL(prm, "polyDrawBW"))
   { DpolyDrawBW = w; return; }
 else if (EQUAL(prm, "zoomDrawW"))
   { zoomDrawW = w; return; }
 else if (EQUAL(prm, "drawPrmW"))
   { drawPrmW = w; return; }
 else if (EQUAL(prm, "zoomAspectW"))
   { zoomAspectW = w; return; }
 else if (EQUAL(prm, "zoomScaleW"))
   { zoomScaleW = w; return; }
 else if (EQUAL(prm, "inchmmLW"))
   { inchmmLW = w; return; }
 else if (EQUAL(prm, "curveDW"))
   { curveDW = w; return; }
 printf("%s???\n", (char *)clientData);
}

/*ARGSUSED */
static void RegisterApplication(XtAppContext app)
{
 /* -- Useful shorthand for registering things with the Wcl library */
#define RCB(name, func) WcRegisterCallback(app, name, func, NULL);
 /* -- Register application specific callbacks */
 RCB("DeleteWindowCB", DeleteWindowCB)
 RCB("NewFileCB", NewFileCB)
 RCB("loadPictureCB", loadPictureCB)
 RCB("getWidgetIDCB", getWidgetIDCB)
 RCB("reallyExitCB", reallyExitCB)
 RCB("savePictureCB", savePictureCB)
 RCB("initDrawAreaCB", initDrawAreaCB)
 RCB("initColSelCB", initColSelCB)
 RCB("redrawDAreaCB", redrawDAreaCB)
 RCB("redrawColSelCB", redrawColSelCB)
 RCB("toolBarCB", toolBarCB)
 RCB("shiftPictureCB", shiftPictureCB)
 RCB("undoCB", undoCB)
 RCB("mirrorPictureCB", mirrorPictureCB)
 RCB("ChSizePicCB", ChSizePicCB)
 RCB("InitChSizeDlgCB", InitChSizeDlgCB)
 RCB("selectCB", selectCB)
 RCB("cutCopyCB", cutCopyCB)
 RCB("pastemoveCB", pastemoveCB)
 RCB("line_widthCB", line_widthCB)
 RCB("printPictureCB", printPictureCB)
 RCB("colorToolsCB", colorToolsCB)
 RCB("restoreColorCB", restoreColorCB)
 RCB("setColorNoCB", setColorNoCB)
 RCB("rotateCB", rotateCB)
 RCB("doContourCB", doContourCB)
 RCB("saveDrawingCB", saveDrawingCB)
 RCB("readFloppyCB", readFloppyCB)
 RCB("loadPictButtonCB", loadPictButtonCB)
 RCB("changeTitleCB", changeTitleCB)
 RCB("aspectRatioCB", aspectRatioCB)
 RCB("customizeMessageBoxCB", customizeMessageBoxCB)
 RCB("helpCB", helpCB)
 RCB("redrawZoomDAreaCB", redrawZoomDAreaCB)
 RCB("handleZoomDrawCB", handleZoomDrawCB)
 RCB("setButtonIconCB", setButtonIconCB)
 RCB("getRepeatPatternCB", getRepeatPatternCB)
 RCB("setZoomFactorCB", setZoomFactorCB)
 RCB("checkKbdInputCB", checkKbdInputCB)
 RCB("incDecTextFieldCB", incDecTextFieldCB)
 RCB("multiplePasteCB", multiplePasteCB)
 RCB("SetWidgetStrCB", SetWidgetStrCB)
 RCB("SetWidgetStrDotsCB", SetWidgetStrDotsCB)
 RCB("transpProtColCB", transpProtColCB)
 RCB("initPrtInterfaceCB", initPrtInterfaceCB)
 RCB("cancelPrintCB", cancelPrintCB)
 RCB("setPrintToFileCB", setPrintToFileCB)
 RCB("cancelLoadCB", cancelLoadCB)
 RCB("inpFontCB", inpFontCB)
 RCB("repeatCB", repeatCB)
 RCB("historyListCB", historyListCB)
 RCB("drawPrmCB", drawPrmCB)
 RCB("toolsCB", toolsCB)
 RCB("UDcolRemapCB", UDcolRemapCB)
}

/* unsuccessful init callback */
static void initErrorCB(Widget w, XtPointer clientData, XtPointer callData)
{
 int initError = (int) clientData, currErr;
 char msg[2048], *p = msg;

 currErr = initError;
 *p = 0;
 if (CFG_FILE_ERR & currErr)
   {
    char e[10];

    p = strcpy (e, "Error!");

    p = strcpy (msg, "Missing configuration file!");
    p = strcat (msg, "\nSave it using function \"Save setup\"\nin ArahWeave program!");
    currErr = 0;
    manageInfoDialog(msg);
    return;
   }
 if (MSG_FILE_ERR & currErr)
   {
    char e[10];

    p = strcpy (e, "Error!");

    p = strcpy (msg, "Message file is damaged!");
    p = strcat (msg, "\nTerminating application.");
    currErr = 0;
    setDialogTitle(errorDialogW, e);
   }
 if (currErr & CREATE_PIXMAP_ERR)
   {
    p = strcat (msg, getTxtI8N (PIXMAP_CREATION_ERROR_MSG));
    p = strcat (msg, "\n");
    currErr -= CREATE_PIXMAP_ERR;
   }
 if (currErr & COLOR_ALLOC_ERR)
   {
    p = strcat (msg, getTxtI8N (NOT_ENOUGH_COLORS_MSG));
    p = strcat (msg, "\n");
    currErr -= COLOR_ALLOC_ERR;
   }
 if (currErr & PROGRAM_PATH_ERR)
   {
    p = strcat (msg, getTxtI8N (PROG_PATH_ERROR_MSG));
    p = strcat (msg, "\n");
    currErr -= PROGRAM_PATH_ERR;
   }
 if (currErr & NOT_ENOUGH_MEMORY_ERR)
   {
    p = strcat (msg, getTxtI8N (NOT_ENOUGH_MEMORY_MSG));
    p = strcat (msg, "\n");
    currErr -= NOT_ENOUGH_MEMORY_ERR;
   }
 if (currErr & WRONG_COLOR_DEPTH_ERR)
   {
    p = strcat (msg, getTxtI8N (NOT_ENOUGH_COLORS_MSG));
    p = strcat (msg, "\n");
    currErr -= WRONG_COLOR_DEPTH_ERR;
   }
 manageErrorDialog(msg, reallyExitCB, "1", getTxtI8N(PROGRAM_EXIT_MSG));
}

#define CARDINAL(argc) (&argc)

int main(int argcc, String argvv[])
{
 Atom WM_DELETE_WINDOW;
 Widget appShell;
 XtAppContext app;
 char *p;
 int result = 0, i, argc;
 char *argv[30], *class;
 char addition[10][200];
 /* program's name, so it can rename itself, if it was called under different name */
 char apaint[] = "apaint";

 firstInit();
 if (!ReadCFGFile (argvv[0]))
   result = CFG_FILE_ERR;
 argvv[0] = apaint;

 if (!ReadMsgFile (ApData.msgFile))
   result += MSG_FILE_ERR;

 for (i = 0; i < argcc; i++)
   argv[i] = argvv[i];
 strcpy (addition[0], "-rf");
 argv[argcc] = addition[0];
 p = strrchr (argvv[0], '/');
 if (p)
   p += 2;
 else
   p = argvv[0] + 1;
 strcpy (addition[1], p);
 argv[argcc + 1] = addition[1];

 strcpy (addition[2], "-xrm");
 argv[argcc + 2] = addition[2];

 if (result == 0)
   strcat (ApData.fontList, getTxtI8N (FONT_ISO_MSG));
 sprintf (addition[3], "*fontList:%s", ApData.fontList);
 argv[argcc + 3] = addition[3];

 argc = argcc + 4;
 /*  -- Intialize Toolkit creating the application shell */
 class = WcAppClass(argc, argv); /* to prevent memory leak on calling WcAppClass */
 appShell = XtInitialize (NULL, class, options, XtNumber (options), CARDINAL (argc), argv);
 XtFree(class);
 installErrorHandler();
 app = XtWidgetToApplicationContext(appShell);

 /*  -- Register all application specific callbacks and widget classes */
 RegisterApplication(app);

 /*  -- Register all Motif classes, constructors, and Xmp CBs & ACTs */
 XmpRegisterAll(app);

 result += initMyXstructures (appShell);

 /*  -- Create widget tree below toplevel shell using Xrm database */
 if (WcWidgetCreation(appShell))
   exit(1);
 /*  -- Realize the widget tree */
 XtRealizeWidget(appShell);

 xinfo.W = XtWindow(appShell);

 if (!allocateBuffers(TRUE))
  result |= NOT_ENOUGH_MEMORY_ERR;

 drawAreaWin = XtWindow(drawAreaWID);
 ChangeCursor(XtWindow (mainW), ApData.oldCursorShape, True);
 ChangeCursor(drawAreaWin, ApData.oldCursorShape, True);

 WM_DELETE_WINDOW = XInternAtom (XtDisplay (appShell), "WM_DELETE_WINDOW", False);
 XmAddWMProtocolCallback (appShell, WM_DELETE_WINDOW, reallyExitCB, "0");
 /*set program icon picture and name - when iconized */
 XtAddEventHandler(appShell, StructureNotifyMask, False, (XtEventHandler) stateWatcher, NULL);

 XtVaSetValues (appShell, XmNdeleteResponse, XmDO_NOTHING, NULL);

 if (result)
  initErrorCB(appShell, (XtPointer) result, NULL);
 initHistoryList(FALSE);
 setLabelString(inchmmLW, metricSystem ? "mm" : "in");
 /* set default tool to freehand drawing */
 toolBarCB(curveDW, "0", NULL);
 /* and finally, enter the main application loop */
 XtMainLoop();
 return(0);
}
