#define MIRROR
#include <Xm/Xm.h>
#include <stdlib.h>
#include "paint.h"

/* mirror area of srcImg into dstImg, allocated by the caller */
DMIRROR void mirrorAreaImgHorizontal(XImage *dstImg, XImage *srcImg, int x, int y, 
 unsigned int width, unsigned int height)
{
 unsigned char *lPtr, *rPtr, *dstPtr;
 int i, bplSrc, bplDst;
 Rectangle srcImgR, reqR, resR;

 if ((srcImg == NULL) || (dstImg == NULL) || (dstImg == srcImg))
   return;
 if ((srcImg->data == NULL) || (dstImg->data == NULL))
   return;

 srcImgR.x = srcImgR.y = 0;
 srcImgR.width = srcImg->width;
 srcImgR.height = srcImg->height;
 reqR.x = x;
 reqR.y = y;
 reqR.width = width;
 reqR.height = height;

 if (!intersectRect(&srcImgR, &reqR, &resR))
   return;

 resR.width = MIN (resR.width, dstImg->width);
 resR.height = MIN (resR.height, dstImg->height);
 bplSrc = srcImg->bytes_per_line;
 bplDst = dstImg->bytes_per_line;
 for (i = 0; i < resR.height; i++)
  {
   lPtr = rPtr = &srcImg->data[bplSrc * (i + resR.y) + resR.x];
   rPtr += resR.width;
   dstPtr = &dstImg->data[bplDst * i + resR.width - 1];
   while (lPtr < rPtr)
    {
     *dstPtr = *lPtr;
     dstPtr--;
     lPtr++;
    }
  }
}

DMIRROR void mirrorImgHorizontal(XImage * img, unsigned int rows)
{
 unsigned char *lPtr, *rPtr;
 int i, bpl;

 if (img == NULL)
   return;
 if (img->data == NULL)
   return;
 if (rows > img->width)
   rows = img->width;
 bpl = img->bytes_per_line;
 for (i = 0; i < img->height; i++)
  {
   lPtr = &img->data[bpl * i];
   rPtr = &img->data[bpl * i + rows - 1];
   while (lPtr < rPtr)
    {
     SWAP(*lPtr, *rPtr);
     rPtr--;
     lPtr++;
    }
  }
}

/* mirror area of srcImg into dstImg, allocated by the caller */
DMIRROR void mirrorAreaImgVertical(XImage *dstImg, XImage *srcImg, int x, int y, 
 unsigned int width, unsigned int height)
{
  unsigned char *srcPtr, *dstPtr;
  int i, bplSrc, bplDst;
  Rectangle srcImgR, reqR, resR;

  if ((srcImg == NULL) || (dstImg == NULL) || (dstImg == srcImg))
    return;
  if ((srcImg->data == NULL) || (dstImg->data == NULL))
    return;
    
  srcImgR.x = srcImgR.y = 0;
  srcImgR.width = srcImg->width;
  srcImgR.height = srcImg->height;
  reqR.x = x;
  reqR.y = y;
  reqR.width = width;
  reqR.height = height;
  
  if (!intersectRect(&srcImgR, &reqR, &resR))
    return;
  
  resR.width = MIN (resR.width, dstImg->width);
  resR.height = MIN (resR.height, dstImg->height);
  bplSrc = srcImg->bytes_per_line;
  bplDst = dstImg->bytes_per_line;
  for (i = 0; i < resR.height; i++)
    {
     srcPtr = &srcImg->data[bplSrc*(i + resR.y) + resR.x];
     dstPtr = &dstImg->data[bplDst*(resR.height - i - 1)];
     memcpy (dstPtr, srcPtr, resR.width);
     srcPtr += bplSrc;
     dstPtr -= bplDst;
    }
}

DMIRROR void mirrorAreaImg(XImage *dstImg, XImage *srcImg, int x, int y, 
 unsigned int width, unsigned int height, repeatType how)
{
 switch (how)
  {
   case normal:
    break;
   case mirrorX:
    mirrorAreaImgVertical(dstImg, srcImg, x, y, width, height);
    break;
   case mirrorY:
    mirrorAreaImgHorizontal(dstImg, srcImg, x, y, width, height);
    break;
   case mirrorXY:
    mirrorAreaImgVertical(dstImg, srcImg, x, y, width, height);
    mirrorImgHorizontal(dstImg, dstImg->width);
    break;
  }
}


DMIRROR void mirrorImgVertical(XImage * img, unsigned int lines)
{
 unsigned char *line, *tPtr, *bPtr;
 int bpl;

 if (img == NULL)
  return;
 if (img->data == NULL)
  return;
 if (lines > img->height)
  lines = img->height;
 bpl = img->bytes_per_line;
 line = (unsigned char *) malloc (bpl);
 if (line == NULL)
  return;
 tPtr = &img->data[0];
 bPtr = &img->data[bpl * (lines - 1)];
 while (tPtr < bPtr)
  {
   memcpy (line, tPtr, bpl);
   memcpy (tPtr, bPtr, bpl);
   memcpy (bPtr, line, bpl);
   tPtr += bpl;
   bPtr -= bpl;
  }
 free(line);
}

DMIRROR void mirrorImgFlip(XImage *img, int wid, int hei)
{
 int x, y, newBPL;
 unsigned char *data1, *data2, *newImage;

 if (img == NULL)
  return;
 if (img->data == NULL)
  return;
 if (wid > img->width)
  wid = img->width;
 if (hei > img->height)
  hei = img->height;

 newBPL = img->height;
 newImage = malloc(img->width*newBPL);
 if (!newImage)
  return;
 for (y=0; y<hei; y++)
  {
   data1 = &img->data[y*img->bytes_per_line];
   data2 = &newImage[y];
   for (x=0; x<wid; x++, data2+=newBPL)
    *data2=*data1++;
  }

 free(img->data);
 img->data=newImage;
 img->bytes_per_line=newBPL;
 SWAP(img->width, img->height);
}

DMIRROR void mirrorImg(XImage *img, repeatType how)
{
 switch (how)
  {
   case normal:
    break;
   case mirrorX:
    mirrorImgVertical(img, img->height);
    break;
   case mirrorY:
    mirrorImgHorizontal(img, img->width);
    break;
   case mirrorXY:
    mirrorImgVertical(img, img->height);
    mirrorImgHorizontal(img, img->width);
    break;
  }
}

/* mirror picture */
DMIRROR void mirrorPictureCB(Widget w, XtPointer clientData, XtPointer callData)
{
 int i, reason = atoi((char *) clientData);

 if (!Select.selected)
  selectCB(w, "2", NULL);
 if (Select.selected)
  {
   if (Select.selected && (!Select.rect_on))
     drawSelectionBoundariesWrapped(drawAreaWin, 1, 1, 0, 0);

   Select.rect_on = True;
   highlightButton (NULL);
   if ((Undo.tool != selectA) && (Undo.tool != selectPolygonA) && (Undo.tool != moveA) &&
       (Undo.tool != mirrorHor) && (Undo.tool != mirrorVer) && (Undo.tool != mirrorXY) &&
       (Undo.tool != pasteA) && (Undo.tool != rotateA))
     finalDraw (ApData.pixy);
   disableUndo();

   if (reason==0)
    ApData.tool = mirrorVer;
   if (reason==1)
    ApData.tool = mirrorHor;
   if (reason==2)
    ApData.tool = mirrorFlip;

   drawSelectionBoundariesWrapped(drawAreaWin, 1, 1, 0, 0);

   if (!Select.imgAllocated)
    {
     Select.img = MyGetImage(ApData.img, Select.x, Select.y,
      Select.width, Select.height, ApData.viewMode);
     if (!checkXImage(Select.img))
       return;
     Select.imgAllocated = True;
    }
   if (reason==0)
    {
     mirrorImgVertical(Select.img, Select.img->height);
     if (!Select.rectangular)
      for (i = 0; i < Select.npoints; i++)
       Select.points[i].y = 2 * Select.y + Select.height - Select.points[i].y;
    }

   if (reason==1)
    {
     mirrorImgHorizontal(Select.img, Select.img->width);
     if (!Select.rectangular)
      for (i = 0; i < Select.npoints; i++)
       Select.points[i].x = 2 * Select.x + Select.width - Select.points[i].x;
    }
   if (reason==2)
    {
     int wholePicture;

     wholePicture = ((Select.height == ApData.height) &&
      (Select.y == 0) && (Select.x == 0) && (Select.width == ApData.width));
     mirrorImgFlip(Select.img, Select.img->width, Select.img->height);
     for (i = 0; i < Select.npoints; i++)
      {
       SWAP(Select.points[i].x, Select.points[i].y);
       Select.points[i].x+=Select.x - Select.y;
       Select.points[i].y+=Select.y - Select.x;
      }
     SWAP(Select.width, Select.height);

     if (wholePicture)
      {
       ApData.DAheight = ApData.height = Select.height;
       ApData.DAwidth = ApData.width = Select.width;

       changeTitleCB(NULL, ApData.fname, NULL);
       freeMainImage(&ApData.img);
       ApData.img = createMainImage(ApData.width, ApData.height);
       XtVaSetValues(drawAreaWID, XmNheight, ApData.height, XmNwidth, ApData.width, NULL);
      }

    }

   ApData.pic_change = True;
   opBounds.x = Undo.points[Undo.npoints].x = (short) Select.x;
   opBounds.y = Undo.points[Undo.npoints++].y = (short) Select.y;
   if (reason==2)
    {
     opBounds.width = Undo.points[Undo.npoints].x = (short) max(Select.width, Select.height);
     opBounds.height = Undo.points[Undo.npoints++].y = opBounds.width;
    }
   else
    {
     opBounds.width = Undo.points[Undo.npoints].x = (short) Select.width;
     opBounds.height = Undo.points[Undo.npoints++].y = (short) Select.height;
    }
 
   if (!Select.rectangular)
    {
     XDestroyRegion (Select.reg);
     XSetClipMask (xinfo.dpy, Select.moveGC, None);

     Select.reg = XPolygonRegion (Select.points, Select.npoints, WindingRule);
     XSetRegion (xinfo.dpy, Select.moveGC, Select.reg);
     refreshDrawingArea(False, True);
    }
   Undo.tool = ApData.tool;
   segmentDrawOp();
   calcOpBounds();
   refreshZoomWin(opBounds.x, opBounds.y, opBounds.width, opBounds.height, True);
   LastDrawOnAllWin(False);
  }
}

/* roll the XImage by the values passed */
static void rollXIM(XImage *xim, int offX, int offY)
{
 unsigned char buff[MAX_IMAGE];
 unsigned short swapdst[MAX_IMAGE], swapval[MAX_IMAGE];
 int bpl, remainX, width, height, x, y, i, j;
 unsigned char *data, *datb;

 bpl = xim->bytes_per_line;
 width = xim->width;
 height = xim->height;

 /* first prepare the swap tables */
 remainX = width - offX;
 if ((offY > 0) || (offX > 0))
   {
     for (y = 0; y < height; y++)
       swapval[y] = (unsigned short) y;
     for (y = 0; y < height - offY; y++)
       {
	 SWAP (swapval[y], swapval[y + offY]);
	 swapdst[y] = y + offY;
       }
     j = 0;
     for (y = height - offY; y < height; y++)
       {
	 i = y;
	 while (swapval[i] != j)
	   i++;
	 swapdst[y] = (unsigned short) i;
	 if (swapdst[y] != y)
	   SWAP (swapval[y], swapval[i]);
	 j++;
       }

     /* than perform rolling in x and y dimensions in paralell */
     for (y = 0; y < height; y++)
       {
	 if (swapdst[y] != y)
	   {
	     data = (unsigned char *) xim->data + y * bpl;
	     datb = (unsigned char *) xim->data + swapdst[y] * bpl;
	     for (x = 0; x < width; x++)
	       {
		 SWAP (*data, *datb) data++;
		 datb++;
	       }
	   }
	 /* do it for x dimension while it is still in cache */
	 if (offX > 0)
	   {
	     data = (unsigned char *) xim->data + y * bpl;
	     memcpy (buff, data, offX);
	     memcpy (data, &data[offX], remainX);
	     memcpy (&data[remainX], buff, offX);
	   }
       }
   }
}

DMIRROR void ShiftPic(int w, int h)
{
 unsigned int x1, y1, x2, y2, w1, h1, w2, h2, xd1, xd2, yd1, yd2;
 Boolean OK = True;

 if (w == 0)
   {
     w2 = w1 = ApData.width;
     h1 = ApData.height - h;
     h2 = h;
     y1 = x1 = x2 = 0;
     y2 = h;
     yd1 = xd1 = xd2 = 0;
     yd2 = h1;
     if (h1 < h2)
       {
	 SWAP (h1, h2);
	 SWAP (y1, y2);
	 SWAP (yd1, yd2);
       }
   }
 else
   {
     w1 = ApData.width - w;
     h2 = h1 = ApData.height;
     w2 = w;
     x1 = y1 = y2 = 0;
     x2 = w;
     xd1 = yd1 = yd2 = 0;
     xd2 = w1;
     if (w1 < w2)
       {
	 SWAP (w1, w2);
	 SWAP (x1, x2);
	 SWAP (xd1, xd2);
       }
   }

 rollXIM(ApData.img, w, h);

 if (OK)
   {
     refreshDrawingArea(True, True);

     if (w == 0)
       {
	Undo.points[Undo.npoints].x = 0;
	Undo.points[Undo.npoints].y = ApData.height - h;
       }
     else
       {
	Undo.points[Undo.npoints].x = ApData.width - w;
	Undo.points[Undo.npoints].y = 0;
       }
     Undo.npoints++;
     Undo.tool = ApData.tool;
     enableUndo (False);
     refreshZoomWin (zoomDraw.clipWin11.x, zoomDraw.clipWin11.y,
       zoomDraw.clipWin11.width + 2, zoomDraw.clipWin11.height + 2, True);
   }
}

DMIRROR void shiftPictureCB(Widget w, XtPointer clientData, XtPointer callData)
{
 if (Select.rect_on)
  {
   drawSelectionBoundariesWrapped (drawAreaWin, 1, 1, 0, 0);
   drawSelectionBoundariesWrapped (XtWindow (zoomDrawAreaW), zoomDraw.factorX,
    zoomDraw.factorY, zoomDraw.clipWin.x, zoomDraw.clipWin.y);
  }
 Select.rect_on = Select.selected = False;

 if (atoi((char *) clientData))
  ApData.tool = shiftVer;
 else
  ApData.tool = shiftHor;

 setToolEventHandler(ApData.tool, 0);
 highlightButton(w);
}

