#define MISC
#include <Xm/Xm.h>
#include <signal.h>
#include <stdlib.h>
#include <X11/keysym.h>
#include <Xm/ScrollBar.h>
#include <Xm/MessageB.h>
#include <Xm/BulletinB.h>
#include <Xm/TextF.h>
#include <Xm/ArrowB.h>
#include <Xm/DialogS.h>
#include <Xm/SelectioB.h>
#include <Xm/PushB.h>
#include <ctype.h>
#include <sys/utsname.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <unistd.h>
#include "paint.h"

#ifdef MITSHM
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>
#endif

typedef struct
 {
  Window wid;
  Cursor cursor;
  unsigned int c_shape, old_shape;
 }
WinCursorDef;

#define WIN_CURSORS 10

static short usedCursors = 0;
static WinCursorDef WinCursors[WIN_CURSORS];

/* copies the shared image (Pixmap) or Ximage to the Window */
DMISC void xPutBuffer(Window win, int srcx, int srcy, int dstx, int dsty, int wid, int hei)
{
 if (pixmapIsXimage)
  {
   XCopyArea(xinfo.dpy, ApData.scrPix, win, xinfo.gc, srcx, srcy, wid, hei, dstx, dsty);
   XSync(xinfo.dpy, FALSE);
  }
 else
  XPutImage(xinfo.dpy, win, xinfo.gc, ApData.scrImg, srcx, srcy, dstx, dsty, wid, hei);
}


/* allocate one full screen buffer for screen redraw, use MIT_SHM if possible; in this */
/* case also create Pixmap, which can be used for XCopyArea instead of expensive XPutImage */
DMISC int allocateBuffers(int allocate)
{
 int SIMON=0;

 pixmapIsXimage=FALSE;
 
 if ((Shm.present) && (SIMON))
  {
#ifdef MITSHM
   static XShmSegmentInfo shmInfo;
   int allOK=TRUE;

   if (!allocate)
    {
     XShmDetach(xinfo.dpy, &shmInfo);
     if (shmctl(shmInfo.shmid, IPC_RMID, 0)==0)
      printf("Shared memory free!\n");
     else
      perror("shmctl");
     return(TRUE);
    }
   ApData.scrImg = XShmCreateImage(xinfo.dpy, xinfo.visual, xinfo.depth, ZPixmap,
    0, &shmInfo, xinfo.dpyw, xinfo.dpyh);
   if (ApData.scrImg==NULL)
    allOK=FALSE;
   if (allOK)
    shmInfo.shmid = shmget(IPC_PRIVATE, ApData.scrImg->bytes_per_line*ApData.scrImg->height,
     IPC_CREAT|0777);
   printf("shmid %d\n", shmInfo.shmid);
   if (shmInfo.shmid < 0)
    { perror("shmget"); allOK=FALSE; }
   if (allOK)
    shmInfo.shmaddr = (char *)shmat(shmInfo.shmid, 0, 0);
   if (allOK)
    if (shmInfo.shmaddr == ((char *)-1))
     { perror("shmat"); allOK=FALSE; }
   ApData.scrImg->data = shmInfo.shmaddr;
   if (allOK)
    {
     XShmAttach(xinfo.dpy, &shmInfo);
     ApData.scrPix = XShmCreatePixmap(xinfo.dpy, xinfo.W, shmInfo.shmaddr,
      &shmInfo, ApData.scrImg->width, ApData.scrImg->height, ApData.scrImg->depth);
     pixmapIsXimage=TRUE;
    }
   return((allOK)&&(ApData.scrPix));
#endif /* MITSHM */
  }
 else
  {
   if (!allocate)
    return(TRUE);
   ApData.scrImg = doXImage(xinfo.depth, xinfo.dpyw,  xinfo.dpyh);
   ApData.scrPix = safeCreatePixmap(xinfo.dpy, xinfo.W, xinfo.dpyw, xinfo.dpyh, xinfo.depth);
   return((ApData.scrImg)||(ApData.scrPix));
  }
}

/* painlessly sets XmString string to resource of widget */
DMISC void xtVaSetString(Widget w, String resource, char *setting)
{
 XtVaSetValues(w, XtVaTypedArg, resource, XmRString, setting, strlen(setting)+1, NULL);
}

/* sets the passed string to widget as titleString */
DMISC void setTitleString(Widget w, char *str)
{
 xtVaSetString(w, XmNtitleString, str);
}

/* sets the passed string to widget as labelString */
DMISC void setLabelString(Widget w, char *str)
{
 xtVaSetString(w, XmNlabelString, str);
}

/* extracts a normal C string from XmString */
DMISC char *xmStringToCstring(XmString motifString)
{
 XmStringContext context;
 XmStringCharSet charSet;
 XmStringDirection dir;
 Boolean separator;
 char *cString;

 XmStringInitContext(&context, motifString);
 if (!XmStringGetNextSegment(context, &cString, &charSet, &dir, &separator))
  cString = NULL;
 XmStringFreeContext(context);
 return(cString);
}

/* returns an XImage of specified depth and dimensions */
DMISC XImage *doXImage(int depth, int x, int y)
{
 char *imagedata = NULL;
 XImage *tmpXImage;

// printf("creating XIMAGE %dx%d depth %d\n", x, y, depth);

 tmpXImage = XCreateImage(xinfo.dpy, xinfo.visual, depth, ZPixmap, 0, imagedata,
  x, y, BitmapPad(xinfo.dpy), 0);
 if (tmpXImage!=NULL)
  {
   imagedata = (char *) malloc((unsigned long) y*(tmpXImage->bytes_per_line));
   if (imagedata == NULL)
    {
     printf("%s\n", getTxtI8N(NOT_ENOUGH_MEMORY_MSG));
     XDestroyImage(tmpXImage);
     tmpXImage = NULL;
    }
   else
    tmpXImage->data = imagedata;
  }
 return(tmpXImage);
}

/* frees the XImage structure and image data */
DMISC void undoXImage(XImage **theXImage)
{
 if (*theXImage)
  XDestroyImage(*theXImage);
 *theXImage = NULL;
}

/* return TRUE if widget is viewable (i.e. it is sensible to draw in it) */
DMISC int widgetIsViewable(Widget w)
{
 return(XmGetVisibility(w)!=XmVISIBILITY_FULLY_OBSCURED);
}

/* low level error handler; just prints the error, does nothing, does not exit */
static int errorHandler(Display * theDisplay, XErrorEvent * theErrorEvent)
{
 int bufferLength = 120;
 char theBuffer[130];

 XGetErrorText(theDisplay, theErrorEvent->error_code, theBuffer, bufferLength);
 printf("X Error: %s\n", theBuffer);
 printf("Serial number of request: %ld Op Code: %d.%d Error Code: %d\n",
  theErrorEvent->serial, theErrorEvent->request_code,
  theErrorEvent->minor_code, theErrorEvent->error_code);
 printf("Resource ID of failed request: %ld on display %s.\n",
  theErrorEvent->resourceid, DisplayString(theDisplay));
 return(0);
}

static void freeAndExit(void)
{
 clearSHM();
 allocateBuffers(FALSE);
 exit(0);
}

static void errorSEGV()
{
 printf("\n%s: Segmentation fault\n", ApData.version);
 freeAndExit();
}

static void errorFPE()
{
 printf("\n%s: Floating point error\n", ApData.version);
 freeAndExit();
}

static void errorBUS()
{
 printf("\n%s: Bus error\n", ApData.version);
 freeAndExit();
}

static void xtErrorHandler(String msg)
{
 printf("\n X Toolkit error %s\n", msg);
 freeAndExit();
}

/* installs error handler; prints X protocol errors and prevents core dumps */
DMISC void installErrorHandler()
{
  /* this one will prevent exiting on not too fatal X Protocol errors */
  XSetErrorHandler(errorHandler);
  /* this one will prevent exiting on not too fatal XToolkit errors */
  XtSetErrorHandler((XtErrorHandler) xtErrorHandler);
  /* these will prevent the generation of core */
  signal(SIGSEGV, (void (*)(int)) errorSEGV);
  signal(SIGFPE, (void (*)(int)) errorFPE);
  signal(SIGBUS, (void (*)(int)) errorBUS);
}

/* RGB to HSL transformation by Ken Fiskhin of Pixar, Graphics Gems I p. 763 */
/* r,g,b: 0..255, h: 0..359, s,l: 0..1000 */
DMISC void RGBtoHSL(int r, int g, int b, int *h, int *s, int *l)
{
  double v, m, vm, r1, g1, b1, h1, s1, l1, r2, g2, b2;

  r1 = (double) r / 255.0;
  g1 = (double) g / 255.0;
  b1 = (double) b / 255.0;

  v = MAX(r1, g1);
  v = MAX(v, b1);
  m = MIN(r1, g1);
  m = MIN(m, b1);

  l1 = (m + v) / 2.0;
  if (l1 <= 0.0)
    {
      *l = (int) (l1 * 1000);
      *h = 0;
      *s = 0;
      return;
    }

  s1 = vm = v - m;
  if (s1 > 0.0)
    {
      s1 /= (l1 <= 0.5) ? (v + m) : (2.0 - v - m);
    }
  else
    {
      *h = 0;
      *l = (int) (l1 * 1000);
      *s = (int) (s1 * 1000);
      return;
    }

  r2 = (v - r1) / vm;
  g2 = (v - g1) / vm;
  b2 = (v - b1) / vm;

  if (r1 == v)
    h1 = (g1 == m ? 5.0 + b2 : 1.0 - g2);
  else if (g1 == v)
    h1 = (b1 == m ? 1.0 + r2 : 3.0 - b2);
  else
    h1 = (r1 == m ? 3.0 + g2 : 5.0 - r2);

  h1 /= 6.0;

  *h = (int) (h1 * 359 + 0.5);
  *l = (int) (l1 * 1000 + 0.5);
  *s = (int) (s1 * 1000 + 0.5);
}

/* HSL to RGB transformation by Ken Fiskhin of Pixar, Graphics Gems I p. 763 */
/* r,g,b: 0..255, h: 0..359, s,l: 0..1000 */
DMISC void HSLtoRGB(int h, int s, int l, int *r, int *g, int *b)
{
  double h1, s1, l1, r1, g1, b1, v;

  h1 = h / 360.0;
  s1 = s / 1000.0;
  l1 = l / 1000.0;

  v = (l1 <= 0.5) ? (l1 * (1.0 + s1)) : (l1 + s1 - l1 * s1);
  if (v <= 0)
   r1 = g1 = b1 = 0.0;
  else
   {
    double m, sv, fract, vsf, mid1, mid2;
    int sextant;

    m = l1 + l1 - v;
    sv = (v - m) / v;
    h1 *= 6.0;
    sextant = (int) h1;
    fract = h1 - sextant;
    vsf = v * sv * fract;
    mid1 = m + vsf;
    mid2 = v - vsf;
    switch (sextant)
     {
      default:
      case 0:
	r1 = v;
	g1 = mid1;
	b1 = m;
	break;
      case 1:
	r1 = mid2;
	g1 = v;
	b1 = m;
	break;
      case 2:
	r1 = m;
	g1 = v;
	b1 = mid1;
	break;
      case 3:
	r1 = m;
	g1 = mid2;
	b1 = v;
	break;
      case 4:
	r1 = mid1;
	g1 = m;
	b1 = v;
	break;
      case 5:
	r1 = v;
	g1 = m;
	b1 = mid2;
	break;
     }
   }
  *r = (int) (r1 * 255 + 0.5);
  *g = (int) (g1 * 255 + 0.5);
  *b = (int) (b1 * 255 + 0.5);
}

/* IMPORTANT !!!! PATCH If the scrollbar position is changed, working
 * area of scrolled window is not aware of this and that's why
 * we setup callback parameters for page increment and call
 * the corresponding callback to set up the value in scrolled window */
DMISC void patchScrollBarPos(Widget hsb, Widget vsb, int x, int y)
{
 XmScrollBarCallbackStruct *cb;

 cb = (XmScrollBarCallbackStruct *) malloc(sizeof(*cb));
 cb->reason = XmCR_PAGE_INCREMENT;
 cb->value = x;
 XtCallCallbacks(hsb, XmNpageIncrementCallback, (XtPointer) cb);
 cb->value = y;
 XtCallCallbacks(vsb, XmNpageIncrementCallback, (XtPointer) cb);
 free((char *) cb);
}

/* change cursor shape in window win to a new shape */
DMISC void ChangeCursor(Window win, unsigned int shape, Boolean StoreOld)
{
 int i = 0;

 while ((WinCursors[i].wid != win) && (i < usedCursors))
  i++;

 if (i == usedCursors)
  {
   WinCursors[usedCursors].wid = win;
   WinCursors[usedCursors].old_shape = WinCursors[usedCursors].c_shape = XC_left_ptr;
   if (usedCursors < WIN_CURSORS)
     usedCursors++;
  }

 if (WinCursors[i].cursor != 0)
   XFreeCursor(xinfo.dpy, WinCursors[i].cursor);
 if (StoreOld)
  {
   ApData.oldCursorShape = ApData.cursorShape;
   WinCursors[i].old_shape = WinCursors[i].c_shape;
   WinCursors[i].c_shape = shape;
   ApData.cursorShape = shape;
  }
 WinCursors[i].cursor = XCreateFontCursor(xinfo.dpy, shape);
 XDefineCursor(xinfo.dpy, win, WinCursors[i].cursor);
}

/* restore cursor shape of window win to the old value */
DMISC void RestoreCursor(Window win)
{
 int i = 0;
 unsigned int shape = XC_left_ptr;;

 while ((WinCursors[i].wid != win) && (i < usedCursors))
  i++;

 if (i != usedCursors)	
  shape = WinCursors[i].c_shape = WinCursors[i].old_shape;

 if (WinCursors[i].cursor!=0)
  XFreeCursor(xinfo.dpy, WinCursors[i].cursor);
 WinCursors[i].cursor = XCreateFontCursor(xinfo.dpy, shape);
 ApData.cursorShape = ApData.oldCursorShape;
 XDefineCursor(xinfo.dpy, win, WinCursors[i].cursor);
}

/* get Cursor id from window */
DMISC Cursor getWindowCursor(Window win)
{
  int i = 0;

  while ((WinCursors[i].wid != win) && (i < usedCursors))
    i++;
  if (i == usedCursors)
    return(0);
  else
    return(WinCursors[i].cursor);
}

/*refresh area of window from pixmap data */
DMISC void refreshWindow(Window win, XImage * src, int src_x, int src_y,
 int width, int height, int dest_x, int dest_y, Boolean final_draw)
{
 int diffX, diffY, srcX, srcY;

 diffX = diffY = 0;
 srcX = src_x;
 srcY = src_y;
 if (src_x < 0)
   { diffX = -src_x; srcX = 0; }
 if (src_y < 0)
   { diffY = -src_y; srcY = 0; }

 MyPutImage(win, xinfo.gc, src, srcX, srcY,
  dest_x + diffX, dest_y + diffY, width - diffX, height - diffY, NULL);
 if (final_draw)
   finalDraw(drawAreaWin);
}

/* refresh whole drawing area */
DMISC void refreshDrawingArea(Boolean wholeWindow, Boolean final_draw)
{
 Rectangle src, clip, dst;

 getClipWinDims(clipWin, scrollBX, scrollBY, ApData.DAwidth, ApData.DAheight, &ApData.clipWin);
 switch (ApData.viewMode)
  {
   case repeatView:
     refreshRepeatView(drawAreaWin, &ApData.clipWin, wholeWindow, final_draw);
     break;
   case normalView:
     if (wholeWindow)
      refreshWindow(drawAreaWin, ApData.img, ApData.clipWin.x, ApData.clipWin.y,
       ApData.clipWin.width, ApData.clipWin.height, ApData.clipWin.x, ApData.clipWin.y, final_draw);
     else
      {
       src.x=opBounds.x;
       src.y=opBounds.y;
       src.width=opBounds.width;
       src.height=opBounds.height;
       clip.x=ApData.clipWin.x;
       clip.y=ApData.clipWin.y;
       clip.width=ApData.clipWin.width;
       clip.height=ApData.clipWin.height;
       if (intersectRect(&src, &clip, &dst))
	refreshWindow(drawAreaWin, ApData.img, dst.x, dst.y, dst.width, dst.height, dst.x, dst.y, final_draw);
      }
     break;
    default:
  }
}

/* get ClipWindow width, height and x,y offsets */
DMISC void getClipWinDims(Widget clipWinW, Widget scrollBarX, Widget scrollBarY,
 unsigned int maxWidth, unsigned int maxHeight, Rectangle * clipWinR)
{
 Dimension hCW, wCW;
 int wcpy, hcpy, junk;

 XtVaGetValues(clipWinW, XmNheight, &hCW, XmNwidth, &wCW, NULL);
 clipWinR->width = wcpy = (int) wCW;
 clipWinR->height = hcpy = (int) hCW;
 if ((wcpy > maxWidth) && (hcpy > maxHeight))	/* no clip window, window shows all picture */
  {
   clipWinR->width = maxWidth;
   clipWinR->height = maxHeight;
   clipWinR->x = clipWinR->y = 0;
   return;
  }

 if (maxWidth < wcpy)
  clipWinR->width = maxWidth;
 else
  {
   XmScrollBarGetValues(scrollBarX, &wcpy, &junk, &junk, &junk);
   clipWinR->x = wcpy;
  }
 if (maxHeight < hcpy)
  clipWinR->height = maxHeight;
 else
  {
   XmScrollBarGetValues(scrollBarY, &hcpy, &junk, &junk, &junk);
   clipWinR->y = hcpy;
  }
}

/* tests if file exists ... actually, if it can be read, but that's the same
   for our purposes  */
DMISC Boolean fileExists(char *filename)
{
 FILE *fp;

 if ((fp = fopen (filename, "r")) == NULL)
  return(FALSE);
 else
  {
   fclose(fp);
   return(TRUE);
  }
}


/*create and manage warning dialog by setting callbacks to buttons */
DMISC void manageWarningDialog(char *message, XtCallbackProc OKCallback, char *OKData,
 char *OKStr, XtCallbackProc CancelCallback, char *CancelData, char *CancelStr,
 XtCallbackProc HelpCallback, char *HelpData, char *HelpStr)
{
 if (XtHasCallbacks(warnDialogW, XmNokCallback) == XtCallbackHasSome)
   XtRemoveAllCallbacks(warnDialogW, XmNokCallback);
 if (XtHasCallbacks(warnDialogW, XmNcancelCallback) == XtCallbackHasSome)
   XtRemoveAllCallbacks(warnDialogW, XmNcancelCallback);

 XtVaSetValues(warnDialogW, XtVaTypedArg, XmNmessageString, XmRString,
  message, strlen(message) + 1,
 XtVaTypedArg, XmNokLabelString, XmRString, OKStr, strlen(OKStr) + 1,
  XtVaTypedArg, XmNcancelLabelString, XmRString, CancelStr, strlen(CancelStr) + 1,
  XtVaTypedArg, XmNhelpLabelString, XmRString, HelpStr, strlen(HelpStr) + 1, NULL);
 if (OKCallback != NULL)
   XtAddCallback(warnDialogW, XmNokCallback, OKCallback, OKData);
 if (CancelCallback != NULL)
   XtAddCallback(warnDialogW, XmNcancelCallback, CancelCallback, CancelData);
 if (HelpCallback != NULL)
   {
    customizeMessageBoxCB(warnDialogW, "YesHelp", NULL);
    XtAddCallback(warnDialogW, XmNhelpCallback, HelpCallback, HelpData);
   }
 else
   customizeMessageBoxCB(warnDialogW, "NoHelp", NULL);
 XtManageChild(warnDialogW);
}

DMISC Widget GetShellParent(Widget w)
{
 Widget tmp = w;

while (tmp && !XtIsShell(tmp))
  tmp = XtParent(tmp);
return(tmp);
}

DMISC Widget GetDialogParent(Widget w)
{
 Widget tmp = w;

 while ((tmp != NULL) && !XmIsDialogShell(tmp))
  tmp = XtParent(tmp);
 return(tmp);
}

/* create and manage error dialog box with message */
DMISC void manageErrorDialog(char *message, XtCallbackProc OKCallback, char *OKData, char *OKStr)
{
 XtVaSetValues(errorDialogW, XtVaTypedArg, XmNmessageString, XmRString, message, strlen(message) + 1,
  XtVaTypedArg, XmNokLabelString, XmRString, OKStr, strlen(OKStr) + 1, NULL);
 if (XtHasCallbacks(errorDialogW, XmNokCallback) == XtCallbackHasSome)
   XtRemoveAllCallbacks(errorDialogW, XmNokCallback);

 if (OKCallback != NULL)
   XtAddCallback(errorDialogW, XmNokCallback, OKCallback, OKData);
 XtManageChild(errorDialogW);
}

/* create and information error dialog box with message */
DMISC void manageInfoDialog(char *message)
{
 XtVaSetValues(infoDialogW, XtVaTypedArg, XmNmessageString, XmRString,
  message, strlen(message) + 1, NULL);
 XtManageChild(infoDialogW);
}
/* unmanage unneccessary buttons in message box widget */
DMISC void customizeMessageBoxCB(Widget w, XtPointer clientData, XtPointer callData)
{
 Widget NoW = NULL;
 char *par = (char *) clientData;
 Boolean unmanage;

 if (strncmp(par, "No", 2) == 0)
   unmanage = TRUE;
 else
   unmanage = FALSE;
 if (strcmp(par, "NoCancel") == 0)
   NoW = XmMessageBoxGetChild(w, XmDIALOG_CANCEL_BUTTON);
 else if (strcmp(par, "NoHelp") == 0)
   NoW = XmMessageBoxGetChild(w, XmDIALOG_HELP_BUTTON);
 else if (strcmp(par, "NoIcon") == 0)
   NoW = XmMessageBoxGetChild(w, XmDIALOG_SYMBOL_LABEL);
 else if (strcmp(par, "NoMessage") == 0)
   NoW = XmMessageBoxGetChild(w, XmDIALOG_MESSAGE_LABEL);
 else if (strcmp(par, "YesHelp") == 0)
   NoW = XmMessageBoxGetChild(w, XmDIALOG_HELP_BUTTON);

 if (NoW)
  {
   if (unmanage)
     XtUnmanageChild(NoW);
   else
     XtManageChild(NoW);
  }
}

DMISC void setButtonIconCB(Widget w, XtPointer clientData, XtPointer callData)
{
 Pixmap picture;
 unsigned int width, height;
 Pixel fore, back;
 char fn[255];
 Window win;

 XtVaGetValues(w, XmNbackground, &back, XmNforeground, &fore, NULL);
 sprintf(fn, "./icons/%s", (char *) clientData);
 if (!fileExists(fn))
  sprintf(fn, "../common/icons/%s", (char *) clientData);
 win = RootWindow(xinfo.dpy, XDefaultScreen(xinfo.dpy));

 picture = loadBitmapFile(fn, xinfo.dpy, win, &width, &height, fore, back);
 if (picture)
  XtVaSetValues(w, XmNlabelType, XmPIXMAP, XmNlabelPixmap, picture, NULL);
 else
  fprintf(stdout, "Icon file %s???\n", fn);
}

DMISC void enableUndo(Boolean clearImg)
{
 if ((clearImg) && (Undo.imgAllocated))
  {
   undoXImage(&Undo.img);
   Undo.imgAllocated = FALSE;
  }
 XtVaSetValues(UndoBW, XmNsensitive, TRUE, NULL);
 if (XtIsRealized(undoZoomBW))
  XtVaSetValues(undoZoomBW, XmNsensitive, TRUE, NULL);
}

/* Disable undo, release backup images ... */
DMISC void disableUndo(void)
{
 if (Undo.imgAllocated)
  {
   undoXImage(&Undo.img);
   Undo.imgAllocated = FALSE;
  }
 XtVaSetValues(UndoBW, XmNsensitive, FALSE, NULL);
 XtVaSetValues(undoZoomBW, XmNsensitive, FALSE, NULL);
 Undo.area.x = Undo.area.y = Undo.area.width = Undo.area.height = 0;
 Undo.tool = none;
 Undo.npoints = 0;
}

/* if image was really allocated and report error if neccessary */
DMISC Boolean checkXImage(XImage * img)
{
 if (img != NULL)
  if (img->data != NULL)
   return(TRUE);
 manageErrorDialog(getTxtI8N(NOT_ENOUGH_MEMORY_MSG), NULL, NULL, getTxtI8N(OK_MSG));
 return(FALSE);
}

/* if pixmap was really created and report error if neccessary */
DMISC Boolean checkPixmap(Pixmap pixy)
{
 if (pixy == 0)
  {
   manageErrorDialog(getTxtI8N(NOT_ENOUGH_MEMORY_MSG), NULL, NULL, getTxtI8N(OK_MSG));
   return(FALSE);
  }
 return(TRUE);
}

/* Handle mouse motion inside drawing area */
DMISC void mousePos(Widget w, XtPointer unused, XEvent * event)
{
  char pos[20], *ptr = pos;
  static short xMP, yMP;

  switch (event->type)
   {
    case MotionNotify:
     if (ApData.viewMode == densityView)
      return;
     xMP = event->xmotion.x;
     yMP = event->xmotion.y;
     ptr += sprintf(pos, "%4d", xMP + 1);
     sprintf(ptr, "%5d", ApData.DAheight - yMP);
     setLabelString(posXYW, pos);
     break;
    case LeaveNotify:
     setLabelString(posXYW, getTxtI8N(HELP_MSG));
     xMP = yMP = -1;
     break;
    case KeyRelease:
     {
      KeySym key;
      char buf[5];
      int bufSize = 5, len;

      len = XLookupString((XKeyEvent *) event, buf, bufSize, &key, NULL);

      if (key == XK_Left)
       moveSelection(-1, 0);
      if (key == XK_Right)
       moveSelection(1, 0);
      if (key == XK_Up)
       moveSelection(0, -1);
      if (key == XK_Down)
       moveSelection(0, 1);

      if (len == 0)
       return;

      if ((xMP != -1) && (key == XK_space))
       changeColorKbd(drawAreaWin, xMP, yMP, Button1);
     }
     break;
   }
}

#define DATAPATH	0
#define METRICSYSTEM	1
#define INITWIDTH	2
#define INITHEIGHT	3
#define SCREENWIDTHMM	4
#define SCREENHEIGHTMM	5
#define LANGUAGE	6
#define MAXWIDTH	7
#define MAXHEIGHT	8
#define ENABLESHM	9
#define PRINTCOMMAND	10
#define DEFAULTPRINTER	11
#define FONT_SCALE	12
#define FONT_NORMAL_PT	13
#define COPY2DOSCOMMAND	14
#define MESH_THICK	15

static char *scaleFont;
static int fontSize;

/*looks in line for keywords and returns value if found */
static Boolean getSetKey(char *ln, int *paramNo)
{
  char *cfgKeyWords[] =
  {"PATH_IMAGES", "METRIC_SYSTEM", "INIT_WIDTH", "INIT_HEIGHT", 
   "SCREEN_WIDTH_MM", "SCREEN_HEIGHT_MM", "MESSAGES_FILE",
   "MAXWIDTH", "MAXHEIGHT", "ENABLE_SHM",
   "PRINT_COMMAND", "PRINTER_TYPE", "FONT_SCALE", "FONT_NORMAL_PT",
   "COPY_TO_DOS_COMMAND", "MESH_THICK"}; /* WARNING: Fix keyword type below, when adding/deleting fields */
  Boolean keyWordType[] =
  {FALSE, TRUE, TRUE, TRUE,
   TRUE, TRUE, FALSE,
   TRUE, TRUE, TRUE,
   FALSE, FALSE, FALSE, FALSE,
   FALSE, TRUE, };
  int i = 0, keyNo, j = 0, res = 0;
  Boolean notFound = TRUE;
  char *p = NULL;

  keyNo = XtNumber(cfgKeyWords);
  while ((i < keyNo) && notFound)
    {
      if ((p = strstr(ln, cfgKeyWords[i])))
	notFound = FALSE;
      else
	i++;
    }
  if (notFound)
    return(FALSE);

  res = 0;
  p = p + strlen(cfgKeyWords[i]) + 1 /* = sign */ ;
  if (keyWordType[i])		/* numeric */
    res = atoi(p);
  else
    while (strncmp(&p[j], " ", 1) == 0)
      p++;

  *paramNo = i + 1;

  switch (i)
    {
    case DATAPATH:
      if (ApData.imgPath)
	if (strlen(ApData.imgPath) > 0)
	  free(ApData.imgPath);
      if ((ApData.imgPath = calloc(strlen(p) + 1, sizeof(char))))
	  strcpy(ApData.imgPath, p);
      break;
    case METRICSYSTEM:
      metricSystem = res;
      break;
    case INITWIDTH:
      initWidth = res;
      break;
    case INITHEIGHT:
      initHeight = res;
      break;
    case MAXWIDTH:
      maxWidth = res;
      break;
    case MAXHEIGHT:
      maxHeight = res;
      break;
    case SCREENWIDTHMM:
      screenWidthMM = res;
      break;
    case SCREENHEIGHTMM:
      screenHeightMM = res;
      break;
    case MESH_THICK:
      zoomDraw.thickX = zoomDraw.thickY = res;
      break;
    case LANGUAGE:
      {
       char fn[2048], *ptr;
       int l;

       ptr = strstr(p, ".txt");
       if (ptr == NULL)
	 break;
       strcpy(fn, "./msg/apaint.");
       l = strlen(p) - strlen(ptr);
       strncat(fn, p, l);
       strcat(fn, ".msg");
       if (fileExists(fn))
	 {
	   if (ApData.msgFile)
	     free(ApData.msgFile);
	   if ((ApData.msgFile = calloc(strlen(fn) + 1, sizeof(char))))
	       strcpy(ApData.msgFile, fn);
	 }
       ptr = strstr(fn, ".msg");
       strcpy(ptr, ".hlp");
       if (fileExists(fn))
	 {
	   if (ApData.helpFile)
	     free(ApData.helpFile);
	   if ((ApData.helpFile = calloc(strlen(fn) + 1, sizeof(char))))
	       strcpy(ApData.helpFile, fn);
	 }
      }
      break;
    case PRINTCOMMAND:
      if (printCommand)
       free(printCommand);
      printCommand = calloc(strlen(p) + 1, sizeof(char));
      strcpy(printCommand, p);
      break;
    case ENABLESHM:
      enableSHM = (res > 0);
      break;
    case DEFAULTPRINTER:
      if (defaultPrinter)
	free(defaultPrinter);
      if ((defaultPrinter = calloc(strlen(p) + 1, sizeof(char))))
	  strcpy(defaultPrinter, p);
      break;
    case COPY2DOSCOMMAND:
      if (copy2dosCommand)
	free(copy2dosCommand);
      copy2dosCommand = strdup(p);
      break;
    case FONT_SCALE:
      if (scaleFont)
	free(scaleFont);
      if ((scaleFont = calloc(strlen(p) + 1, sizeof(char))))
	  strcpy(scaleFont, p);
      break;
    case FONT_NORMAL_PT:
       fontSize = atoi(p);
      break;
    }
  return(TRUE);
}

/* Read cfg file and set program parameters */
DMISC int ReadCFGFile(char *argv0)
{
 char *p;
 int i;
 FILE *f;
 char buf[5000], fname[255], cfgFile[] = ".arahne";
 char programPath[255];
 Boolean homeCFGFile;
 struct stat statBuf;
 int haveCfg = FALSE;

 scaleFont = strdup("-*-helvetica-bold-r-normal--");
 fontSize = 140;

 /* check if started as link and follow it to find executable */

 lstat(argv0, &statBuf);
 if (S_ISLNK(statBuf.st_mode))
  readlink(argv0, programPath, 255);
 else
  strcpy(programPath, argv0);

 /* chdir to executable's directory */
 strcpy(fname, programPath);
 chdir(dirName(fname));

 /* try for cfg on user's home directory */
 sprintf(fname, "%s/%s", getenv("HOME"), cfgFile);
 homeCFGFile = fileExists(fname);

 if (!homeCFGFile)
  sprintf(fname, "%s/%s", pathCommon, cfgFile);
 
  f = fopen(fname, "r");
  if (f)
   {
    while ((!feof(f)))
     if (fgets(buf, 5000, f) != NULL)
      {
       p = strchr(buf, '\n');
       if (p)
	 memset(p, 0, 1);
       p = strchr(buf, '#');
       if (p == buf)
	if (!getSetKey(&buf[1], &i))
	 i = 0;
      }
    fclose(f);
    haveCfg = TRUE;
   }
  sprintf(ApData.fontList, "%s*-%d-*-*-*-*", scaleFont, fontSize); 
  if (scaleFont)
    free(scaleFont);
  return(haveCfg);
}

/* verify function for text input: checks if input data is numeric */
DMISC void checkKbdInputCB(Widget w, XtPointer clientData, XtPointer callData)
{
 XmTextVerifyCallbackStruct *tvptr;
 XmTextBlockRec *tblock;
 int i, prm = atoi((char *) clientData);

 tvptr = (XmTextVerifyCallbackStruct *) callData;
 tblock = (XmTextBlockRec *) tvptr->text;

 if (prm == 0)	/*numeric */
   for (i = 0; i < tblock->length; i++)
     if (!((isdigit(tblock->ptr[i])) || (tblock->ptr[i] == '.')))
       tvptr->doit = FALSE;
}

/* gets the text value from the widget and returns it as integer, checks for */
/* upper and lower bounds and corrects source accordingly */
DMISC int getIntFromWidget(Widget w, int lowerB, int upperB)
{
 char *numStr = NULL;
 int value = 0;

 if (XmIsTextField(w))
   numStr = XmTextFieldGetString(w);

 if (XmIsSelectionBox(w))
   {
    XmString xms;

    XtVaGetValues(w, XmNtextString, &xms, NULL);
    numStr = xmStringToCstring(xms);
    XmStringFree(xms);
   }

 if (numStr != NULL)
   {
    value = atoi(numStr);
    XtFree(numStr);
   }
 if ((value > upperB) || (value < lowerB))
   value = setIntToWidget(w, value, lowerB, upperB);
 return(value);
}

/* sets the integer value to the text widget within bounds and returns value */
DMISC int setIntToWidget(Widget w, int value, int lowerB, int upperB)
{
 char numString[200], *nSp;
 int res;

 nSp = numString;
 if (value < lowerB)
   value = lowerB;
 if (value > upperB)
   value = upperB;
 res = sprintf(nSp, "%d", value);
 if (res > 0)
   XtVaSetValues(w, XmNvalue, nSp, NULL);
 return(value);
}

/* gets the text value from the widget and returns it as double, checks for */
/* upper and lower bounds and corrects source accordingly */
DMISC double getDoubleFromWidget(Widget w, double lowerB, double upperB, unsigned char decimals)
{
 char *numStr = NULL;
 double value = 0.0;

 if (XmIsTextField(w))
   numStr = XmTextFieldGetString(w);
 else if (XmIsSelectionBox(w))
  {
   XmString xms;

   XtVaGetValues(w, XmNtextString, &xms, NULL);
   numStr = xmStringToCstring(xms);
   XmStringFree(xms);
  }

 if (numStr != NULL)
  {
   value = atof(numStr);
   XtFree(numStr);
  }
 if ((value > upperB) || (value < lowerB))
   value = setDoubleToWidget(w, value, lowerB, upperB, decimals);
 return(value);
}


/* sets the double value to the text widget within bounds and returns value */
DMISC double setDoubleToWidget(Widget w, double value, double lowerB, double upperB, unsigned char decimals)
{
 char numString[200], *nSp;
 int res;

 nSp = numString;
 if (value < lowerB)
   value = lowerB;
 if (value > upperB)
   value = upperB;
 res = sprintf(nSp, "%.*f", decimals, value);
 if (res > 0)
   XtVaSetValues(w, XmNvalue, nSp, NULL);
 return(value);
}


static struct
{
 int changeValue;
 int ind;
 Boolean rearm;
}
rearmS;

static void rearmArrowButton(int clientData, XtIntervalId * id)
{
 int timeOut = (int) clientData, v;
 inpNumStruct *p = &inpNumArr[rearmS.ind];

 if (timeOut == 0)
  timeOut = 500;
 if (rearmS.rearm)
  {
   v = p->value + rearmS.changeValue;
   if ((v < p->lowerB) || (v > p->upperB))
     rearmS.rearm = FALSE;
   else
     p->value = setIntToWidget(p->inpField, v, p->lowerB, p->upperB);
  }
 if (timeOut > 50)
  timeOut -= timeOut / 2;
 if (rearmS.rearm)
  XtAddTimeOut(timeOut, (XtTimerCallbackProc) *rearmArrowButton, (XtPointer) timeOut);
}

DMISC inpNumStruct *initInputText(Widget inpFieldW, int value, int lowerB, int upperB)
{
 int i = 0;
 inpNumStruct *p = NULL;

 while ((i < INP_NUM_COUNT) && (inpNumArr[i].inpField != NULL))
   i++;
 if ((i == INP_NUM_COUNT) && (inpNumArr[i].inpField != NULL))
  {
   fprintf(stderr, "Too many messages - table too small!");
   return(p);
  }
 p = &inpNumArr[i];
 p->lowerB = lowerB;
 p->inpField = inpFieldW;
 p->upperB = upperB;
 p->enableChange = TRUE;

 setInpArrVal(p, value);
 rearmS.ind = -1;
 return(p);
}

DMISC void setInpArrVal(inpNumStruct * ind, int value)
{
 ind->value = setIntToWidget(ind->inpField, value, ind->lowerB, ind->upperB);
}

#define INP_FIELD_INC 0
#define INP_FIELD_DEC 1
#define INP_FIELD_VALUE_CHANGED 2
#define INP_FIELD_DISARM 3
#define INP_FIELD_FOCUS_LOST 4
DMISC void incDecTextFieldCB(Widget w, XtPointer clientData, XtPointer callData)
{
 WidgetList wList = NULL; /* it is not allocated by getValue, so do not free it !!! */

 int numChildren = 0, i = 0, prm = atoi((char *) clientData);
 Widget tmp = w;

 if (prm == INP_FIELD_DISARM)
   {
     rearmS.rearm = FALSE;
     rearmS.ind = -1;
     return;
   }
 if (XmIsArrowButton(w))
   if (rearmS.ind == -1)
     {
      XtVaGetValues(XtParent(w), XmNnumChildren, &numChildren, XmNchildren, &wList, NULL);
      while ((i < numChildren) && (!XmIsTextField(*wList)))
	{
	  i++;
	  wList++;
	}
      if (i == numChildren)
	return;
      tmp = *wList;
      i = 0;
     }
   else
     i = rearmS.ind;
 else
   i = 0;

 if (i == 0) /* search index only if Text widget or first click on arrow b */
   {
    while ((i < INP_NUM_COUNT) && (inpNumArr[i].inpField != tmp))
      i++;
    if ((i == INP_NUM_COUNT) && (inpNumArr[i].inpField != tmp))
      return;
   }

 if ((XmIsArrowButton(w)) && (inpNumArr[i].enableChange))
   {
    if (prm == INP_FIELD_INC)
      rearmS.changeValue = 1;
    else if (prm == INP_FIELD_DEC)
      rearmS.changeValue = -1;
    rearmS.ind = i;
    rearmS.rearm = TRUE;
    rearmArrowButton(0, NULL);
    return;
   }
 if (prm == INP_FIELD_VALUE_CHANGED)
   inpNumArr[i].value = getIntFromWidget(tmp, -MAX_IMAGE, MAX_IMAGE);
 else if (prm == INP_FIELD_FOCUS_LOST)
   inpNumArr[i].value = getIntFromWidget(tmp, inpNumArr[i].lowerB, inpNumArr[i].upperB);
}

static void mergeImages(XImage *dstImg, XImage *srcImg, int x, int y,
 int width, int height, int dstX, int dstY)
{
 unsigned char *d, *s;
 int r, c;
 if (!ApData.specialPutImg)
   return;

 if (x < 0)
   x = 0;
 if (y < 0)
   y = 0;
 if (dstX < 0)
   dstX = 0;
 if (dstY < 0)
   dstY = 0;

 if ((x + width) > srcImg->width)
   width = srcImg->width - x;
 if ((dstX + width) > dstImg->width)
   width = dstImg->width - dstX;
 if ((y + height) > srcImg->height)
   height = srcImg->height - y;
 if ((dstY + height) > dstImg->height)
   height = dstImg->height - dstY;
 for (r = 0; r < height; r++)
  {
   s = (unsigned char *) srcImg->data + srcImg->bytes_per_line*(y + r) + x;
   d = (unsigned char *) dstImg->data + dstImg->bytes_per_line*(dstY + r) + dstX;
   for (c = 0; c < width; c++)
    *d++ = ApData.transfTbl[*s++][*d];
  }
}

DMISC Boolean putXImageTP(XImage *baseImg, XImage *srcImg, Drawable drwbl, GC gc,
 int srcX, int srcY, int width, int height, int pixX, int pixY, int dstX, int dstY, Region region)
{
 XImage *img;
 int newWidth = width, newHeight = height, clipX = pixX, clipY = pixY;
 int diffX = 0, diffY = 0;

 if ((baseImg == NULL) || (srcImg == NULL))
   return(FALSE);
 if (ApData.specialPutImg)
  {
   if (pixX < 0)
    {
     clipX = 0;
     newWidth = pixX + width;
     diffX = newWidth - width;
    }
   if (pixY < 0)
    {
     clipY = 0;
     newHeight = pixY + height;
     diffY = newHeight - height;
    }
   newWidth = ((clipX + width) > ApData.width) ? ApData.width - clipX : newWidth;
   newHeight = ((clipY + height) > ApData.height) ? ApData.height - clipY : newHeight;
   if ((newWidth <= 0) || (newHeight <= 0) || ((dstX - diffX + newWidth) <= 0)
       || ((dstY - diffY + newHeight) <= 0))
     return(TRUE);
   img = SubImage(baseImg, clipX, clipY, newWidth, newHeight);

   if (!checkXImage(img))
     return(FALSE);
   mergeImages(img, srcImg, srcX - diffX, srcY - diffY, newWidth, newHeight, 0, 0);
   MyPutImage(drwbl, gc, img, 0, 0, dstX - diffX, dstY - diffY, newWidth, newHeight, region);
   undoXImage(&img);
  }
 else
  MyPutImage(drwbl, gc, srcImg, srcX, srcY, dstX, dstY, width, height, region);
 return(TRUE);
}

/* return TRUE if MITSHM extension supported and local connection to server */
DMISC int checkLocalAndMITSHM()
{
 int dummy;
 char *dname;
 struct utsname unixid;

#ifdef MITSHM
 if (XQueryExtension(xinfo.dpy, "MIT-SHM", &dummy, &dummy, &dummy))
  {
   Bool shmPixy;

   XShmQueryVersion(xinfo.dpy, &dummy, &dummy, &shmPixy);
   if (!shmPixy)
     return(FALSE);
   dname = XDisplayString(xinfo.dpy);
   uname(&unixid);
   if (strlen(dname) < 2)
     return(FALSE);
   strtok(dname, ":");
   if (!strcmp(":0", dname))
     return(TRUE);
   if (!strcmp(":0.0", dname))
     return(TRUE);
   if (!strcmp("unix", dname))
     return(TRUE);
   if (!strcmp(unixid.nodename, dname))
     return(TRUE);
  }
#endif
 return(FALSE);
}

/*clear allocated shared memory - if any */
DMISC void clearSHM()
{
 int res;

#ifdef MITSHM
 if((Shm.present) && (Shm.alloc))
  {
   res = shmdt(Shm.Info.shmaddr);
   if (res != 0)
     printf("Error detaching shared memory!\n");
   XShmDetach(xinfo.dpy, &Shm.Info);
   if (shmctl(Shm.Info.shmid, IPC_RMID, 0) == -1)
     printf("Error freeing shm with shmctl\n");
   Shm.alloc = FALSE;
   Shm.img->data = NULL;
   XDestroyImage(Shm.img);
   XSync(xinfo.dpy, FALSE);
   Shm.img = NULL;
   if (res)
    printf("Unable to free shared memory %d!!!\n", Shm.Info.shmid);
  }
#endif
}


DMISC Pixmap CreatePixmap(Display *display, Drawable drawable, unsigned int width,
 unsigned int height, unsigned int depth)
{
#ifdef MITSHM
 if (Shm.present)
  {
   XImage *img;
   int res;
   Pixmap pixy = 0;

// printf("creating SHARED pixmap %dx%d depth %d\n", width, height, depth);
   XSync(xinfo.dpy, FALSE);
   if (Shm.alloc)
     clearSHM();
   else
     Shm.alloc = FALSE;
   img = XShmCreateImage(display, xinfo.visual, depth, ZPixmap, 0, &Shm.Info, width, height);

   Shm.Info.shmid = shmget(IPC_PRIVATE, (img->bytes_per_line) * img->height, IPC_CREAT | 0777);
   if (Shm.Info.shmid < 0)
    {
     printf("Shared memory request not granted %d %d\n", width, height);
     img->data = NULL;
     XDestroyImage(img);
     return(safeCreatePixmap(display, drawable, width, height, depth));
    }
   Shm.Info.shmaddr = (char *) shmat(Shm.Info.shmid, 0, 0);
   if (Shm.Info.shmaddr == ((char *) -1))
    {
     res = shmctl(Shm.Info.shmid, IPC_RMID, 0);
     if (res != 0)
       printf("Error freeing segment\n");
     return(safeCreatePixmap(display, drawable, width, height, depth));
    }

   img->data = Shm.Info.shmaddr;
   res = XShmAttach(xinfo.dpy, &Shm.Info);
   if (res != 1)
    {
     res = shmctl(Shm.Info.shmid, IPC_RMID, 0);
     printf("Error attaching shared segment!\n");
     return(safeCreatePixmap(display, drawable, width, height, depth));
    }
   Shm.alloc = TRUE;
   Shm.img = img;
   pixy = XShmCreatePixmap(display, drawable, Shm.Info.shmaddr,
     &Shm.Info, width, height, depth);
   if (pixy == 0)
    {
     clearSHM();
     printf("Error creating shm pixmap! \n");
     return(safeCreatePixmap(display, drawable, width, height, depth));
    }
   else
    return(pixy);
  }
 else
  return(safeCreatePixmap(display, drawable, width, height, depth));
#else
 return(safeCreatePixmap(display, drawable, width, height, depth));
#endif
}

static Boolean PixmapAllocError = FALSE;

static int pixmapErrorHandler(Display * dpy, XErrorEvent * error)
{
 PixmapAllocError = TRUE;
 return(TRUE);
}

/* create pixmap and wait for success, then return pixmap id */
DMISC Pixmap safeCreatePixmap(Display * display, Drawable drawable,
 int width, int height, int depth)
{
 Pixmap thePixmap;
 Drawable root;
 int x, y;
 unsigned int rw, rh, bw, rd;
 XErrorHandler oldHandler = NULL;

// printf("creating safe pixmap %dx%d depth %d\n", width, height, depth);

 PixmapAllocError = FALSE;
 /* install our error handler */
 oldHandler = XSetErrorHandler(pixmapErrorHandler);
 /* attempt to create pixmap */
 thePixmap = XCreatePixmap(display, drawable, width, height, depth);
 XGetGeometry(display, thePixmap, &root, &x, &y, &rw, &rh, &bw, &rd);
 /* wait to see if server allocated pixmap */
 XSync(xinfo.dpy, FALSE);
 /* restore standard error handler */
 XSetErrorHandler(oldHandler);
 return(PixmapAllocError ? (Pixmap) 0 : thePixmap);
}

DMISC void FreePixmap(Display * display, Pixmap pixy)
{
 if ((Shm.present) && (Shm.alloc))
  {
   XSync(xinfo.dpy, FALSE);
   XFreePixmap(display, pixy);
   clearSHM();
  }
 else
  XFreePixmap(display, pixy);
}

/* do the intersection of first and second rectangle, returning the resulting
   rectangle; if the intersection is empty, return FALSE, TRUE otherwise */
DMISC Boolean intersectRect(Rectangle *first, Rectangle *second, Rectangle *result)
{
 int diff, width, height;

 if ((!first) || (!second) || (!result))
   return(FALSE);

 width = result->width = first->width;
 height = result->height = first->height;
 result->x = first->x;
 result->y = first->y;
 if (first->x < second->x)
   {
     width -= second->x - first->x;
     result->x = second->x;
   }
 if (first->y < second->y)
   {
     height -= second->y - first->y;
     result->y = second->y;
   }
 diff = result->x + width - second->x - second->width;
 if (diff > 0)
   width -= diff;
 diff = result->y + height - second->y - second->height;
 if (diff > 0)
   height -= diff;

 if ((width <= 0) || (height <= 0))
   return(FALSE);

 result->width = (unsigned int) width;
 result->height = (unsigned int) height;
 return(TRUE);
}

DMISC XImage *SubImage(XImage *src, int x, int y, unsigned int width, unsigned int height)
{
 XImage *dst;
 int bpl, i, bplD, j = 0;

 if ((src == NULL) || (src->data == NULL) ||
   ((x + width) > src->width) || ((y + height) > src->height))
  return(NULL);
 dst = doXImage(src->depth, width, height);
 if (dst == NULL)
  return(NULL);

 bpl = src->bytes_per_line;
 bplD = dst->bytes_per_line;
 for (i = y; j < height; i++, j++)
  memcpy(&dst->data[bplD * j], &src->data[bpl * i + x], width);
 return(dst);
}

DMISC void highlightButton(Widget w)
{
 Pixel color1, color2;
 static Widget oldButton = NULL;

 if (oldButton)
  {
   XtVaGetValues(oldButton, XmNtopShadowColor, &color1,
    XmNbottomShadowColor, &color2, NULL);
   XtVaSetValues(oldButton, XmNtopShadowColor, color2,
    XmNbottomShadowColor, color1, NULL);
   oldButton = NULL;
  }
 if ((w == NULL) || (!XmIsPushButton(w)))
   return;
 XtVaGetValues(w, XmNtopShadowColor, &color1, XmNbottomShadowColor, &color2, NULL);
 XtVaSetValues(w, XmNtopShadowColor, color2, XmNbottomShadowColor, color1, NULL);
 oldButton = w;
}

/* reimplementation of dirname for systems that do not have it */
DMISC char *dirName(char *filen)
{
 static char dirn[260];
 int i;
 
 strncpy(dirn, filen, 260);
 i=strlen(dirn);
 while ((i>0)&&(dirn[i]!='/')) i--;
 dirn[i]=0;
 return(dirn);
}

/* reimplementation of basename for systems that do not have it */
DMISC char *baseName(char *filen)
{
 int i;
 
 i=strlen(filen);
 while ((i>0)&&(filen[i]!='/')) i--;
 if (filen[i]=='/') i++;
 return(&filen[i]);
}

/* check if point is inside rectangle */
DMISC Boolean pointInRect(int x, int y, Rectangle * r)
{
 if ((x >= r->x) && (y >= r->y) &&
   (x < (r->x + (int) r->width)) && (y < (r->y + (int) r->height)))
  return(TRUE);
 else
  return(False);
}

/* function similar to XCopyArea, just XImage is used instead of Drawables */
DMISC void CopyAreaImg(XImage * src, XImage * dest, Region clipRegion, int clipX,
 int clipY, int src_x, int src_y, unsigned int width, unsigned int height, int dest_x, int dest_y)
{
 Rectangle srcR, tmpR, dstR, imgR;
 int i;
 unsigned char *srcP, *dstP;
 int width8, width2do;

 if ((src == NULL) || (dest == NULL))
   return;
 if (src->depth != dest->depth)
   return;

 /*check first if the copyied area lies inside the src image */
 imgR.x = src_x;
 imgR.y = src_y;
 imgR.width = width;
 imgR.height = height;
 tmpR.x = tmpR.y = 0;
 tmpR.width = src->width;
 tmpR.height = src->height;
 if (!intersectRect(&imgR, &tmpR, &srcR))
   return;

 /*check if the copyied area lies inside the dst image */
 imgR.x = dest_x;
 imgR.y = dest_y;
 imgR.width = srcR.width;
 imgR.height = srcR.height;
 tmpR.x = tmpR.y = 0;
 tmpR.width = dest->width;
 tmpR.height = dest->height;
 if (!intersectRect(&imgR, &tmpR, &dstR))
   return;
 if (clipRegion == NULL)
  for (i = 0; i < dstR.height; i++)
   {
    srcP = &src->data[src->bytes_per_line * (srcR.y + i) + srcR.x];
    dstP = &dest->data[dest->bytes_per_line * (dstR.y + i) + dstR.x];
    memcpy(dstP, srcP, dstR.width);
   }
 else
   { /* simulate clipping of GC when puting Ximage on Ximage */
     /* create GC region, draw filled rectangle in it - only */
     /* pixels inside region will be drawn - then use this */
     /* pixmap for selective copying from src to dst */
    int j, k, diffEnd, diffStart, xoffs, yoffs, startBit, byte1len;
    unsigned char *bitP;
    XImage *tmpImg;
    Boolean LSB;

    XSetRegion(xinfo.dpy, xinfo.gc1bit, clipRegion);
    XSetClipOrigin(xinfo.dpy, xinfo.gc1bit, clipX, clipY);
    XSetForeground(xinfo.dpy, xinfo.gc1bit, 1);

    XFillRectangle(xinfo.dpy, ApData.pixy, xinfo.gc1bit, dstR.x, dstR.y,
		    dstR.width, dstR.height);
    XSetClipMask(xinfo.dpy, xinfo.gc1bit, None);
    XSetClipOrigin(xinfo.dpy, xinfo.gc1bit, 0, 0);
    if (Shm.alloc)
     {
      XSync(xinfo.dpy, False);
      tmpImg = Shm.img;
      LSB = ((tmpImg->bitmap_bit_order == LSBFirst) &&
	     (tmpImg->byte_order == LSBFirst));
      diffStart = dstR.x % 8;
      startBit = diffStart;
      yoffs = dstR.y;
      xoffs = dstR.x / 8;
      byte1len = diffStart ? 8 - diffStart : 0;
     }
    else
     {
      diffStart = xoffs = yoffs = startBit = byte1len = 0;
      tmpImg = XGetImage(xinfo.dpy, ApData.pixy, dstR.x,
       dstR.y, dstR.width, dstR.height, AllPlanes, ZPixmap);
      if (tmpImg == NULL)
	return;
      LSB = ((tmpImg->bitmap_bit_order == LSBFirst) && (tmpImg->byte_order == LSBFirst));
     }
    width8 = dstR.width;
    byte1len = byte1len > dstR.width ? dstR.width : byte1len;
    diffEnd = (dstR.width - byte1len) % 8;
    width8 -= byte1len + diffEnd;
    if (LSB)
     {
      for (j = 0; j < dstR.height; j++)
       {
	dstP = &dest->data[dest->bytes_per_line * (j + dstR.y) + dstR.x];
	bitP = &tmpImg->data[tmpImg->bytes_per_line * (j + yoffs) + xoffs];
	srcP = &src->data[src->bytes_per_line * (srcR.y + j) + srcR.x];
	width2do = width8;
	if (byte1len)
	 {
	  for (k = startBit; k < byte1len + startBit; k++, dstP++, srcP++)
	    if ((1 << k) & *bitP)
	      *dstP = *srcP;
	  bitP++;
	 }
	while (width2do > 0)
	  {
	    if (*bitP)
	     {
	      for (k = 0; k < 8; k++, dstP++, srcP++)
		if ((1 << k) & *bitP)
		  *dstP = *srcP;
	     }
	    else
	     {
	      dstP += 8;
	      srcP += 8;
	     }
	    bitP++;
	    width2do -= 8;
	  }
	/* check out the remaining part of the byte */
	for (k = 0; k < diffEnd; k++, dstP++, srcP++)
	  if ((1 << k) & *bitP)
	    *dstP = *srcP;
       }
     }
    if (!Shm.alloc)
      undoXImage(&tmpImg);
    /* clear pixmap */
    XSetForeground(xinfo.dpy, xinfo.gc1bit, 0);
    XFillRectangle(xinfo.dpy, ApData.pixy, xinfo.gc1bit, dest_x, dest_y, width, height);
    XSetForeground(xinfo.dpy, xinfo.gc1bit, 1);
   }
}

DMISC void CopyAreaImgTP(XImage *src, XImage *dest, Region clipRegion, int clipX,
 int clipY, int src_x, int src_y, unsigned int width, unsigned int height, int dest_x, int dest_y)
{
 if (ApData.specialPutImg)
  {
   Rectangle srcR, tmpR, dstR, imgR;
   XImage *mrgImg;

   if ((src == NULL) || (dest == NULL))
     return;
   if (src->depth != dest->depth)
     return;
   /*check first if the copyied area lies inside the src image */
   imgR.x = src_x;
   imgR.y = src_y;
   imgR.width = width;
   imgR.height = height;
   tmpR.x = tmpR.y = 0;
   tmpR.width = src->width;
   tmpR.height = src->height;
   if (!intersectRect(&imgR, &tmpR, &srcR))
     return;

   /* check if the copied area lies inside the dst image */
   imgR.x = dest_x;
   imgR.y = dest_y;
   imgR.width = srcR.width;
   imgR.height = srcR.height;
   tmpR.x = tmpR.y = 0;
   tmpR.width = dest->width;
   tmpR.height = dest->height;
   if (!intersectRect(&imgR, &tmpR, &dstR))
     return;

   mrgImg = SubImage(dest, dstR.x, dstR.y, dstR.width, dstR.height);
   if (!checkXImage(mrgImg))
     return;
   mergeImages(mrgImg, src, srcR.x, srcR.y, dstR.width, dstR.height, 0, 0);
   CopyAreaImg(mrgImg, dest, clipRegion, clipX, clipY, 0, 0, mrgImg->width,
    mrgImg->height, dest_x, dest_y);
   undoXImage(&mrgImg);
  }
 else
  CopyAreaImg(src, dest, clipRegion, clipX, clipY, src_x, src_y, width, height, dest_x, dest_y);
}

/* get XImage from drawable depending on the viewMode */
DMISC XImage *MyGetImage(XImage * img, int x, int y, unsigned int width,
 unsigned int height, viewType viewMode)
{
 XImage *res;

 res = NULL;
 switch (viewMode)
  {
   case repeatView:
    res = GetRepeatImage(img, x, y, width, height);
    break;
   default:
    res = SubImage(img, x, y, width, height);
    break;
  }
 return (res);
}

/* for all bits set inside Pixmap pixy set equivalent bytes in image */
/* actually it transfers the result of a X Server drawing into a pixmap */
/* into an XImage. This is because the server can not draw in XImage and */
/* because 8 bit Pixmaps in >= 16bpp mode are not supported */
DMISC void draw2Image(XImage *img, Pixmap pixy, Boolean shared, int x, int y, unsigned int width,
 unsigned int height, unsigned char drawColorInd)
{
 XImage *tmpImg;
 unsigned char *imgP, *bitP, colorInd;
 int j, k, diffEnd, diffStart, xoffs, yoffs, startBit, byte1len;
 Boolean LSB;
 int width8, width2do;
 Rectangle pixyR, opR, clipR;

 if ((img == NULL) || (img->data == NULL))
   return;
 pixyR.x = pixyR.y = 0;
 opR.x = x;
 opR.y = y;
 opR.width = width;
 opR.height = height;
 if (shared)
  {
   XSync(xinfo.dpy, FALSE);
   pixyR.width = Shm.img->width;
   pixyR.height = Shm.img->height;
   if (!intersectRect(&opR, &pixyR, &clipR))
     return;
   tmpImg = Shm.img;
   LSB = ((tmpImg->bitmap_bit_order==LSBFirst) && (tmpImg->byte_order==LSBFirst));
   diffStart = clipR.x % 8;
   startBit = diffStart;
   yoffs = clipR.y;
   xoffs = clipR.x / 8;
   byte1len = diffStart ? 8 - diffStart : 0;
  }
 else
  {
   Window win;
   int dummy1, dummy2;

   if (XGetGeometry(xinfo.dpy, pixy, &win, &dummy1, &dummy1, &pixyR.width,
     &pixyR.height, &dummy2, &dummy1) == 0)
     return;
   diffStart = xoffs = yoffs = startBit = byte1len = 0;
   if (!intersectRect(&opR, &pixyR, &clipR))
     return;
   tmpImg = XGetImage(xinfo.dpy, pixy, clipR.x, clipR.y,
     clipR.width, clipR.height, AllPlanes, ZPixmap);
   if (tmpImg == NULL)
     return;
   LSB = ((tmpImg->bitmap_bit_order==LSBFirst) && (tmpImg->byte_order==LSBFirst));
  }

 colorInd = (unsigned char) xinfo.allocCells[drawColorInd].pixel;

 width8 = clipR.width;
 byte1len = byte1len > clipR.width ? clipR.width : byte1len;
 diffEnd = (clipR.width - byte1len) % 8;
 width8 -= byte1len + diffEnd;
 if (LSB)
   for (j = 0; j < clipR.height; j++)
     {
       imgP = &img->data[img->bytes_per_line * (j + clipR.y) + clipR.x];
       bitP = &tmpImg->data[tmpImg->bytes_per_line * (j + yoffs) + xoffs];
       width2do = width8;
       if (ApData.specialPutImg)
	 {
	   if (byte1len)
	     {
	       for (k = startBit; k < (byte1len + startBit); k++, imgP++)
		 if ((1 << k) & *bitP)
		   *imgP = ApData.transfTbl[colorInd][*imgP];
	       bitP++;
	     }
	   while (width2do > 0)
	     {
	       if (*bitP == 0)
		 imgP += 8;
	       else
		 for (k = 0; k < 8; k++, imgP++)
		   if ((1 << k) & *bitP)
		     *imgP = ApData.transfTbl[colorInd][*imgP];
	       bitP++;
	       width2do -= 8;
	     }
	   for (k = 0; k < diffEnd; k++, imgP++)
	     if ((1 << k) & *bitP)
	       *imgP = ApData.transfTbl[colorInd][*imgP];
	 }
       else
	 {
	   if (byte1len)
	     {
	       for (k = startBit; k < (byte1len + startBit); k++, imgP++)
		 if ((1 << k) & *bitP)
		   *imgP = colorInd;
	       bitP++;
	     }
	   while (width2do > 0)
	     {
	       switch (*bitP)
		{
		 case 0:
		   imgP += 8;
		   break;
		 case 0xff:
		   for (k = 0; k < 8; k++, imgP++)
		     *imgP = colorInd;
		   break;
		 default:
		   for (k = 0; k < 8; k++, imgP++)
		     if ((1 << k) & *bitP)
		       *imgP = colorInd;
		   break;
		}
	       bitP++;
	       width2do -= 8;
	     }
	   for (k = 0; k < diffEnd; k++, imgP++)
	     if ((1 << k) & *bitP)
	       *imgP = colorInd;
	 }
     }

 if (!shared)
   undoXImage(&tmpImg);
 XSetForeground(xinfo.dpy, xinfo.gc1bit, 0);
 XFillRectangle(xinfo.dpy, pixy, xinfo.gc1bit, x, y, width, height);
 XSetForeground(xinfo.dpy, xinfo.gc1bit, 1);
}

DMISC void freeMainImage(XImage ** img)
{
 undoXImage(img);
 FreePixmap(xinfo.dpy, ApData.pixy);
 ApData.pixy = 0;
}

DMISC XImage *createMainImage(unsigned int width, unsigned int height)
{
 XImage *tmpImg;

 tmpImg = NULL;
 tmpImg = doXImage(8, width, height);
 if (tmpImg == NULL)
   return(NULL);
 ApData.pixy = CreatePixmap(xinfo.dpy, RootWindow(xinfo.dpy, xinfo.screenNo),
  width, height, 1);
 if (ApData.pixy == 0)
  {
   undoXImage(&tmpImg);
   return(NULL);
  }
 /*clear picture */
 memset(tmpImg->data, xinfo.allocCells[ColorMatrix.BckColorInd].pixel,
  tmpImg->bytes_per_line * height);
 /*clear pixmap */
 XSetForeground(xinfo.dpy, xinfo.gc1bit, 0);
 XFillRectangle(xinfo.dpy, ApData.pixy, xinfo.gc1bit, 0, 0, width, height);
 XSetForeground(xinfo.dpy, xinfo.gc1bit, 1);
 return(tmpImg);
}

/* procedure which makes XPutImage depending on the depth of the drawable */
/* if depth == 8, uses normal XPutImage, else it creates a second XImage */
/* in the correct depth and puts it on drawable */
/* all this because image is always of depth 8 while drawable can have bigger depth */
DMISC void MyPutImage(Drawable drawable, GC gc, XImage *image, int src_x,
 int src_y, int dest_x, int dest_y, unsigned int width,
 unsigned int height, Region region)
{
 if (!image)
  return;
 if (xinfo.trueColor)
  {
   int dummy1;
   unsigned int dummy2, depth;
   Window win;

   if (XGetGeometry(xinfo.dpy, drawable, &win, &dummy1, &dummy1, &dummy2,
    &dummy2, &dummy2, &depth) == 0)
     return;
   if (depth == image->depth)
    {
     if (image==ApData.scrImg)
      {
       if ((!pixmapIsXimage)||(drawable!=ApData.scrPix))
        xPutBuffer(drawable, src_x, src_y, dest_x, dest_y, width, height);
      }
     else
      XPutImage(xinfo.dpy, drawable, gc, image, src_x, src_y, dest_x, dest_y,
       width, height);
    }
   else if (depth > 8)
    {
     XGCValues values;

     convertImg2TrueColor(image, ApData.scrImg, src_x, src_y, width, height, 1, 1);
     if (region)
      XSetRegion(xinfo.dpy, xinfo.gc, region);
     XGetGCValues(xinfo.dpy, gc, GCClipXOrigin | GCClipYOrigin, &values);
     XChangeGC(xinfo.dpy, xinfo.gc, GCClipXOrigin | GCClipYOrigin, &values);
     XPutImage(xinfo.dpy, drawable, xinfo.gc, ApData.scrImg, 0, 0, dest_x,
      dest_y, width, height);
     values.clip_x_origin = values.clip_y_origin = 0;
     XChangeGC(xinfo.dpy, xinfo.gc, GCClipXOrigin | GCClipYOrigin, &values);
     XSetClipMask(xinfo.dpy, xinfo.gc, None);
    }
  }
 else
  XPutImage(xinfo.dpy, drawable, gc, image, src_x, src_y, dest_x, dest_y, width, height);
}

/* return a value of timeout for Motif XtAddTimeout functions depending on */
/* how busy is the system. We need to give XtMainloop some time to do its job. */
/* Test it by round trip XSync. Minimum is 1 (1000/sec), max is 100 (10/sec) */
DMISC int adaptiveTimeout()
{
 int ms1, ms2;
 struct timeval tv;
 struct timezone tz;

 tz.tz_minuteswest = 0;
 tz.tz_dsttime = 0;
 gettimeofday(&tv, &tz);
 ms1 = tv.tv_usec;
 XSync(xinfo.dpy, FALSE);
 gettimeofday(&tv, &tz);
 ms2 = tv.tv_usec;
 /* 500 is a round trip on PII233, which works fine with value of 1 */
 /* as delay value, so scale it to this value */
 return(max(1, min(100, abs(ms2 - ms1) / 500)));
}

/* sets the passed string to widget as dialogTitle */
DMISC void setDialogTitle(Widget w, char *str)
{
 while ((!XmIsBulletinBoard(w)) && XtIsWidget(w))
  w=XtParent(w);
 xtVaSetString(w, XmNdialogTitle, str);
}

/* set icon image and name */
DMISC void setIcon(Widget shell, XImage *xim)
{
 int i,j, wid, hei;
 Pixmap pix;
 XIconSize *sizeList;
 XGCValues values;
 GC myGC;
 unsigned char val;
 char name[250];

 if (!xim)
  return;
 if (XGetIconSizes(xinfo.dpy, xinfo.W, &sizeList, &i))
  { wid = sizeList->max_width; hei = sizeList->max_height; XFree(sizeList); }
 else
  wid=hei=64;
 myGC = XCreateGC(xinfo.dpy, xinfo.W, 0, &values);

 pix = safeCreatePixmap(xinfo.dpy, xinfo.W, wid, hei, xinfo.depth);
 if (pix)
  for (i=0; i<wid; i++)
   for (j=0; j<hei; j++)
    {
     val = XGetPixel(xim, (i*xim->width)/wid, (j*xim->height)/hei);
     XSetForeground(xinfo.dpy, myGC, (xinfo.trueColor) ? xinfo.pixelTC[val] : val);
     XDrawPoint(xinfo.dpy, pix, myGC, i, j);
    }
 else
  return;
 sprintf(name, "%s: %s", ApData.version, baseName(ApData.fname));
 XtVaSetValues(shell, XmNiconPixmap, pix, XmNiconName, name, NULL);
 XFreeGC(xinfo.dpy, myGC);
}

/* returns the position of widget in the menu */
DMISC int getMenuIndexFromWidget(Widget w)
{
 int i, numChildren;
 WidgetList wList; /* it is not allocated by getValue, so do not free it !!! */
 int selected=0;

 XtVaGetValues(XtParent(w), XmNnumChildren, &numChildren, XmNchildren, &wList, NULL);
 for (i=0; i<numChildren; i++)
  if (w==wList[i])
   selected=i;
 return(selected);
}

/* sets the state on toggle button widgets */
DMISC void setState(Widget w, int state)
{
 XtVaSetValues(w, XmNset, state, NULL);
}

/* makes the widget sensitive to input */
DMISC void setSensitive(Widget w, int state)
{
 XtVaSetValues(w, XmNsensitive, state, NULL);
}
