/* epson.h header file for common printer functions and global variables */
/* in epson.c								*/

#ifndef Boolean
typedef int Boolean;
#define True 1
#define False 0
#endif

#define CYANOFFSET 0
#define MAGENTAOFFSET 1
#define YELLOWOFFSET 2
#define BLACKOFFSET 3
#define CYANLIGHTOFFSET 4
#define MAGENTALIGHTOFFSET 5


#define NO_COMPRESSION 0
#define COMPRESSION 1

#define DPI_360 10
#define DPI_720 5
#define DPI_180 20

#define DEFAULT_DOT_SIZE	0
#define MICRO_DOT_SIZE		1
#define SINGLE_DOT_SIZE		2
#define DOUBLE_DOT_SIZE		3
#define STANDARD_DOT_SIZE	2
#define SUPER_MICRO_DOT_SIZE	4
#define VARIABLE_DOT_SIZE_1	16
#define VARIABLE_DOT_SIZE_2	17
#define MULTISHOT_DOT_SIZE	18

#define MONOCHROME_MODE 1
#define COLOR_MODE 2

#define DEFAULT_SPEED 0
#define SPEED_20IPS 1
#define SPEED_10IPS 2

#define SGN(a)  (((a)<0) ? -1: +1)

#define GRAPHICSMODE_LEN 6
#define SETUNIT_LEN 6
#define SETPAGELENGTHU_LEN 7
#define SETDIRECTION_LEN 3
#define SETPAGEFORMAT_LEN 9
#define SETABSYPOS_LEN 6

/* these functions are from print.c */

/* registers printer with printer driver */
extern void registerPrinter (getPrinterInfoProc getPrtInfo,
	 initPrintProc initPrt, printBandProc prtBand, endPrintProc endPrt);

/* put the files in the print queue of the currently selected printer */
extern int enqueuePrintFile(char *file1);

/* cleanup of temporary files */
extern void removeFileFreeFilename(char *filename);

/* allocates and returns the dither matrix of size 256*256 and distribution */
/* of values ranging from 0 to 255 according to the value of distri[256] */
/* array. The sum of all values in the distribution array must be 65536 */
extern unsigned char *makeDither(unsigned short *distr);

/*-----------------------COMMANDS------------------------------------*/
extern unsigned char *ResetPrn(unsigned char *bufPtr);
extern unsigned char *FormFeed(unsigned char *bufPtr);
extern unsigned char *SetColorCMYK(unsigned char CMYKcomp, unsigned char *bufPtr);
extern unsigned char *SetColorCCMMYK(unsigned char CCMMYKcomp, unsigned char *bufPtr);
extern unsigned char *SetGraphicsMode(unsigned char *bufPtr);
extern unsigned char *SetMicroWeave(Boolean microWeaveOn, unsigned char *bufPtr);
extern unsigned char *SetUnit(unsigned char unitS, unsigned char *bufPtr);
extern unsigned char *SetPageLengthU(int PageLength, unsigned char *bufPtr);
extern unsigned char *SetDirection(unsigned char d, unsigned char *bufPtr);
extern unsigned char *SetPageFormat(int topM, int bottomM, unsigned char *bufPtr);
extern unsigned char *SetAbsXPos(int position, unsigned char *bufPtr);
extern unsigned char *SetRelXPos(int position, unsigned char *bufPtr);
extern unsigned char *SetRelXPosNew(int position, unsigned int unit, unsigned char *bufPtr);
extern unsigned char *SetAbsYPos(int position, unsigned char *bufPtr);
extern unsigned char *SetRelYPos(int position, unsigned char *bufPtr);
extern unsigned char *SetDotSize(unsigned char dotSize, unsigned char *bufPtr);
extern unsigned char *PrintRasterGraphics(char compression, char vDotDens, char hDotDens,
		     char m, int width, unsigned char *bufPtr);
unsigned char *SetResolutionRasterImage(int xdpi, int ydpi, unsigned char *bufPtr);
extern unsigned char *TransferRasterImage(char color, char compression, char bitsPerPixel, 
 int width, int height, unsigned char *bufPtr);
extern unsigned char *SelectPrintingMode(char mode, unsigned char *bufPtr);
extern unsigned char *SelectPrintingSpeed(char speed, unsigned char *bufPtr);

/*-----------------------COMMON FUNCTIONS------------------------------------*/
extern unsigned char *compressLine(unsigned char *lnPtr, unsigned char *printBuf, 
 unsigned char *RLLBuf, int diff);

extern unsigned char *linearizeDither(float *point, float bleed);
extern void setDots (unsigned int *field, unsigned char *m);
