/* posts.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "print.h"

#define IMAGE_CODEWIDTH 60
#define MAX_CELLS  256
#define NCOLORS 256
#define DEPTH 8

/* registers printer with printer driver */
extern void registerPrinter(getPrinterInfoProc getPrtInfo,
	initPrintProc initPrt, printBandProc prtBand, endPrintProc endPrt);

/* put the files in the print queue of the currently selected printer */
extern int enqueuePrintFile(char *file1);

/* cleanup of temporary files */
extern void removeFileFreeFilename(char *filename);

static FILE *fout;
static char *fname;
static int printModeType;
static int Height;

static int initPrint(int printType, int complexImage, int xpixels,
	int ypixels, int xdpi, int ydpi, int xmargin, int ymargin)
{
 double pageWidth, pageHeight;
 double horizMargin, vertMargin;
 double horizInset, vertInset;
 double xscale, yscale;
 time_t t;

 printModeType = printType;
 if (printModeType > IMG_TRUE_COL_DITH)
  return(PRT_WRONG_IMG);
 fname = tempnam("/tmp","prt");
 if ((fname == NULL) || ((fout = fopen(fname, "w" )) == NULL))
  return(PRT_IO_ERROR);

 Height = ypixels;

 fprintf(fout, "%%!PS-Adobe-2.0\n");
 fprintf(fout, "%%%%Pages: 1\n");
 fprintf(fout, "%%%%File: %s\n", fname);
 time(&t);
 fprintf(fout, "%%%%CreationDate: %s", ctime(&t));
 fprintf(fout, "%%%%EndComments\n");
 fprintf(fout, "%%%%EndProlog\n");
 fprintf(fout, "%%%%Page: 1 1\n");

 fputs("\ngsave\n\n", fout);
 fputs("/inch { 72 mul } def\n", fout);

 pageWidth = 8.2;   /* inches are assumed A4 */
 pageHeight = 11.7;
 horizMargin = ((double) xmargin) / ((double) (xdpi/1000));
 vertMargin =  ((double) ymargin) / ((double) (ydpi/1000));
 horizInset = pageWidth - horizMargin - horizMargin;
 vertInset  = pageHeight - vertMargin - vertMargin;
 xscale = ((double) xpixels) / ((double) (xdpi/1000));
 yscale = ((double) ypixels) / ((double) (ydpi/1000));
 if (xscale > horizInset)
  {
   yscale *= horizInset / xscale;
   xscale = horizInset;
  }
 if (yscale > vertInset)
  {
   xscale *= vertInset / yscale;
   yscale = vertInset;
  }

 fprintf(fout, "%1.2g inch %1.2g inch translate\n", 
         horizMargin, pageHeight - vertMargin - yscale);
 fprintf(fout, "%1.2g inch %1.2g inch scale\n\n", xscale, yscale);
 fputs("/endstr 1 string def\n", fout);
 fputs("/ccount 0 def\n", fout);

 /* non-encoded output:
    define a drawcolorimage procedure geared to this image */
 fprintf(fout, "/buffer %d string def\n", 1);
 fprintf(fout, "/line   %d string def\n", xpixels);
 fprintf(fout, "/rgbmap %d string def\n", NCOLORS * 3);
 fputs("/onepixel 3 string store\n\n", fout);

 fputs("/drawcolormappedimage {\n", fout);
 fputs("  systemdict /colorimage known userdict /colorimage known or not {\n", fout);
 fputs("    %% convert colors to greyscale\n", fout);
 fprintf(fout,"    /ncolors %d store\n", NCOLORS);
 fputs("    /ridx 0 store\n", fout);
 fputs("    /greys ncolors string def\n", fout);
 fputs("    /linecount 0 store\n", fout);
 fputs("    0 1 ncolors 1 sub {\n", fout);
 fputs("      /gidx exch store\n", fout);
 fputs("      rgbmap ridx get .3 mul\n", fout);
 fputs("      rgbmap ridx 1 add get .59 mul add\n", fout);
 fputs("      rgbmap ridx 2 add get .11 mul add\n", fout);
 fputs("      cvi\n", fout);
 fputs("      /agrey exch store\n", fout);
 fputs("      greys gidx agrey put\n", fout);
 fputs("      /ridx ridx 3 add store\n", fout);
 fputs("    } for\n", fout);
 fprintf(fout, "    %d %d %d", xpixels, ypixels, DEPTH);
 fprintf(fout, "  [%d 0 0 -%d 0 %d]\n", xpixels, ypixels, ypixels);
 fputs("    %% define a block of code to read and decode the input stream\n", fout);
 fputs("    { currentfile read not { 0 } if\n", fout);
 fputs("      greys exch 1 getinterval\n", fout);
 fputs("    } image\n", fout);
 fputs("  } {  %% ifelse\n", fout);
 fprintf(fout, "    %d %d %d", xpixels, ypixels, DEPTH);
 fprintf(fout, "  [%d 0 0 -%d 0 %d]\n", xpixels, ypixels, ypixels);
 fputs("    %% define a block of code to read and decode the input stream\n", fout);
 fputs("    { rgbmap   currentfile read not { 0 } if   3 mul 3 getinterval\n", fout);
 fputs("    } false 3 colorimage\n", fout);
 fputs("  } ifelse\n", fout);
 fputs("} bind def\n\n", fout);

 return(PRT_OK);
}

static int printBand(int xpixels, char *data, int bpl, int ypos,
	int lines, char bandType, unsigned char *cmap)
{
 register int x, y;

 if (data == NULL)
  return(PRT_INTERNAL_ERROR);

 if (printModeType < bandType)
  return(PRT_WRONG_IMG);

 if (ypos == 0)
  {
   /* write the rgb map */
   fputs("\n%% get the rgb map\n", fout);
   fputs("currentfile rgbmap readhexstring\n", fout);
   for (x = 0; x < NCOLORS; x++)
    fprintf(fout, "%02x%02x%02x\n", cmap[3*x], cmap[3*x+1], cmap[3*x+2]);
   fputs("pop pop\n\n", fout);

   fprintf(fout, "%%%%BeginData: %d Binary Bytes\n", 22 + Height * xpixels);
   fputs("drawcolormappedimage\n", fout);
  }

 /* write the map indexes */
 for (y = 0; y < lines; y++)
  for (x = 0; x < xpixels; x++)
   fputc(data[bpl * y + x], fout);

 return(PRT_OK);
}

static int endPrint(int printStatus)
{
 if (printStatus == PRT_OK)
  if (printModeType <= IMG_TRUE_COL_DITH)
   {
    fputs("\n\n%%EndData\n", fout);
    fputs("\ngrestore\nshowpage\n%%Trailer\n", fout);
    fclose(fout);
    printStatus = enqueuePrintFile(fname);
   }
  else 
   printStatus = PRT_WRONG_IMG;
 else
  if (printModeType <= IMG_TRUE_COL_DITH)
   fclose(fout);
  else
   printStatus = PRT_WRONG_IMG;

 removeFileFreeFilename(fname);

 return(printStatus);
}

/* fill in the preallocated printer modes for PostScript */
static void getPrtInfo(printerMode *printerMod, int *numModes,
	printerDensity *prtDensityX, int *numDensitX,
	printerDensity *prtDensityY, int *numDensitY,
	int *nonprintableTop, int *nonprintableBottom, int *nonprintableLeft,
	int *dpiVariable, char **prtName, char **profileName)
{
 int i;

 printerMod[0].printType = IMG_CMAP;

 *numModes=1;

 *numDensitY=5;

 *numDensitX = *numDensitY = 6;
 for (i=0; i<6; i++)
  {
   prtDensityX[i].pixels = prtDensityY[i].pixels = 20000;
   prtDensityX[i].dpi = prtDensityY[i].dpi = 100000*(i+1);
  }

 *nonprintableTop=0;
 *nonprintableBottom=0;
 *nonprintableLeft=0;
 *dpiVariable=FALSE;
 *prtName="PostScript";
 *profileName=NULL;
}

/* register printing to PostScript file */
void registerPostScript()
{
 registerPrinter(*getPrtInfo, *initPrint, *printBand, *endPrint);
}
